
local Healcsi = Healcsi
local Media   = Healcsi.Media
local Gui     = Healcsi.Gui
local L       = Healcsi.Locale
local Editor  = {}

Healcsi.Editor = Editor

local MAX_SPELL_ID = 250000 -- for legion

Editor.frame = CreateFrame("Frame")

Editor.frame:SetScript("OnEvent", function(self, event, ...)
  self[event](...)
end)

Editor.frame:RegisterEvent("PLAYER_REGEN_ENABLED")
Editor.frame:RegisterEvent("PLAYER_REGEN_DISABLED")

function Editor.frame:PLAYER_REGEN_ENABLED(...)
	Editor.InCombat = false
	Healcsi:BroadcastEvent("OutOfCombat")
	Healcsi:BroadcastEvent("CombatStateChanged")
end

function Editor.frame:PLAYER_REGEN_DISABLED(...)
	Editor.InCombat = true
	Healcsi:BroadcastEvent("InCombat")
	Healcsi:BroadcastEvent("CombatStateChanged")
end

-- 

function Editor:CreateInlineGroup(frame, isFullWidth, width, margin, bgMargin, bgColor)

	margin   = margin or {0, 0, 0, 0}
	bgMargin = bgMargin or {0, 5, 0, 5}
	bgColor  = bgColor or {r=0,g=0,b=0,a=0}

	local group = Gui:Create("Healcsi-InlineGroup")

	group:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then group:SetWidth(width) end

	group:SetLayout("Flow")
	group:SetMargin(margin)
	group:SetBackground(bgMargin, "Solid", bgColor)

	frame:AddChild(group)

	return group
end

function Editor:CreateGroup(frame, isFullWidth, width, margin, bgMargin, colorName, alpha)

	margin    = margin or {0, 0, 0, 0}
	bgMargin  = bgMargin or {0, 5, 0, 5}
	colorName = colorName or "DarkBlue"

	local group = Gui:Create("Healcsi-InlineGroup")

	group:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then group:SetWidth(width) end

	group:SetLayout("List")
	group:SetMargin(margin)
	group:SetBackground(bgMargin, "Solid", colorName, alpha or 0)

	frame:AddChild(group)

	return group

end

function Editor:CreateButton(frame, labelText)
	local button = Gui:Create("Healcsi-Button")
	button:SetText(labelText)
	frame:AddChild(button)

	return button
end

function Editor:CreateEditBox(frame, labelText)
	local editBox = Gui:Create("Healcsi-EditBox")
	editBox:SetLabel(labelText)

	frame:AddChild(editBox)

	return editBox
end

function Editor:CreateHeaderlabel(frame, labelText, colorName, alpha)
	local label = Gui:Create("Healcsi-Label")
	label:SetFont(Healcsi:GetFont("Friz Quadrata TT"), 16)
	label:SetFontColor(Healcsi:GetColor(colorName or "BlueHovered", alpha or 1))
	label:SetText(labelText)
	label:SetFullWidth(true)

	if(frame) then frame:AddChild(label) end

	return label
end

function Editor:CreateLabel(frame, isFullWidth, width, labelText, colorName, alpha, size)
	local label = Gui:Create("Healcsi-Label")

	label:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then label:SetWidth(width) end

	label:SetFont(Healcsi:GetFont("Friz Quadrata TT"), size or 14)
	label:SetFontColor(Healcsi:GetColor(colorName or "BlueHovered", alpha or 1))
	label:SetText(labelText)

	if(frame) then frame:AddChild(label) end

	return label
end

function Editor:CreateCheckBox(frame, isFullWidth, width, labelText, db, key, callbackName, broadcast, disableInCombat, colorName, alpha)
	local checkBox = Gui:Create("Healcsi-CheckBox")

	checkBox:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then checkBox:SetWidth(width) end
	checkBox:SetHeight(30)

	checkBox:SetLabel(labelText)
	checkBox:SetLabelFontColor(Healcsi:GetColor(colorName or "DarkBlue", alpha or 1))

	checkBox:SetValue(db[key] or false)

	checkBox:SetCallback("OnValueChanged", function(self, event, value)
		db[key] = value
		self:SetValue(db[key])
		Healcsi:BroadcastEvent(callbackName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	disableInCombat = disableInCombat or false

	Healcsi:RegisterEventCallback("CombatStateChanged", function()
		if (Editor.InCombat) then
			if (disableInCombat) then
				checkBox:SetDisabled(true)
			end
		else
			checkBox:SetDisabled(false)
		end
	end)

	if (Editor.InCombat) then
		if (disableInCombat) then
			checkBox:SetDisabled(true)
		end
	end

	if(frame) then frame:AddChild(checkBox) end

	return checkBox
end

function Editor:CreateColorPicker(frame, isFullWidth, width, hasAlpha, labelText, db, key, callbackName, broadcast, disableInCombat)
	local colorPicker = Gui:Create("Healcsi-ColorPicker")

	colorPicker:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then colorPicker:SetWidth(width) end

	colorPicker:SetLabel(labelText)
	colorPicker:SetHasAlpha(hasAlpha or true)
	colorPicker:SetHandlerName(callbackName)
	colorPicker:SetRGBAColorObj(db[key])

	colorPicker:SetCallback("OnValueChanged", function(self, event, red, green, blue, alpha)
		if (hasAlpha) then
			db[key] = {r=red,g=green,b=blue,a=alpha}
		else
			db[key] = {r=red,g=green,b=blue}
		end
		self:SetRGBAColorObj(db[key])
		Healcsi:BroadcastEvent(callbackName)

		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	--disableInCombat = disableInCombat or false
	--TODO: disable in combat stuff

	if(frame) then frame:AddChild(colorPicker) end


	return colorPicker
end

function Editor:CreateSlider(frame, isFullWidth, width, labelText, minVal, maxVal, stepVal, isPercent, db, key, callbackName, broadcast, disableInCombat)
	local slider = Gui:Create("Healcsi-Slider")

	slider:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then slider:SetWidth(width) end

	slider:SetLabel(labelText)
	slider:SetSliderValues(minVal, maxVal, stepVal)
	slider:SetIsPercent(isPercent)
	slider:SetValue(db[key])
	slider:SetCallback("OnValueChanged", function(self, event, value)
		db[key] = value
		self:SetValue(db[key])
		Healcsi:BroadcastEvent(callbackName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	disableInCombat = disableInCombat or false

	Healcsi:RegisterEventCallback("CombatStateChanged", function()
		if (Editor.InCombat) then
			if (disableInCombat) then
				slider:SetDisabled(true)
			end
		else
			slider:SetDisabled(false)
		end
	end)

	if (Editor.InCombat) then
		if (disableInCombat) then
			slider:SetDisabled(true)
		end
	end

	if(frame) then frame:AddChild(slider) end

	return slider
end

function Editor:CreateDropDown(frame, isFullWidth, width, labelText, list, db, key, callbackName, broadcast, disableInCombat)
	local dropDown = Gui:Create("Healcsi-Dropdown")

	dropDown:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then dropDown:SetWidth(width) end

	dropDown:SetLabel(labelText)
	dropDown:SetList(list)
	dropDown:SetValue(db[key])
	dropDown:SetCallback("OnValueChanged", function(self, event, value)
		db[key] = value
		self:SetValue(db[key])
		Healcsi:BroadcastEvent(callbackName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	disableInCombat = disableInCombat or false

	Healcsi:RegisterEventCallback("CombatStateChanged", function()
		if (Editor.InCombat) then
			if (disableInCombat) then
				dropDown:SetDisabled(true)
			end
		else
			dropDown:SetDisabled(false)
		end
	end)

	if (Editor.InCombat) then
		if (disableInCombat) then
			dropDown:SetDisabled(true)
		end
	end

	if(frame) then frame:AddChild(dropDown) end

	return dropDown
end

function Editor:CreateHorizLine(frame)
	local horizLine = Gui:Create("Healcsi-HorizLine")
	frame:AddChild(horizLine)

	return horizLine
end

function Editor:CreateBackground(frame, isFullWidth, width, labelText, db, key, callbackName, broadcast, disableInCombat)
	local background = Gui:Create("Healcsi-Background")

	background:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then background:SetWidth(width) end

	background:SetHeight(54)
	background:SetList(Media:HashTable("background"))
	background:SetLabel(labelText)
	background:SetValue(db[key])
	background:SetCallback("OnValueChanged", function(self, event, value)
		db[key] = value
		self:SetValue(db[key])
		Healcsi:BroadcastEvent(callbackName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	frame:AddChild(background)

	return background
end

function Editor:CreateBorder(frame, isFullWidth, width, labelText, db, key, callbackName, broadcast, disableInCombat)
	local border = Gui:Create("Healcsi-Border")

	border:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then border:SetWidth(width) end

	border:SetHeight(54)
	border:SetList(Media:HashTable("border"))
	border:SetLabel(labelText)
	border:SetValue(db[key])
	border:SetCallback("OnValueChanged", function(self, event, value)
		db[key] = value
		self:SetValue(db[key])
		Healcsi:BroadcastEvent(callbackName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	frame:AddChild(border)

	return border
end

function Editor:CreateStatusbar(frame, isFullWidth, width, labelText, db, key, callbackName, broadcast, disableInCombat)
	local statusbar = Gui:Create("Healcsi-Statusbar")

	statusbar:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then statusbar:SetWidth(width) end

	--statusbar:SetHeight(130) -- caused issues
	statusbar:SetList(Media:HashTable("statusbar"))
	statusbar:SetLabel(labelText)
	statusbar:SetValue(db[key])
	statusbar:SetCallback("OnValueChanged", function(self, event, value)
		db[key] = value
		self:SetValue(db[key])
		Healcsi:BroadcastEvent(callbackName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	frame:AddChild(statusbar)

	return statusbar
end

function Editor:CreateFont(frame, isFullWidth, width, labelText, db, key, callbackName, broadcast, disableInCombat)
	local font = Gui:Create("Healcsi-Font")

	font:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then font:SetWidth(width) end

	--font:SetHeight(130) -- caused issues
	font:SetList(Media:HashTable("font"))
	font:SetLabel(labelText)
	font:SetValue(db[key])
	font:SetCallback("OnValueChanged", function(self, event, value)
		db[key] = value
		self:SetValue(db[key])
		Healcsi:BroadcastEvent(callbackName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	frame:AddChild(font)

	return font
end

function Editor:CreateCustomSelector(frame, isFullWidth, width, labelText, list, db, key, callbackName, broadcast, disableInCombat)
	local customBorder = Gui:Create("Healcsi-Background")

	customBorder:SetFullWidth(isFullWidth)
	if (not(isFullWidth) and width) then customBorder:SetWidth(width) end

	customBorder:SetHeight(54)
	customBorder:SetList(list)
	customBorder:SetLabel(labelText)
	customBorder:SetValue(db[key])
	customBorder:SetCallback("OnValueChanged", function(self, event, value)
		db[key] = value
		self:SetValue(db[key])
		Healcsi:BroadcastEvent(callbackName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
	end)

	frame:AddChild(customBorder)

	return customBorder
end

function Editor:UpdateFrameDecoration(data)
	data.holder:ReleaseChildren()
	if (data.db.showBackground) then
		data = Editor:CreateFrameDecorationEditor(data.frame, data.db, data.id, data.broadcast, data)
	else
		Editor:CreateLabel(data.holder, true, 300, L["EnableBgToActivate"])
	end
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateFrameDecorationEditor(frame, db, id, broadcast, hasPrevData)

	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		holder:SetLineBackground(4, 0, "Solid", "DarkBlue", 0.8)
		data.holder = holder
	end

	local line1 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})

	Editor:CreateColorPicker(line1, false, 300, true, L["BackgroundColor"], db.background, "color", "UnitFrames"..id.."Color", broadcast)
	Editor:CreateBackground(line1, false, 300, L["BackgroundTex"], db.background, "file", "UnitFramesBgFile", broadcast)

	local tileOptions = {
		["tile"]    = "Tile",
		["stretch"] = "Stretch"
	}
	local tileDropDown = Editor:CreateDropDown(line1, false, 300, L["BgTexMode"], tileOptions, db.background, "tile", "UnitFrames"..id.."Tile", broadcast)

	if (db.background.tile == "tile") then
		Editor:CreateSlider(line1, false, 300, L["TileSize"], 0, 2000, 1, false, db.background, "tileSize", "UnitFrames"..id.."TileSize", broadcast)
	end

	Editor:CreateHorizLine(data.holder)

	local line2 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})

	local icb = Editor:CreateCheckBox(line2, true, 300, L["CommonInset"], db.background, "commonInset", "UnitFrames"..id.."CommonInset", broadcast)
	icb:SetDescription(L["BackgroundInsetDesc"])

	if (db.background.commonInset) then
		Editor:CreateSlider(line2, false, 300, L["OneInset"], 0, 50, 0.5, false, db.background.insets, "top", "UnitFrames"..id.."Inset", broadcast)
	else
		Editor:CreateSlider(line2, false, 300, L["TopInset"],    0, 50, 0.5, false, db.background.insets, "top", "UnitFrames"..id.."InsetTop", broadcast)
		Editor:CreateSlider(line2, false, 300, L["RightInset"],  0, 50, 0.5, false, db.background.insets, "right", "UnitFrames"..id.."InsetRight", broadcast)
		Editor:CreateSlider(line2, false, 300, L["BottomInset"], 0, 50, 0.5, false, db.background.insets, "bottom", "UnitFrames"..id.."InsetBottom", broadcast)
		Editor:CreateSlider(line2, false, 300, L["LeftInset"],   0, 50, 0.5, false, db.background.insets, "left", "UnitFrames"..id.."InsetLeft", broadcast)
	end

	Editor:CreateHorizLine(data.holder)

	local line3 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})

	Editor:CreateColorPicker(line3, false, 300, true, L["BorderColor"], db.border, "color", "UnitFrames"..id.."BorderColor", broadcast)
	Editor:CreateBorder(line3, false, 300, L["BorderTex"], db.border, "file", "UnitFrames"..id.."BorderFile", broadcast)
	Editor:CreateSlider(line3, false, 300, L["BorderThickness"], 0, 50, 0.5, false, db.border, "edgeSize", "UnitFrames"..id.."BorderSize", broadcast)

	tileDropDown:SetCallback("OnValueChanged", function(self, event, value)
		db.background.tile = value
		Healcsi:BroadcastEvent("UnitFrames"..id.."Tile")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateFrameDecoration(data)
	end)

	icb:SetCallback("OnValueChanged", function(self, event, value)
		db.background.commonInset = value
		Healcsi:BroadcastEvent("UnitFrames"..id.."CommonInset")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateFrameDecoration(data)
	end)

	return data
end

function Editor:UpdateUnitHealthDecoration(data)
	data.holder:ReleaseChildren()
	if (data.db.show) then
		data = Editor:CreateUnitHealthDecorationEditor(data.frame, data.db, data.id, data.broadcast, data)
	else
		Editor:CreateLabel(data.holder, true, 300, L["EnableHealthBgToActivate"])
	end
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateShadeEditor(frame, db, id, labelText, broadcast)

	local info = {}

	if (labelText) then
		Editor:CreateHeaderlabel(frame, labelText)
	end

	local colorModes = {
		["class"]  = "Class",
		["health"] = "Health",
		["fix"]    = "Fix"
	}
	info.colorMode = Editor:CreateDropDown(frame, false, 300, L["ColorMode"], colorModes, db.color, "mode", "UnitButton"..id.."ColorMode", broadcast)
	info.shadeMode = nil

	local line1 = Editor:CreateGroup(frame, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})

	if (db.color.mode == "fix") then
		Editor:CreateColorPicker(line1, false, 300, true, L["FixColor"], db.color, "fix", "UnitButton"..id.."FixColor", broadcast)
	elseif (db.color.mode == "class") or (db.color.mode == "health") then

		local colorShadeModes = {
			["rgbShade"]  = "RGB shade",
			["hsvShade"] = "HSV shade"
		}
		info.shadeMode = Editor:CreateDropDown(line1, false, 300, L["ColorShadeMode"], colorShadeModes, db.color, "shadeMode", "UnitButton"..id.."ShadeColor", broadcast)

		local line2 = Editor:CreateGroup(line1, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})

		if (db.color.shadeMode == "rgbShade") then
			Editor:CreateSlider(line2, true, 300, L["Red"], -1, 1, 0.01, true, db.color.rgb, "r", "UnitButton"..id.."Red", broadcast)
			Editor:CreateSlider(line2, true, 300, L["Green"], -1, 1, 0.01, true, db.color.rgb, "g", "UnitButton"..id.."Green", broadcast)
			Editor:CreateSlider(line2, true, 300, L["Blue"], -1, 1, 0.01, true, db.color.rgb, "b", "UnitButton"..id.."Blue", broadcast)
		elseif (db.color.shadeMode == "hsvShade") then
			Editor:CreateSlider(line2, true, 300, L["Hue"], -360, 360, 1, false, db.color.hsv, "h", "UnitButton"..id.."Hue", broadcast)
			Editor:CreateSlider(line2, true, 300, L["Saturation"], -1, 1, 0.01, true, db.color.hsv, "s", "UnitButton"..id.."Saturation", broadcast)
			Editor:CreateSlider(line2, true, 300, L["Lightness"], -1, 1, 0.01, true, db.color.hsv, "v", "UnitButton"..id.."Lightness", broadcast)
		end
		Editor:CreateSlider(line2, true, 300, L["Alpha"], 0, 1, 0.01, true, db.color, "alpha", "UnitButton"..id.."Alpha", broadcast)
	end

	return info
end

function Editor:CreatePowerShadeEditor(frame, db, id, labelText, broadcast)

	local info = {}

	if (labelText) then
		Editor:CreateHeaderlabel(frame, labelText)
	end

	local colorModes = {
		["auto"]  = "Auto",
		["fix"]    = "Fix"
	}
	info.colorMode = Editor:CreateDropDown(frame, false, 300, L["ColorMode"], colorModes, db.color, "mode", "UnitButton"..id.."ColorMode", broadcast)
	info.shadeMode = nil

	local line1 = Editor:CreateGroup(frame, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})

	if (db.color.mode == "fix") then
		Editor:CreateColorPicker(line1, false, 300, true, L["FixColor"], db.color, "fix", "UnitButton"..id.."FixColor", broadcast)
	elseif (db.color.mode == "auto") then

		local colorShadeModes = {
			["rgbShade"]  = "RGB shade",
			["hsvShade"] = "HSV shade"
		}
		info.shadeMode = Editor:CreateDropDown(line1, false, 300, L["ColorShadeMode"], colorShadeModes, db.color, "shadeMode", "UnitButton"..id.."ShadeColor", broadcast)

		local line2 = Editor:CreateGroup(line1, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})

		if (db.color.shadeMode == "rgbShade") then
			Editor:CreateSlider(line2, true, 300, L["Red"], -1, 1, 0.01, true, db.color.rgb, "r", "UnitButton"..id.."Red", broadcast)
			Editor:CreateSlider(line2, true, 300, L["Green"], -1, 1, 0.01, true, db.color.rgb, "g", "UnitButton"..id.."Green", broadcast)
			Editor:CreateSlider(line2, true, 300, L["Blue"], -1, 1, 0.01, true, db.color.rgb, "b", "UnitButton"..id.."Blue", broadcast)
		elseif (db.color.shadeMode == "hsvShade") then
			Editor:CreateSlider(line2, true, 300, L["Hue"], -360, 360, 1, false, db.color.hsv, "h", "UnitButton"..id.."Hue", broadcast)
			Editor:CreateSlider(line2, true, 300, L["Saturation"], -1, 1, 0.01, true, db.color.hsv, "s", "UnitButton"..id.."Saturation", broadcast)
			Editor:CreateSlider(line2, true, 300, L["Lightness"], -1, 1, 0.01, true, db.color.hsv, "v", "UnitButton"..id.."Lightness", broadcast)
		end
		Editor:CreateSlider(line2, true, 300, L["Alpha"], 0, 1, 0.05, true, db.color, "alpha", "UnitButton"..id.."Alpha", broadcast)
	end

	return info

end

function Editor:CreateUnitHealthDecorationEditor(frame, db, id, broadcast, hasPrevData)

	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		holder:SetLineBackground(4, 0, "Solid", "DarkBlue", 0.8)
		data.holder = holder
	end

	local line1 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})
	Editor:CreateStatusbar(line1, false, 400, L["HealthTex"], db, "file", "UnitButton"..id.."File", broadcast)
	
	Editor:CreateHorizLine(data.holder)
	local line2 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})
	local colorStuff1 = Editor:CreateShadeEditor(line2, db, id.."Normal", "Normal Unit", broadcast)

	colorStuff1.colorMode:SetCallback("OnValueChanged", function(self, event, value)
		db.color.mode = value
		Healcsi:BroadcastEvent("UnitButton"..id.."NormalColorMode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateUnitHealthDecoration(data)
	end)

	if (colorStuff1.shadeMode) then
		colorStuff1.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.shadeMode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."NormalShadeColor")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateUnitHealthDecoration(data)
		end)
	end

	Editor:CreateHorizLine(data.holder)
	local line4 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})
	local colorStuff3 = Editor:CreateShadeEditor(line4, db.color.inVehicle, id.."Vehicle", "Unit in Vehicle", broadcast)

	colorStuff3.colorMode:SetCallback("OnValueChanged", function(self, event, value)
		db.color.inVehicle.color.mode = value
		Healcsi:BroadcastEvent("UnitButton"..id.."VehicleColorMode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateUnitHealthDecoration(data)
	end)

	if (colorStuff3.shadeMode) then
		colorStuff3.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.inVehicle.color.shadeMode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."VehicleShadeColor")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateUnitHealthDecoration(data)
		end)
	end

	Editor:CreateHorizLine(data.holder)
	local line5 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})
	local colorStuff4 = Editor:CreateShadeEditor(line5, db.color.pet, id.."Pet", "Pet", broadcast)

	colorStuff4.colorMode:SetCallback("OnValueChanged", function(self, event, value)
		db.color.pet.color.mode = value
		Healcsi:BroadcastEvent("UnitButton"..id.."PetColorMode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateUnitHealthDecoration(data)
	end)

	if (colorStuff4.shadeMode) then
		colorStuff4.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.pet.color.shadeMode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."PetShadeColor")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateUnitHealthDecoration(data)
		end)
	end

end

function Editor:UpdateUnitPowerDecoration(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateUnitPowerDecorationEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateUnitPowerDecorationEditor(frame, db, id, broadcast, hasPrevData)

	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		holder:SetLineBackground(4, 0, "Solid", "DarkBlue", 0.8)
		data.holder = holder
	end

	local line1 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})
	Editor:CreateStatusbar(line1, false, 400, L["HealthTex"], db, "file", "UnitButton"..id.."File", broadcast)
	
	Editor:CreateHorizLine(data.holder)
	local line2 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})
	local colorStuff1 = Editor:CreatePowerShadeEditor(line2, db, id, nil, broadcast) --"Power")

	colorStuff1.colorMode:SetCallback("OnValueChanged", function(self, event, value)
		db.color.mode = value
		Healcsi:BroadcastEvent("UnitButton"..id.."ColorMode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateUnitPowerDecoration(data)
	end)

	if (colorStuff1.shadeMode) then
		colorStuff1.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.shadeMode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."ShadeColor")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateUnitPowerDecoration(data)
		end)
	end

end

function Editor:UpdateUnitDamageAbsorbDecoration(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateUnitDamageAbsorbDecorationEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateUnitDamageAbsorbBorderEditor(frame, db, data, id, sideName, broadcast)

	Editor:CreateHeaderlabel(frame, L[sideName])

	local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 10, 0})
	local holder = Editor:CreateGroup(padding, true, nil, {10, 0, 10, 0})
	holder:SetLineBackground(4, 0, "Solid", "DarkBlue", 0.8)

	local line1 = Editor:CreateGroup(holder, true, nil, {0, 10, 0, 20})

	local show = Editor:CreateCheckBox(line1, false, 500, L["ShowDmgBorder" .. sideName], db, "show", "UnitButton"..id.."ShowShieldBorder"..sideName, broadcast) --, disableInCombat, colorName, alpha)
	
	show:SetCallback("OnValueChanged", function(self, event, value)
		db.show = value
		Healcsi:BroadcastEvent("UnitButton"..id.."ShowShieldBorder"..sideName)
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateUnitDamageAbsorbDecoration(data)
	end)

	if (db.show) then
		local colorStuff = Editor:CreateShadeEditor(line1, db, id.."Border"..sideName, nil, broadcast)
		colorStuff.colorMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.mode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."BorderColorMode")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateUnitDamageAbsorbDecoration(data)
		end)
		if (colorStuff.shadeMode) then
			colorStuff.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
				db.color.shadeMode = value
				Healcsi:BroadcastEvent("UnitButton"..id.."BorderShadeColor")
				if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
				Editor:UpdateUnitDamageAbsorbDecoration(data)
			end)
		end

		local line2 = Editor:CreateInlineGroup(line1, true, nil, {5, 0, 5, 0})

		Editor:CreateCustomSelector(line2, false, 300, L["DmgAbsorbTex"], Media.borderTextures, db, "file", "UnitButton"..id.."ShieldBorder"..sideName.."File", broadcast)
		Editor:CreateSlider(line2, false, 300, L["Size"], 0, 20, 1, false, db, "size", "UnitButton"..id.."ShieldBorder"..sideName.."FileSize", broadcast)

	end
	
end

function Editor:CreateUnitDamageAbsorbDecorationEditor(frame, db, id, broadcast, hasPrevData)

	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	local show = Editor:CreateCheckBox(data.holder, false, 300, L["ShowDmgAbsorbBar"], db, "showAbsorbBar", "UnitButton"..id.."Show", broadcast)

	if (db.showAbsorbBar) then
		Editor:CreateHorizLine(data.holder)
		local line1 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})
		line1:SetLineBackground(4, 0, "Solid", "DarkBlue", 0.8)
		local colorStuff1 = Editor:CreateShadeEditor(line1, db, id, "Shield", broadcast)

		colorStuff1.colorMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.mode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."ColorMode")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateUnitDamageAbsorbDecoration(data)
		end)

		if (colorStuff1.shadeMode) then
			colorStuff1.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
				db.color.shadeMode = value
				Healcsi:BroadcastEvent("UnitButton"..id.."ShadeColor")
				if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
				Editor:UpdateUnitDamageAbsorbDecoration(data)
			end)
		end

		local line2 = Editor:CreateGroup(line1, true, nil, {3, 10, 0, 10})

		Editor:CreateStatusbar(line2, false, 300, L["DmgAbsorbTex"], db, "file", "UnitButton"..id.."Texture", broadcast)

	else
		-- turned off
	end

	local line2 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 0}, {5, 0, 5, 0})

	Editor:CreateUnitDamageAbsorbBorderEditor(line2, db.border.left,   data, id, "Left", broadcast)
	Editor:CreateUnitDamageAbsorbBorderEditor(line2, db.border.right,  data, id, "Right", broadcast)
	Editor:CreateUnitDamageAbsorbBorderEditor(line2, db.border.top,    data, id, "Top", broadcast)
	Editor:CreateUnitDamageAbsorbBorderEditor(line2, db.border.bottom, data, id, "Bottom", broadcast)


	show:SetCallback("OnValueChanged", function(self, event, value)
		db.showAbsorbBar = value
		Healcsi:BroadcastEvent("UnitButton"..id.."Show")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateUnitDamageAbsorbDecoration(data)
	end)

	--local line1 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10}, {5, 0, 5, 0})
end


function Editor:UpdateFontEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateFontEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateFontEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	local line1 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 10})

	local anchorList = {
		["CENTER"]   = "Center",
		["LEFT"]     = "Left",
		["TOP"]      = "Top",
		["RIGHT"]    = "Right",
		["BOTTOM"]   = "Bottom",

		["TOPLEFT"]  = "Top left",
		["TOPRIGHT"]  = "Top right",
		["BOTTOMLEFT"]  = "Bottom left",
		["BOTTOMRIGHT"]  = "Bottom right"
	}
	Editor:CreateDropDown(line1, false, 300, L["TextAnchor"], anchorList, db, "anchor", "UnitButton"..id.."Anchor", broadcast)

	local line2 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
	Editor:CreateSlider(line2, false, 300, L["OffsetX"], -500, 500, 1, false, db.offset, "x", "UnitButton"..id.."OffsetX", broadcast)
	Editor:CreateSlider(line2, false, 300, L["OffsetY"], -500, 500, 1, false, db.offset, "y", "UnitButton"..id.."OffsetY", broadcast)

	Editor:CreateHorizLine(data.holder)
	
	local line3 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
	Editor:CreateSlider(line3, false, 300, L["FontSize"], 8, 90, 1, false, db.font, "size", "UnitButton"..id.."FontSize", broadcast)
	Editor:CreateFont(line3, false, 300, L["FontType"], db.font, "file", "UnitButton"..id.."FontType", broadcast)
	Editor:CreateCheckBox(line3, false, 500, L["FontOutline"], db.font, "outline", "UnitButton"..id.."FontOutline", broadcast)

	Editor:CreateHorizLine(data.holder)

	local line4 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
	local colorStuff1 = Editor:CreateShadeEditor(line4, db, "NameText", nil, broadcast)

	Editor:CreateHorizLine(line4)

	Editor:CreateHeaderlabel(line4, L["TextShadow"])

	local colorStuff2= Editor:CreateShadeEditor(line4, db.font.shadow, "NameTextShadow", nil, broadcast)
	Editor:CreateSlider(line4, false, 300, L["ShadowOffsetX"], -5, 5, 1, false, db.font.shadow.offset, "x", "UnitButton"..id.."ShadowOffsetX", broadcast)
	Editor:CreateSlider(line4, false, 300, L["ShadowOffsetY"], -5, 5, 1, false, db.font.shadow.offset, "y", "UnitButton"..id.."ShadowOffsetY", broadcast)

	colorStuff1.colorMode:SetCallback("OnValueChanged", function(self, event, value)
		db.color.mode = value
		Healcsi:BroadcastEvent("UnitButton"..id.."ColorMode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateFontEditor(data)
	end)

	if (colorStuff1.shadeMode) then
		colorStuff1.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.shadeMode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."ShadeColor")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateFontEditor(data)
		end)
	end

	colorStuff2.colorMode:SetCallback("OnValueChanged", function(self, event, value)
		db.font.shadow.color.mode = value
		Healcsi:BroadcastEvent("UnitButton"..id.."ShadowColorMode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateFontEditor(data)
	end)

	if (colorStuff2.shadeMode) then
		colorStuff2.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
			db.font.shadow.color.shadeMode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."ShadowShadeColor")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateFontEditor(data)
		end)
	end


end

function Editor:UpdateHealthTextEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateHealthTextEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateHealthTextEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

--[[
	local line0 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	local show = Editor:CreateCheckBox(line0, false, 500, L["ShowHealtText"], db, "show", "UnitButton"..id.."Show", broadcast)

	show:SetCallback("OnValueChanged", function(self, event, value)
		db.show = value
		Healcsi:BroadcastEvent("UnitButton"..id.."Show")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateHealthTextEditor(data)
	end)

	if (not(db.show)) then return end
]]

	local hpModes = {
		["remaining"] = "Remaining",
		["missing"]   = "Missing",
		["percent"]   = "Percent"
	}
	local line1 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	local mode = Editor:CreateDropDown(line1, false, 300, L["HealtTextMode"], hpModes, db, "mode", "UnitButton"..id.."Mode", broadcast)

	mode:SetCallback("OnValueChanged", function(self, event, value)
		db.mode = value

		if (value == "percent" and (db.value == "healthAndShield" or db.value == "healthPlusShield")) then
			db.value = "includeShield"
		elseif (db.value == "includeShield") then
			db.value = "healthAndShield"
		end

		Healcsi:BroadcastEvent("UnitButton"..id.."Mode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateHealthTextEditor(data)
	end)

	local valueModes = {}
	if (db.mode == "percent") then
		valueModes = {
			["healthOnly"]    = "Health only",
			["includeShield"] = "Include shield"
		}
	else
		valueModes = {
			["healthOnly"]       = "Health only",
			["healthAndShield"]  = "Health + Shield as one number",
			["healthPlusShield"] = "Health then Shield separately"
		}
	end
	if (db.mode ~= "missing") then
		local line2 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
		local value = Editor:CreateDropDown(line2, false, 300, L["HealtTextValue"], valueModes, db, "value", "UnitButton"..id.."Value", broadcast)

		value:SetCallback("OnValueChanged", function(self, event, value)
			db.value = value

			if ((value == "healthOnly" or value == "healthAndShield") and (db.display ~= "sH" or db.display ~= "fH")) then
				db.display = "sH"
			elseif (db.display == "sH" or db.display == "fH") then
				db.display = "sHsS"
			end

			Healcsi:BroadcastEvent("UnitButton"..id.."Value")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateHealthTextEditor(data)
		end)

	end

	if (db.mode ~= "percent") then
		local displayModes = {}
		if (db.value == "healthOnly" or db.value == "healthAndShield") then
			displayModes = {
				["sH"] = "Short hp",
				["fH"]   = "Full hp"
			}
		else
			displayModes = {
				["sHsS"] = "Short hp, short shield",
				["sHfS"]   = "Short hp, full shield",
				["fHsS"] = "Full hp, short shield",
				["fHfS"]   = "Full hp, full shield"
			}
		end
		local line3 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 10})
		Editor:CreateDropDown(line3, false, 300, L["HealtTextDisplay"], displayModes, db, "display", "UnitButton"..id.."Display", broadcast)
	end

	Editor:CreateHorizLine(data.holder)

	Editor:CreateFontEditor(data.holder, db, "HealthText", broadcast)
end


function Editor:UpdateThreatEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateThreatEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateThreatEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	local line1 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:CreateCustomSelector(line1, false, 300, L["ThreatTex"], Media.aggroTextures, db, "file", "UnitButton"..id.."File", broadcast)
	
	local use = Editor:CreateCheckBox(line1, false, 500, L["UseDefaultAggroColor"], db.color.default, "use", "UnitButton"..id.."Use", broadcast)

	use:SetCallback("OnValueChanged", function(self, event, value)
		db.color.default.use = value
		Healcsi:BroadcastEvent("UnitButton"..id.."Use")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateThreatEditor(data)
	end)

	local line2 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	if (db.color.default.use) then
		Editor:CreateSlider(line2, true, 300, L["AggroOvernukingAlpha"], 0, 1, 0.01, true, db.color.default.alpha, "overnuking", "UnitButton"..id.."AggroOvernukingAlpha", broadcast)
		Editor:CreateSlider(line2, true, 300, L["AggroLosingAlpha"], 0, 1, 0.01, true, db.color.default.alpha, "losing", "UnitButton"..id.."AggroLosingAlpha", broadcast)
		Editor:CreateSlider(line2, true, 300, L["AggroTankingAlpha"], 0, 1, 0.01, true, db.color.default.alpha, "tanking", "UnitButton"..id.."AggroTankingAlpha", broadcast)
	else
		Editor:CreateColorPicker(line2, false, 300, true, L["AggroOvernukingColor"], db.color, "overnuking", "UnitButton"..id.."AggroOvernukingColor", broadcast)
		Editor:CreateColorPicker(line2, false, 300, true, L["AggroLosingColor"], db.color, "losing", "UnitButton"..id.."AggroLosingColor", broadcast)
		Editor:CreateColorPicker(line2, false, 300, true, L["AggroTankingColor"], db.color, "tanking", "UnitButton"..id.."AggroTankingColor", broadcast)
	end

end


function Editor:UpdateSelectionEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateSelectionEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateSelectionEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	local line1 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:CreateCustomSelector(line1, false, 300, L["SelectionTex"], Media.selectionTextures, db, "file", "UnitButton"..id.."File", broadcast)
	Editor:CreateColorPicker(line1, false, 300, true, L["SelectionColor"], db, "color", "UnitButton"..id.."Color", broadcast)

	Editor:CreateHeaderlabel(data.holder, L["HighLight"])
	local line2 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:CreateColorPicker(line2, false, 300, true, L["HighLightColor"], db, "highlightColor", "UnitButton"..id.."HighLightColor", broadcast)
end

function Editor:BesicIconEditor(frame, db, id, broadcast)
	local line1 = Editor:CreateGroup(frame, true, nil, {5, 0, 5, 10})

	local anchorList = {
		["CENTER"]   = "Center",
		["LEFT"]     = "Left",
		["TOP"]      = "Top",
		["RIGHT"]    = "Right",
		["BOTTOM"]   = "Bottom",

		["TOPLEFT"]  = "Top left",
		["TOPRIGHT"]  = "Top right",
		["BOTTOMLEFT"]  = "Bottom left",
		["BOTTOMRIGHT"]  = "Bottom right"
	}
	Editor:CreateDropDown(line1, false, 300, L["TextAnchor"], anchorList, db, "anchor", "UnitButton"..id.."Anchor", broadcast)

	local line2 = Editor:CreateInlineGroup(line1, true, nil, {10, 0, 10, 10})
	Editor:CreateSlider(line2, false, 300, L["OffsetX"], -500, 500, 1, false, db.offset, "x", "UnitButton"..id.."OffsetX", broadcast)
	Editor:CreateSlider(line2, false, 300, L["OffsetY"], -500, 500, 1, false, db.offset, "y", "UnitButton"..id.."OffsetY", broadcast)

	Editor:CreateSlider(line2, true, 300, L["Size"], 0, 200, 1, false, db, "size", "UnitButton"..id.."Size", broadcast)
	Editor:CreateSlider(line2, true, 300, L["Alpha"], 0, 1, 0.01, true, db, "alpha", "UnitButton"..id.."Alpha", broadcast)

	if (type(db.zoom) == "number") then
		Editor:CreateSlider(line2, true, 300, L["Zoom"], 0, 1, 0.01, true, db, "zoom", "UnitButton"..id.."Zoom", broadcast)
	end

end

function Editor:UpdateIconsEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateIconsEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateIconsEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	Editor:CreateHeaderlabel(data.holder, L["PrivilegeIcon"])
	local line1 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 20})

	Editor:CreateCheckBox(line1, false, 500, L["ShowPrivilegeLeaderIcon"],  db.privilege.leader,  "show", "UnitButtonShowPrivilegeLeaderIcon",  broadcast)
	Editor:BesicIconEditor(line1, db.privilege.leader, "PrivilegeLeader", broadcast)

	Editor:CreateHorizLine(line1)

	Editor:CreateCheckBox(line1, false, 500, L["ShowPrivilegeLootIcon"],  db.privilege.loot,  "show", "UnitButtonShowPrivilegeLootIcon",  broadcast)
	Editor:BesicIconEditor(line1, db.privilege.loot, "PrivilegeLoot", broadcast)

	Editor:CreateHorizLine(line1)

	Editor:CreateCheckBox(line1, false, 500, L["ShowPrivilegeMainTankIcon"],  db.privilege.mainTank,  "show", "UnitButtonShowPrivilegeMainTankIcon",  broadcast)
	Editor:BesicIconEditor(line1, db.privilege.mainTank, "PrivilegeMainTank", broadcast)

	Editor:CreateHorizLine(line1)

	Editor:CreateCheckBox(line1, false, 500, L["ShowPrivilegeMainAssistIcon"],  db.privilege.mainAssist,  "show", "UnitButtonShowPrivilegeMainAssistIcon",  broadcast)
	Editor:BesicIconEditor(line1, db.privilege.mainAssist, "PrivilegeMainAssist", broadcast)

	Editor:CreateCheckBox(line1, false, 500, L["ShowPrivilegeAssistIcon"],  db.privilege.assist,  "show", "UnitButtonShowPrivilegeAssistIcon",  broadcast)
	Editor:BesicIconEditor(line1, db.privilege.assist, "PrivilegeAssist", broadcast)

	Editor:CreateHorizLine(data.holder)

	Editor:CreateHeaderlabel(data.holder, L["RoleIcon"])
	local line2 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:CreateCheckBox(line2, false, 500, L["ShowRoleIcon"],       db.role,       "show", "UnitButtonShowRoleIcon",       broadcast)
	Editor:BesicIconEditor(line2, db.role, "Role", broadcast)
	Editor:CreateHorizLine(data.holder)

	Editor:CreateHeaderlabel(data.holder, L["RaidTargetIcon"])
	local line3 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:CreateCheckBox(line3, false, 500, L["ShowRaidTargetIcon"], db.raidTarget, "show", "UnitButtonShowRaidTargetIcon", broadcast)
	Editor:BesicIconEditor(line3, db.raidTarget, "RaidTarget", broadcast)
	Editor:CreateHorizLine(data.holder)

	Editor:CreateHeaderlabel(data.holder, L["BgFlagIcon"])
	local line4 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:BesicIconEditor(line4, db.flag, "Flag", broadcast)
	Editor:CreateHorizLine(data.holder)

	Editor:CreateHeaderlabel(data.holder, L["RessIcon"])
	local line5 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:BesicIconEditor(line5, db.resurrect, "Resurrect", broadcast)

	Editor:CreateHeaderlabel(data.holder, L["ReadyCheckIcon"])
	local line5 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:BesicIconEditor(line5, db.readyCheck, "ReadyCheck", broadcast)

end


function Editor:UpdateHealAbsorbEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateHealAbsorbEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateHealAbsorbEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	local line1 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:CreateCustomSelector(line1, false, 350, L["HealOnAbsorbTex"], Media.absorbTextures, db, "file", "UnitButton"..id.."File", broadcast)

	local colorStuff= Editor:CreateShadeEditor(line1, db, "AbsorbColor", nil, broadcast)

	colorStuff.colorMode:SetCallback("OnValueChanged", function(self, event, value)
		db.color.mode = value
		Healcsi:BroadcastEvent("UnitButton"..id.."ColorMode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateHealAbsorbEditor(data)
	end)

	if (colorStuff.shadeMode) then
		colorStuff.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.shadeMode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."ShadeColor")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateHealAbsorbEditor(data)
		end)
	end

end





function Editor:UpdateIncHealEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateIncHealEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateIncHealEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	local line1 = Editor:CreateGroup(data.holder, true, nil, {5, 0, 5, 10})
	Editor:CreateStatusbar(line1, false, 300, L["IncHealTex"], db, "file", "UnitButton"..id.."File", broadcast)

	local colorStuff= Editor:CreateShadeEditor(line1, db, "IncomingHealColor", nil, broadcast)

	colorStuff.colorMode:SetCallback("OnValueChanged", function(self, event, value)
		db.color.mode = value
		Healcsi:BroadcastEvent("UnitButton"..id.."ColorMode")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateIncHealEditor(data)
	end)

	if (colorStuff.shadeMode) then
		colorStuff.shadeMode:SetCallback("OnValueChanged", function(self, event, value)
			db.color.shadeMode = value
			Healcsi:BroadcastEvent("UnitButton"..id.."ShadeColor")
			if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
			Editor:UpdateIncHealEditor(data)
		end)
	end

end





function Editor:UpdateClickCastEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateClickCastEditor(data.frame, data.db, data.id, data.broadcast, data.modifiers, data.display, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end


function Editor:CreateClickCastDropDown(frame, db, id, labelText, data, broadcast)

	Editor:CreateLabel(frame, false, 150, labelText, "Black", 1, 16)

	local dropDown = Gui:Create("Healcsi-ClickCastDropdown")
	if(frame) then frame:AddChild(dropDown) end
	dropDown:SetWidth(350)

	local list = {
		["1-1-none"]       = "None",
		["2-1-target"]     = "Target",
		["3-1-togglemenu"] = "Menu",
		["4-1-assist"]     = "Assist",
		["5-1-focus"]      = "Focus", 
		["6-1-follow"]     = "Follow",
		["7-1-transfer"]   = "Transfer extra action button",
		["1-3-spell"]      = "Spell =",
		["2-3-macrotext"]  = "Macro Text =",
		["2-3-macro"]      = "Macro ="
	}

	local playerClass = select(2, UnitClass("player"))
	local playerRace  = string.lower(select(2, UnitRace("player")))
	local spells      = db.spells[playerClass]

	for i = 1, #(spells) do

		local name, rank, icon, castingTime, minRange, maxRange, spellID = GetSpellInfo(spells[i])
		-- let's see if the character can cast it or not !?
		checkedName, _, _, _, _, _, _ = GetSpellInfo(name)
		if (checkedName) then
			list[i.."-2-spell"..i] = checkedName .. "_learnt"
		else
			list[i.."-2-spell"..i] = spells[i] .. "_unlearnt"
		end
	end

	if (playerRace == "draenei") then
		local i = ((#spells) + 1)
		local name, _, _, _, _, _, _ = GetSpellInfo(28880)
		if (name) then
			list[i.."-2-spell"..i] = name .. "_learnt"
		end
	end

	dropDown:SetList(list)

	if (db.clickSet[playerClass][id]) then

		dropDown:SetValue(db.clickSet[playerClass][id])

	else
		if (id == "1") then
			db.clickSet[playerClass][id] = "2-1-target"
			dropDown:SetValue("2-1-target")
		elseif(id == "2") then
			db.clickSet[playerClass][id] = "3-1-togglemenu"
			dropDown:SetValue("3-1-togglemenu")
		else
			db.clickSet[playerClass][id] = "1-1-none"
			dropDown:SetValue("1-1-none")
		end
	end

	dropDown:SetCallback("OnValueChanged", function(self, event, value)
		db.clickSet[playerClass][id] = value
		Healcsi:BroadcastEvent("ClickCast"..id.."Changed")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateClickCastEditor(data)
	end)

	Healcsi:RegisterEventCallback("CombatStateChanged", function()
		if (Editor.InCombat) then
			dropDown:SetDisabled(true)
		else
			dropDown:SetDisabled(false)
		end
	end)

	if (Editor.InCombat) then
		dropDown:SetDisabled(true)
	else
		dropDown:SetDisabled(false)
	end

	local value = db.clickSet[playerClass][id]

	local order, section, name = string.match(value, "(%d+)-(%d+)-(.+)")

	if (tonumber(section) == 3) then

		local line = Editor:CreateInlineGroup(frame, true, nil, {0, 10, 0, 10})

		if (name == "spell") then
			local label = Editor:CreateLabel(line, false, 45, "")

			local spell = Editor:CreateEditBox(line, "Enter 'Spell Id' or 'Spell Name':")
			spell:SetWidth(240)
			local spellText = db.clickSet[playerClass][id.."_spellText"]

			spellText = spellText or ""

			local name, rank, icon, castingTime, minRange, maxRange, spellID = GetSpellInfo(spellText)

			if (not(spellText) or (spellText == "")) then
				spell:SetText("")
			elseif (name) then
				spell:SetText(name)
				db.clickSet[playerClass][id.."_spellText"] = spellID
			else
				spell:SetText(spellText .. (" (Unlearnt)"))
			end

			if (icon) then
				label:SetImage(icon)
				label.image:SetTexCoord(0.08, 0.92, 0.08, 0.92)
				label.image:SetSize(32, 32)
			else
				label:SetImage("Interface\\Icons\\INV_Misc_QuestionMark")
				label.image:SetTexCoord(0.08, 0.92, 0.08, 0.92)
				label.image:SetSize(28, 28)
			end

			spell:SetCallback("OnEnterPressed", function(self, event, value)
				local name, rank, icon, castingTime, minRange, maxRange, spellID = GetSpellInfo(value)

				if (not(value) or (value == "")) then
					-- Clear
					db.clickSet[playerClass][id.."_spellText"] = ""
				elseif (name) then
					db.clickSet[playerClass][id.."_spellText"] = spellID
				else
					db.clickSet[playerClass][id.."_spellText"] = value

				end
				Healcsi:BroadcastEvent("ClickCast"..id.."_spellTextChanged")
				if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
				Editor:UpdateClickCastEditor(data)
			end)

			Healcsi:RegisterEventCallback("CombatStateChanged", function()
				if (Editor.InCombat) then
					spell:SetDisabled(true)
				else
					spell:SetDisabled(false)
				end
			end)

			if (Editor.InCombat) then
				spell:SetDisabled(true)
			else
				spell:SetDisabled(false)
			end


		elseif (name == "macrotext") then
			local macro = Editor:CreateEditBox(line, "Enter Macro text (1 line only):")

			local macroText = db.clickSet[playerClass][id.."_macroText"]
			macroText = macroText or ""
			macro:SetText(macroText)
			macro:SetWidth(485)

			macro:SetCallback("OnEnterPressed", function(self, event, value)
				db.clickSet[playerClass][id.."_macroText"] = value
				Healcsi:BroadcastEvent("ClickCast"..id.."_macroTextChanged")
				if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
				Editor:UpdateClickCastEditor(data)
			end)

			Healcsi:RegisterEventCallback("CombatStateChanged", function()
				if (Editor.InCombat) then
					macro:SetDisabled(true)
				else
					macro:SetDisabled(false)
				end
			end)

			if (Editor.InCombat) then
				macro:SetDisabled(true)
			else
				macro:SetDisabled(false)
			end

		elseif (name == "macro") then
			local macroLabel = Editor:CreateLabel(line, false, 45, "")
			local macro = Editor:CreateEditBox(line, "Enter Macro name:")
			macro:SetWidth(240)

			local macroName = db.clickSet[playerClass][id.."_macroName"]
			macroName = macroName or ""

			macro:SetText(macroName)
			local name, texture, body = GetMacroInfo(macroName)

			macroLabel.name = macroName
			
			if (texture) then
				macroLabel:SetImage(texture)
				macroLabel.image:SetTexCoord(0.08, 0.92, 0.08, 0.92)
				macroLabel.image:SetSize(32, 32)
			else
				macroLabel:SetImage("Interface\\Icons\\INV_Misc_QuestionMark")
				macroLabel.image:SetTexCoord(0.08, 0.92, 0.08, 0.92)
				macroLabel.image:SetSize(28, 28)
			end

			macro:SetCallback("OnEnterPressed", function(self, event, value)

				local name, texture, body = GetMacroInfo(value)

				if (not(value) or (value == "")) then
					-- Clear
					db.clickSet[playerClass][id.."_macroName"] = ""
				elseif (texture and name) then
					db.clickSet[playerClass][id.."_macroName"] = name
				else
					db.clickSet[playerClass][id.."_macroName"] = value
				end

				db.clickSet[playerClass][id.."_macroName"] = value
				Healcsi:BroadcastEvent("ClickCast"..id.."_macroNameChanged")
				if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
				Editor:UpdateClickCastEditor(data)
			end)

			Healcsi:RegisterEventCallback("CombatStateChanged", function()
				if (Editor.InCombat) then
					macro:SetDisabled(true)
				else
					macro:SetDisabled(false)
				end
			end)

			if (Editor.InCombat) then
				macro:SetDisabled(true)
			else
				macro:SetDisabled(false)
			end

		end

	end

end

function Editor:CreateClickCastEditor(frame, db, id, broadcast, modifiers, display, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast
	data.modifiers = modifiers
	data.display   = display

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	for i = 1, #modifiers do
		local modifier = modifiers[i]
		local text = display[i]

		local line = Editor:CreateInlineGroup(data.holder, true, nil, {5, 0, 5, 0})

		if ((tonumber(id) == 6) or (tonumber(id) == 7)) then
			text = text .. "Scroll"
		else
			text = text .. "Click"
		end

		Editor:CreateClickCastDropDown(line, db, modifier..id, text, data, broadcast)

	end

end





function Editor:UpdateAurasEditor(data)
	data.holder:ReleaseChildren()
	data = Editor:CreateAurasEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateAurasEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder = hasPrevData.holder
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	local line1 = Editor:CreateGroup(data.holder, true, nil, {5, 5, 5, 5})
	local show = Editor:CreateCheckBox(line1, false, 300, L["Show"..id], db, "show", "UnitAuras"..id.."Show", broadcast)

	show:SetCallback("OnValueChanged", function(self, event, value)
		db.show = value
		Healcsi:BroadcastEvent("UnitAuras"..id.."Show")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateAurasEditor(data)
	end)

	if (not(db.show)) then
		return
	end

	if (id == "Debuffs") then
		Editor:CreateCheckBox(line1, false, 300, L["ShowOnlyDispellable"], db, "onlyDispellable", "UnitAuras"..id.."ShowOnlyDispellable", broadcast)
	end

	Editor:CreateSlider(line1, true, 300, L["BuffCount"], 1, 20, 1, false, db, "count", "UnitAuras"..id.."Count", broadcast)
	Editor:CreateCheckBox(line1, false, 300, L["Demo"], db, "demo", "UnitAuras"..id.."Demo", broadcast)


	local line2 = Editor:CreateInlineGroup(line1, true, nil, {10, 0, 10, 0})
	local anchorList = {
		--["LEFT"]     = "Left",
		--["TOP"]      = "Top",
		--["RIGHT"]    = "Right",
		--["BOTTOM"]   = "Bottom",

		["TOPLEFT"]  = "Top left",
		["TOPRIGHT"]  = "Top right",
		["BOTTOMLEFT"]  = "Bottom left",
		["BOTTOMRIGHT"]  = "Bottom right"
	}
	local anchor = Editor:CreateDropDown(line2, false, 300, L["Anchor"], anchorList, db, "anchor", "UnitAuras"..id.."Anchor", broadcast)

	local growList = {
		["LEFT"]   = "Left",
		["RIGHT"]  = "Right",
		["UP"]     = "Up",
		["DOWN"]   = "Down",
	}
	local grow = Editor:CreateDropDown(line2, false, 300, L["Grow"], growList, db, "grow", "UnitButton"..id.."Grow", broadcast)

	if (db.grow == "LEFT" and (db.anchor == "LEFT" or db.anchor == "TOPLEFT" or db.anchor == "BOTTOMLEFT")) then
		db.grow = "RIGHT"
		Healcsi:BroadcastEvent("UnitAuras"..id.."Grow")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateAurasEditor(data)
		return
	end

	if (db.grow == "RIGHT" and (db.anchor == "RIGHT" or db.anchor == "TOPRIGHT" or db.anchor == "BOTTOMRIGHT")) then
		db.grow = "LEFT"
		Healcsi:BroadcastEvent("UnitAuras"..id.."Grow")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateAurasEditor(data)
		return
	end

	if (db.grow == "UP" and (db.anchor == "TOP" or db.anchor == "TOPRIGHT" or db.anchor == "TOPLEFT")) then
		db.grow = "DOWN"
		Healcsi:BroadcastEvent("UnitAuras"..id.."Grow")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateAurasEditor(data)
		return
	end

	if (db.grow == "DOWN" and (db.anchor == "BOTTOM" or db.anchor == "BOTTOMLEFT" or db.anchor == "BOTTOMRIGHT")) then
		db.grow = "UP"
		Healcsi:BroadcastEvent("UnitAuras"..id.."Grow")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateAurasEditor(data)
		return
	end

	anchor:SetCallback("OnValueChanged", function(self, event, value)
		db.anchor = value
		Healcsi:BroadcastEvent("UnitAuras"..id.."Anchor")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateAurasEditor(data)
	end)

	grow:SetCallback("OnValueChanged", function(self, event, value)
		db.grow = value
		Healcsi:BroadcastEvent("UnitAuras"..id.."Grow")
		if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
		Editor:UpdateAurasEditor(data)
	end)

	local line3 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
	Editor:CreateSlider(line3, false, 300, L["OffsetX"], -500, 500, 1, false, db.offset, "x", "UnitAuras"..id.."OffsetX", broadcast)
	Editor:CreateSlider(line3, false, 300, L["OffsetY"], -500, 500, 1, false, db.offset, "y", "UnitAuras"..id.."OffsetY", broadcast)

	Editor:CreateSlider(line3, true, 300, L["Size"], 0, 48, 1, false, db, "size", "UnitAuras"..id.."Size", broadcast)
	Editor:CreateSlider(line3, true, 300, L["Alpha"], 0, 1, 0.01, true, db, "alpha", "UnitAuras"..id.."Alpha", broadcast)
	Editor:CreateSlider(line3, true, 300, L["Zoom"], 0, 1, 0.01, true, db, "zoom", "UnitAuras"..id.."Zoom", broadcast)

	local lineA = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
	Editor:CreateLabel(lineA, true, 300, L["AvailableSizeHint"])
	Editor:CreateSlider(lineA, false, 300, L["AvailableWidth"], 0.2, 1.2, 0.01, true, db, "availableWidth", "UnitButton"..id.."AvailableWidth", broadcast)
	Editor:CreateSlider(lineA, false, 300, L["AvailableHeight"], 0.2, 1.2, 0.01, true, db, "availableHeight", "UnitButton"..id.."AvailableHeight", broadcast)

	if (id ~= "Dispels") then

		Editor:CreateHorizLine(data.holder)

		Editor:CreateHeaderlabel(data.holder, L["RemainigTime"])
		local line4 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})


		local anchorListWithCenter = {
			["CENTER"]   = "Center",
			["LEFT"]     = "Left",
			["TOP"]      = "Top",
			["RIGHT"]    = "Right",
			["BOTTOM"]   = "Bottom",

			["TOPLEFT"]  = "Top left",
			["TOPRIGHT"]  = "Top right",
			["BOTTOMLEFT"]  = "Bottom left",
			["BOTTOMRIGHT"]  = "Bottom right"
		}
		Editor:CreateDropDown(line4, false, 300, L["Anchor"], anchorListWithCenter, db.remaining, "anchor", "UnitButton"..id.."RemainingAnchor", broadcast)

		local line5 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
		Editor:CreateSlider(line5, false, 300, L["OffsetX"], -500, 500, 1, false, db.remaining.offset, "x", "UnitButton"..id.."RemainingOffsetX", broadcast)
		Editor:CreateSlider(line5, false, 300, L["OffsetY"], -500, 500, 1, false, db.remaining.offset, "y", "UnitButton"..id.."RemainingOffsetY", broadcast)

		local line6 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
		Editor:CreateSlider(line6, false, 300, L["FontSize"], 4, 90, 1, false, db.remaining, "size", "UnitButton"..id.."RemainingFontSize", broadcast)
		Editor:CreateSlider(line6, true, 300, L["Alpha"], 0, 1, 0.01, true, db.remaining, "alpha", "UnitButton"..id.."Alpha", broadcast)
		Editor:CreateFont(line6, false, 300, L["FontType"], db.remaining, "font", "UnitButton"..id.."RemainingFontType", broadcast)
		Editor:CreateCheckBox(line6, false, 500, L["FontOutline"], db.remaining, "hasOutline", "UnitButton"..id.."RemainingFontOutline", broadcast)

		Editor:CreateHorizLine(data.holder)

		Editor:CreateHeaderlabel(data.holder, L["StackCount"])
		local line7 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})

		Editor:CreateDropDown(line7, false, 300, L["Anchor"], anchorListWithCenter, db.stack, "anchor", "UnitButton"..id.."RemainingAnchor", broadcast)

		local line8 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
		Editor:CreateSlider(line8, false, 300, L["OffsetX"], -500, 500, 1, false, db.stack.offset, "x", "UnitButton"..id.."RemainingOffsetX", broadcast)
		Editor:CreateSlider(line8, false, 300, L["OffsetY"], -500, 500, 1, false, db.stack.offset, "y", "UnitButton"..id.."RemainingOffsetY", broadcast)

		local line9 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
		Editor:CreateSlider(line9, false, 300, L["FontSize"], 4, 90, 1, false, db.stack, "size", "UnitButton"..id.."RemainingFontSize", broadcast)
		Editor:CreateFont(line9, false, 300, L["FontType"], db.stack, "font", "UnitButton"..id.."RemainingFontType", broadcast)
		Editor:CreateCheckBox(line9, false, 500, L["FontOutline"], db.stack, "hasOutline", "UnitButton"..id.."RemainingFontOutline", broadcast)

	end

end

function Editor:UpdateDebuffHighlightEditor(data)
	data.holder:ReleaseChildren()
	local db = data.db
	data = Editor:CreateDebuffHighlightEditor(data.frame, data.db, data.id, data.broadcast, data.UnitAuras, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateDebuffHighlightCustomEditor(frame, db, id, broadcast, UnitAuras, selectedTier, selectedInstanceId, selectedBossId, data)

	if (not(db.lastTierSelected) or not(db.lastInstanceSelected) or not(db.lastBossSelected)) then return end

	local line_header1  = Editor:CreateInlineGroup(frame, true, nil, {10, 0, 0, 10})
	local left_header  = Editor:CreateLabel(line_header1, true, 100, L["DebuffList"], "Black", 1, 16)
	left_header:SetJustifyH("LEFT")

	local line_header2  = Editor:CreateInlineGroup(frame, true, nil, {-18, 0, 0, 10})
	local right_header = Editor:CreateLabel(line_header2, true, 100, L["Prio"], "Black", 1, 16)
	right_header:SetJustifyH("RIGHT")

	local holder = Editor:CreateGroup(frame, true, nil, {10, 10, 5, 20}, {0, 5, 0, 5}, "BlueSelected", 0.5)

	if (data.addCustom) then
		data.addCustom = nil
		local line_add = Editor:CreateInlineGroup(frame, true, nil, {5, 0, 5, 0})
		local addCustom = Editor:CreateEditBox(line_add, "Enter 'Spell Id' or 'Debuff Name':")
		addCustom:SetWidth(360)

		addCustom:SetCallback("OnEnterPressed", function(self, event, value)

			value = strtrim(value or "")
			local name, _, icon, link
			local id = tonumber(value)
			local inserted = false
			if (id) then
				name, _, icon = GetSpellInfo(id)
				if (name) then
					UnitAuras:RegisterDebuff(selectedTier, selectedInstanceId, selectedBossId, id, 2, true)
					inserted = true
				end
			else
				for id = 1, MAX_SPELL_ID do
					name, _, icon, _, _, _, spellID = GetSpellInfo(id)
					if (name == value) then
						UnitAuras:RegisterDebuff(selectedTier, selectedInstanceId, selectedBossId, spellID, 2, true)
						inserted = true
					end
				end
			end

			if (inserted) then
				Editor:UpdateDebuffHighlightEditor(data)
			else
				self:SetText("Debuff or spell not found for: " .. value)
			end

		end)

	end

	local line_btn = Editor:CreateInlineGroup(frame, true, nil, {5, 0, 5, 0})

	local line_addBtn = Editor:CreateInlineGroup(line_btn, false, 260, {5, 5, 5, 5})
	local addBtn = Editor:CreateButton(line_addBtn, "Add Custom Dubuff")

	addBtn:SetCallback("OnClick", function(self)
		data.addCustom = 1
		Editor:UpdateDebuffHighlightEditor(data)
	end)

	local line_removeBtn = Editor:CreateInlineGroup(line_btn, false, 260, {5, 5, 5, 5})
	local removeBtn = Editor:CreateButton(line_removeBtn, "Remove Selected")
	removeBtn:SetDisabled(true)

	removeBtn:SetCallback("OnClick", function(self)
		if (self.data and self.data.id) then
			UnitAuras:DeleteCustomDebuff(selectedTier, selectedInstanceId, self.data.id)
			self.data = nil
			Editor:UpdateDebuffHighlightEditor(data)
		end
	end)

	local line_clearBtn = Editor:CreateInlineGroup(line_btn, true, 460, {5, 5, 5, 5})
	local clearBtn

	if (data.resetAll) then
		clearBtn = Editor:CreateButton(line_clearBtn, "Ples CONFIRM Factory reset! (Click again...)")
	elseif (data.resetted) then
		clearBtn = Editor:CreateButton(line_clearBtn, "Factory reset ALL Debuff data")
		Editor:CreateLabel(line_clearBtn, true, 100, "Reset complete! All custom data is now gone :(", "Black", 1, 16)
	else
		clearBtn = Editor:CreateButton(line_clearBtn, "Factory reset ALL Debuff data")
	end

	clearBtn:SetWidth(460)

	clearBtn:SetCallback("OnClick", function(self)
		if (data.resetAll) then
			UnitAuras:DebuffHighLightReset()
			data.resetAll = nil
			data.resetted = 1
		elseif (data.resetted) then
			data.resetted = nil
			data.resetAll = nil
		else
			data.resetted = nil
			data.resetAll = 1
		end
		Editor:UpdateDebuffHighlightEditor(data)
	end)

	local rawDebuffList = UnitAuras:EnumDebuffs(selectedTier, selectedInstanceId, selectedBossId)

	if (rawDebuffList) then

		local i
		for i=1, #rawDebuffList do
			--Healcsi:Print_r(rawDebuffList[i]["data"]) -- id, level, icon, defLevel, link

			if (i > 1) then
				Editor:CreateHorizLine(holder)
			end

			local line_debuff = Editor:CreateInlineGroup(holder, true, nil, {5, 0, 5, 0})

			local item = Gui:Create("Healcsi-DebuffItem")

			item.data = rawDebuffList[i]["data"]
			item.name = rawDebuffList[i]["name"]

			item:SetFullWidth(true)

			item:SetFont(Healcsi:GetFont("Friz Quadrata TT"), 16)
			item:SetFontColor(Healcsi:GetColor("Black", 1))
			item:SetText(     rawDebuffList[i]["name"])
			item:SetImage(    rawDebuffList[i]["data"]["icon"])
			item:SetPriority( rawDebuffList[i]["data"]["level"])

			line_debuff:AddChild(item)

			item.frame:SetScript("OnEnter", function(frame)
				local self = frame.obj

				local right = frame:GetLeft() < 300
				local bottom = GetScreenHeight() - frame:GetTop() < 360

				local pointX = right and "LEFT" or "RIGHT"
				local pointY = bottom and "TOP" or "BOTTOM"
				local relPointX = right and "RIGHT" or "LEFT"
				local relPointY = bottom and "BOTTOM" or "TOP"

				GameTooltip:SetOwner(frame, "ANCHOR_NONE")
				GameTooltip:ClearAllPoints()
				GameTooltip:SetPoint(pointY..pointX, frame, relPointY..relPointX, right and 4 or -4, bottom and -4 or 4)

				GameTooltip:SetText(self.name)
				GameTooltip:AddLine(GetSpellDescription(self.data.id), 1, 1, 1)
				GameTooltip:AddLine("Spell Id: " .. self.data.id, 1, 1, 1)
				GameTooltip:Show()
			end)

			item.frame:SetScript("OnLeave", function(frame)
				GameTooltip:Hide()
			end)

			item.dropDown:SetCallback("OnValueChanged", function(self, event, value)
				
				UnitAuras:SetDebuffLevel(selectedTier, selectedInstanceId, self.obj.data.id, value)

				if (broadcast) then Healcsi:BroadcastEvent(broadcast) end
				Editor:UpdateDebuffHighlightEditor(data)
			end)

			item:SetCallback("OnSelected", function(self)
				if (self.data and self.data.custom) then
					removeBtn:SetDisabled(false)
					removeBtn.data = self.data
				else
					removeBtn:SetDisabled(true)
				end
			end)

		end

	end

end

function Editor:CreateDebuffHighlightEditor(frame, db, id, broadcast, UnitAuras, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast
	data.UnitAuras = UnitAuras

	if (hasPrevData) then
		data.holder    = hasPrevData.holder
		data.addCustom = hasPrevData.addCustom
		data.resetAll  = hasPrevData.resetAll
		data.resetted  = hasPrevData.resetted
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 10}, {5, 0, 5, 0})
		data.holder = holder
	end

	local line1 = Editor:CreateGroup(data.holder, true, nil, {5, 5, 5, 5})
	Editor:CreateCheckBox(line1, false, 300, L["Enable"], db, "show", "UnitAuras"..id.."Show", broadcast)

	if (not(db.show)) then return end

	local line2 = Editor:CreateGroup(data.holder, true, nil, {5, 5, 5, 5})
	local anchorList = {
		["CENTER"]   = "Center",
		["LEFT"]     = "Left",
		["TOP"]      = "Top",
		["RIGHT"]    = "Right",
		["BOTTOM"]   = "Bottom",

		["TOPLEFT"]  = "Top left",
		["TOPRIGHT"]  = "Top right",
		["BOTTOMLEFT"]  = "Bottom left",
		["BOTTOMRIGHT"]  = "Bottom right"
	}
	Editor:CreateDropDown(line2, false, 300, L["Anchor"], anchorList, db, "anchor", "UnitButton"..id.."Anchor", broadcast)

	local line3 = Editor:CreateInlineGroup(data.holder, true, nil, {10, 0, 10, 10})
	Editor:CreateSlider(line3, false, 300, L["OffsetX"], -500, 500, 1, false, db.offset, "x", "UnitButton"..id.."OffsetX", broadcast)
	Editor:CreateSlider(line3, false, 300, L["OffsetY"], -500, 500, 1, false, db.offset, "y", "UnitButton"..id.."OffsetY", broadcast)
	Editor:CreateSlider(line3, true, 300, L["Scale"], 0, 3, 0.01, true, db, "scale", "UnitButton"..id.."Scale", broadcast)
	Editor:CreateSlider(line3, true, 300, L["Alpha"], 0.2, 1, 0.01, true, db, "alpha", "UnitButton"..id.."Alpha", broadcast)
	Editor:CreateSlider(line3, true, 300, L["Zoom"], 0, 1, 0.01, true, db, "zoom", "UnitButton"..id.."Zoom", broadcast)

	Editor:CreateHorizLine(data.holder)

	local line4 = Editor:CreateGroup(data.holder, true, nil, {10, 0, 10, 10})

-- #############################################################################
	local line_version = Editor:CreateInlineGroup(line4, true, nil, {10, 0, 10, 10})
	Editor:CreateLabel(line_version, false, 100, L["Version"], "Black", 1, 16)

	local tierList = {}
	local i
	for i = 1, UnitAuras:GetNumTiers() do
		tinsert(tierList, UnitAuras:GetTierName(i))
	end
	
	local tierDropDown = Gui:Create("Healcsi-Dropdown")
	tierDropDown:SetWidth(300)

	tierDropDown:SetList(tierList)
	tierDropDown:SetValue(db.lastTierSelected)

	tierDropDown:SetCallback("OnValueChanged", function(self, event, value)
		db.lastTierSelected = value
		db.lastInstanceSelected = nil
		db.lastBossSelected = nil
		Editor:UpdateDebuffHighlightEditor(data)
	end)
	line_version:AddChild(tierDropDown)

-- #############################################################################
	local line_instance = Editor:CreateInlineGroup(line4, true, nil, {10, 0, 10, 10})
	Editor:CreateLabel(line_instance, false, 100, L["Instance"], "Black", 1, 16)

	local instanceList = {}
	local selectedInstanceId
	local rawRaidList = UnitAuras:EnumInstances(tierDropDown:GetValue(), "raid")
	for i = 1, #rawRaidList do
		instanceList[i.."-raid"] = rawRaidList[i]["name"]
		if (db.lastInstanceSelected == (i.."-raid")) then
			selectedInstanceId = rawRaidList[i]["id"]
		end
	end
	local rawPartyList = UnitAuras:EnumInstances(tierDropDown:GetValue(), "party")
	for i = 1, #rawPartyList do
		instanceList[i.."-party"] = rawPartyList[i]["name"]
		if (db.lastInstanceSelected == (i.."-party")) then
			selectedInstanceId = rawPartyList[i]["id"]
		end
	end
	local instanceDropDown = Gui:Create("Healcsi-InstancesDropdown")
	instanceDropDown:SetWidth(300)

	instanceDropDown:SetList(instanceList)
	instanceDropDown:SetValue(db.lastInstanceSelected)

	instanceDropDown:SetCallback("OnValueChanged", function(self, event, value)
		db.lastInstanceSelected = value
		db.lastBossSelected = nil
		Editor:UpdateDebuffHighlightEditor(data)
	end)
	line_instance:AddChild(instanceDropDown)

-- #############################################################################
	local line_boss = Editor:CreateInlineGroup(line4, true, nil, {10, 0, 0, 10})
	Editor:CreateLabel(line_boss, false, 100, L["Boss"], "Black", 1, 16)

	local bossList = {}
	local selectedBossId
	local rawBossList = UnitAuras:EnumBosses(tierDropDown:GetValue(), selectedInstanceId)
	if (rawBossList) then
		for i = 1, #rawBossList  do
			bossList[i] = rawBossList[i]["name"]
			if (db.lastBossSelected == i) then
				selectedBossId = rawBossList[i]["id"]
			end
		end
	end

	local bossDropDown = Gui:Create("Healcsi-Dropdown")
	bossDropDown:SetWidth(300)

	bossDropDown:SetList(bossList)
	bossDropDown:SetValue(db.lastBossSelected)

	bossDropDown:SetCallback("OnValueChanged", function(self, event, value)
		db.lastBossSelected = value
		Editor:UpdateDebuffHighlightEditor(data)
	end)
	line_boss:AddChild(bossDropDown)

	Editor:CreateDebuffHighlightCustomEditor(data.holder, db, id, broadcast, UnitAuras, db.lastTierSelected, selectedInstanceId, selectedBossId, data)

end

function Editor:UpdateExtraBuffsEditor(data)
	data.holder:ReleaseChildren()
	local db = data.db
	data = Editor:CreateExtraBuffsEditor(data.frame, data.db, data.id, data.broadcast, data)
	Healcsi:BroadcastEvent("UpdateEditorLayout")
end

function Editor:CreateExtraBuffsEditor(frame, db, id, broadcast, hasPrevData)
	local data     = {}
	data.frame     = frame
	data.db        = db
	data.id        = id
	data.broadcast = broadcast

	if (hasPrevData) then
		data.holder    = hasPrevData.holder
		data.addCustom = hasPrevData.addCustom
	else
		local padding = Editor:CreateGroup(frame, true, nil, {10, 0, 0, 0})
		local holder = Editor:CreateGroup(padding, true, nil, {0, 10, 10, 0}, {5, 0, 5, 0})
		data.holder = holder
	end

	

	if (data.addCustom) then
		data.addCustom = nil
		local line_add = Editor:CreateInlineGroup(data.holder, true, nil, {5, 0, 5, 0})
		local addCustom = Editor:CreateEditBox(line_add, "Enter 'Spell Id' or 'Debuff Name':")
		addCustom:SetWidth(360)

		addCustom:SetCallback("OnEnterPressed", function(self, event, value)

			value = strtrim(value or "")
			local name, _, icon, link
			local id = tonumber(value)
			local inserted = false
			if (id) then
				name, _, icon = GetSpellInfo(id)
				if (name) then
					tinsert(db.before, {class="ALL", spellId=id})
					inserted = true
				end
			else
				for id = 1, MAX_SPELL_ID do
					name, _, icon, _, _, _, spellID = GetSpellInfo(id)
					if (name == value) then
						tinsert(db.before, {class="ALL", spellId=spellID})
						inserted = true
					end
				end
			end

			if (inserted) then
				Editor:UpdateExtraBuffsEditor(data)
			else
				self:SetText("Debuff or spell not found for: " .. value)
			end

		end)

	end

	local line_addBtn = Editor:CreateInlineGroup(data.holder, false, 260, {5, 5, 10, 5})
	local addBtn = Editor:CreateButton(line_addBtn, "Add Custom Buff")

	addBtn:SetCallback("OnClick", function(self)
		data.addCustom = 1
		Editor:UpdateExtraBuffsEditor(data)
	end)


	local holder = Editor:CreateGroup(data.holder, true, nil, {10, 10, 5, 20}, {0, 5, 0, 5}, "BlueSelected", 0.5)

	local classList = {
		["DRUID"]        = "Druid",
		["HUNTER"]       = "Hunter",
		["MAGE"]         = "Mage",
		["PALADIN"]      = "Paladin",
		["PRIEST"]       = "Priest",
		["ROGUE"]        = "Rogue",
		["SHAMAN"]       = "Shaman",
		["WARLOCK"]      = "Warlock",
		["WARRIOR"]      = "Warrior",
		["DEATHKNIGHT"]  = "Death knight",
		["MONK"]         = "Monk",
		["DEMONHUNTER"]  = "Demon hunter",
		["ALL"]          = "All"
	}

	local prioList = {
		["high"] = "High prio.",
		["low"]  = "Low prio."
	}

	local count = 0
	for key, value in ipairs(db.before) do

		if (value) then

			count = count + 1
			if (key > 1) then Editor:CreateHorizLine(holder) end

			local name, rank, icon, castingTime, minRange, maxRange, spellID = GetSpellInfo(value.spellId)
		

			local label = Gui:Create("Healcsi-Label")
			label:SetFullWidth(isFullWidth)
			label:SetWidth(300)

			label:SetFont(Healcsi:GetFont("Friz Quadrata TT"), 16)
			label:SetFontColor(Healcsi:GetColor("Black", 1))
			label:SetText((name or "?") .. " (" ..(spellID or "?").. ")" )
			label:SetImage(icon)

			holder:AddChild(label)

			local line = Editor:CreateInlineGroup(holder, true, nil, {5, 5, 5, 5})

			local dropDown1 = Gui:Create("Healcsi-Dropdown")
			dropDown1:SetWidth(150)

			dropDown1:SetList(classList)
			dropDown1:SetValue(value.class)
			dropDown1.dataKey = key
			dropDown1:SetCallback("OnValueChanged", function(self, event, value)
				db.before[self.dataKey]["class"] = value
				self:SetValue(value)
				Editor:UpdateExtraBuffsEditor(data)
			end)
			line:AddChild(dropDown1)

			local dropDown2 = Gui:Create("Healcsi-Dropdown")
			dropDown2.data = value
			dropDown2.dataKey = key
			dropDown2:SetWidth(130)
			dropDown2:SetList(prioList)
			dropDown2:SetValue("high")
			dropDown2:SetCallback("OnValueChanged", function(self, event, value)
				if (value == "low") then
					tinsert(db.after, self.data)
					table.remove(db.before, self.dataKey)
					Editor:UpdateExtraBuffsEditor(data)
				end
			end)
			line:AddChild(dropDown2)

			local removeBtn = Editor:CreateButton(line, "Remove")
			removeBtn.dataKey = key
			removeBtn:SetWidth(120)

			removeBtn:SetCallback("OnClick", function(self)
				table.remove(db.before, self.dataKey)
				Editor:UpdateExtraBuffsEditor(data)
			end)

		end

	end

	for key, value in ipairs(db.after) do

		if (value) then

			count = count + 1
			if (count>1 or key > 1) then Editor:CreateHorizLine(holder) end

			local name, rank, icon, castingTime, minRange, maxRange, spellID = GetSpellInfo(value.spellId)

			local label = Gui:Create("Healcsi-Label")
			label:SetFullWidth(isFullWidth)
			label:SetWidth(300)

			label:SetFont(Healcsi:GetFont("Friz Quadrata TT"), 16)
			label:SetFontColor(Healcsi:GetColor("Black", 1))
			label:SetText((name or "?") .. " (" ..(spellID or "?").. ")" )
			label:SetImage(icon)

			holder:AddChild(label)

			local line = Editor:CreateInlineGroup(holder, true, nil, {5, 5, 5, 5})

			local dropDown1 = Gui:Create("Healcsi-Dropdown")
			dropDown1:SetWidth(150)

			dropDown1:SetList(classList)
			dropDown1:SetValue(value.class)
			dropDown1.dataKey = key
			dropDown1:SetCallback("OnValueChanged", function(self, event, value)
				db.after[self.dataKey]["class"] = value
				self:SetValue(value)
				Editor:UpdateExtraBuffsEditor(data)
			end)
			line:AddChild(dropDown1)

			local dropDown2 = Gui:Create("Healcsi-Dropdown")
			dropDown2.data = value
			dropDown2.dataKey = key
			dropDown2:SetWidth(130)
			dropDown2:SetList(prioList)
			dropDown2:SetValue("low")
			dropDown2:SetCallback("OnValueChanged", function(self, event, value)
				if (value == "high") then
					tinsert(db.before, self.data)
					table.remove(db.after, self.dataKey)
					Editor:UpdateExtraBuffsEditor(data)
				end
			end)
			line:AddChild(dropDown2)

			local removeBtn = Editor:CreateButton(line, "Remove")
			removeBtn.dataKey = key
			removeBtn:SetWidth(120)

			removeBtn:SetCallback("OnClick", function(self)
				table.remove(db.after, self.dataKey)
				Editor:UpdateExtraBuffsEditor(data)
			end)

		end

	end

end
