--[[-----------------------------------------------------------------------------
Healcsi Window
-------------------------------------------------------------------------------]]
local Healcsi      = Healcsi
local Gui          = Healcsi.Gui
local TopParent    = Healcsi.UIParent
local HiddenParent = Healcsi.HiddenParent

-- Lua APIs
local pairs, assert, type = pairs, assert, type

-- WoW APIs
local CreateFrame = CreateFrame


do
	local Type = "HealcsiWindow"
	local Version = 1

	local function CloseWindow(this)
		this.obj:Hide()
	end

	local function FrameOnShow(this)
		this.obj:Fire("OnShow")
	end

	local function FrameOnClose(this)
		this.obj:Fire("OnClose")
	end

	local function FrameOnMouseDown(this)
		Gui:ClearFocus()
	end

	local function TitleOnMouseDown(this)
		local frame = this:GetParent()
		frame:StartMoving()
		Gui:ClearFocus()
	end

	local function TitleOnMouseUp(this)
		local frame = this:GetParent()
		frame:StopMovingOrSizing()

		-- ... Furthermore, StopMovingOrSizing() will identify the closest anchor point on the
		-- screen and re-anchor the frame once again to that closest point...

		local self = frame.obj

		self.Database.offset.top = math.floor(frame:GetTop() + 0.5)
		self.Database.offset.left = math.floor(frame:GetLeft() + 0.5)

		local width, height = GetScreenWidth(), GetScreenHeight()
		local top, right, bottom, left = frame:GetTop(), frame:GetRight(), frame:GetBottom(), frame:GetLeft()
		top = height - top - 20
		right = width - right - 20
		bottom = bottom - 20
		left = left - 20

		if (bottom < 0 and top < 0) then
			self.Database.offset.top = height - 10
			self.Database.height = height - 20
			self.Database.height = math.floor(self.Database.height + 0.5)
			frame:SetHeight(self.Database.height)
		elseif (top < 0) then
			self.Database.offset.top = height - 10
		elseif (bottom < 0) then
			self.Database.offset.top = self.Database.height + 10
		end

		if (left < 0 and right < 0) then
			self.Database.offset.left = 10
			self.Database.width = width - 20
			self.Database.width = math.floor(self.Database.width + 0.5)
			frame:SetWidth(self.Database.width)
		elseif (right < 0) then
			self.Database.offset.left = width - self.Database.width - 10
		elseif (left < 0) then
			self.Database.offset.left = 10
		end

		self.Database.offset.left = math.floor(self.Database.offset.left + 0.5)
		self.Database.offset.top = math.floor(self.Database.offset.top + 0.5)

		frame:ClearAllPoints() -- Reset anchor, beacuse why not? Right?!
		frame:SetPoint("TOPLEFT", TopParent, "BOTTOMLEFT", self.Database.offset.left, self.Database.offset.top)


	end

	local function SizerOnMouseDown(this)
		local frame = this:GetParent()
		frame:StartSizing("BOTTOMRIGHT")
		frame.isSizing = 1
		frame.sizingUpdate = 0
		Gui:ClearFocus()
	end

	local function SizerOnMouseUp(this)
		local frame = this:GetParent()
		frame:StopMovingOrSizing()
		frame.isSizing = 2

		local self = frame.obj
		local newWidth, newHeight = frame:GetSize()
		self.Database.width = math.floor(newWidth + 0.5)
		self.Database.height = math.floor(newHeight + 0.5)

		frame:SetWidth(self.Database.width)
		frame:SetHeight(self.Database.height)

		local width, height = GetScreenWidth(), GetScreenHeight()
		local top, right, bottom, left = frame:GetTop(), frame:GetRight(), frame:GetBottom(), frame:GetLeft()
		top = height - top - 20
		right = width - right - 20
		bottom = bottom - 20
		left = left - 20

		if (bottom < 0 and top < 0) then
			self.Database.offset.top = height - 10
			self.Database.height = height - 20
			self.Database.height = math.floor(self.Database.height + 0.5)
			frame:SetHeight(self.Database.height)
		elseif (top < 0) then
			self.Database.offset.top = height - 10
		elseif (bottom < 0) then
			self.Database.offset.top = self.Database.height + 10
		end

		if (left < 0 and right < 0) then
			self.Database.offset.left = 10
			self.Database.width = width - 20
			self.Database.width = math.floor(self.Database.width + 0.5)
			frame:SetWidth(self.Database.width)
		elseif (right < 0) then
			self.Database.offset.left = width - self.Database.width - 10
		elseif (left < 0) then
			self.Database.offset.left = 10
		end

		self.Database.offset.left = math.floor(self.Database.offset.left + 0.5)
		self.Database.offset.top = math.floor(self.Database.offset.top + 0.5)

		frame:ClearAllPoints() -- Reset anchor, beacuse why not? Right?!
		frame:SetPoint("TOPLEFT", TopParent, "BOTTOMLEFT", self.Database.offset.left, self.Database.offset.top)

	end

	local function OnWindowUpdate(frame, elapsed)

		if (not(frame.obj.content.background.texture)) then return end

		if (frame.isSizing ~= nil) then
			frame.sizingUpdate = (frame.sizingUpdate or 0) + elapsed
			if (frame.sizingUpdate > 0.1) then
				if (frame.isSizing == 2) then
					Healcsi:UpdateMiddleBackground(frame.obj.content, frame.obj.content.background.texture, frame.obj.content.background.mediaWidth, frame.obj.content.background.mediaHeight, 4)
					frame.isSizing = nil
				else
					Healcsi:UpdateMiddleBackground(frame.obj.content, frame.obj.content.background.texture, frame.obj.content.background.mediaWidth, frame.obj.content.background.mediaHeight, 4)
				end
			end
		end
	end

	local function UpdateMenu(self)
		if (self.menu.modules.selected) and (self.menu.profiles.selected) then
			self.menu.profiles.selected = nil
		end

		if (self.menu.modules.selected) then
			self.Database.selectedMenu = "modules"
			Healcsi:DecorateTexture(self.menu.modules.texture, "background", "Solid", "BlueSelected", 0.8)
			self.menu.modules.texture:Show()
			self.menu.modules.selector.holder:Show()
			self.menu.profiles.texture:Hide()
		end

		if (self.menu.profiles.selected) then
			self.Database.selectedMenu = "profiles"
			Healcsi:DecorateTexture(self.menu.profiles.texture, "background", "Solid", "BlueSelected", 0.8)
			self.menu.profiles.texture:Show()
			self.menu.modules.texture:Hide()
			self.menu.modules.selector.holder:Hide()
		end

		self:Fire("MenuChanged", self.Database.selectedMenu)
	end

	local function ModulesOnMouseDown(this)
		local frame = this:GetParent()
		local self = frame.obj

		Healcsi:DecorateTexture(self.menu.modules.texture, "background", "Solid", "BluePressed", 0.8)
		self.menu.modules.texture:Show()
		self.menu.modules.pressed = true
	end

	local function ModulesOnMouseUp(this)
		local frame = this:GetParent()
		local self = frame.obj
		if (self.menu.modules.entered) then
			self.menu.modules.selected = true
			self.menu.profiles.selected = nil
		end
		UpdateMenu(self)
	end

	local function ModulesOnEnter(this)
		local frame = this:GetParent()
		local self = frame.obj

		Healcsi:DecorateTexture(self.menu.modules.texture, "background", "Solid", "BlueHovered", 0.8)
		self.menu.modules.texture:Show()
		self.menu.modules.entered = true
	end

	local function ModulesOnLeave(this)
		local frame = this:GetParent()
		local self = frame.obj
		if (self.menu.modules.selected == nil) then
			self.menu.modules.texture:Hide()
		else
			Healcsi:DecorateTexture(self.menu.modules.texture, "background", "Solid", "BlueSelected", 0.8)
		end
		self.menu.modules.entered = nil
	end

	local function ProfilesOnMouseDown(this)
		local frame = this:GetParent()
		local self = frame.obj

		Healcsi:DecorateTexture(self.menu.profiles.texture, "background", "Solid", "BluePressed", 0.8)
		self.menu.profiles.texture:Show()
		self.menu.profiles.pressed = true
	end

	local function ProfilesOnMouseUp(this)
		local frame = this:GetParent()
		local self = frame.obj
		if (self.menu.profiles.entered) then
			self.menu.profiles.selected = true
			self.menu.modules.selected = nil
		end
		UpdateMenu(self)
	end

	local function ProfilesOnEnter(this)
		local frame = this:GetParent()
		local self = frame.obj

		Healcsi:DecorateTexture(self.menu.profiles.texture, "background", "Solid", "BlueHovered", 0.8)
		self.menu.profiles.texture:Show()
		self.menu.profiles.entered = true
	end

	local function ProfilesOnLeave(this)
		local frame = this:GetParent()
		local self = frame.obj
		if (self.menu.profiles.selected == nil) then
			self.menu.profiles.texture:Hide()
		else
			Healcsi:DecorateTexture(self.menu.profiles.texture, "background", "Solid", "BlueSelected", 0.8)
		end
		self.menu.profiles.entered = nil
	end

	local function CloseOnEnter(this)
		local frame = this:GetParent()
		local self = frame.obj

		Healcsi:SetVertexColor(self.closeButton.texture, "BlueHovered", 1)
	end

	local function CloseOnLeave(this)
		local frame = this:GetParent()
		local self = frame.obj

		Healcsi:SetVertexColor(self.closeButton.texture, "DarkBlue", 1)
	end

	local function MemoryUpdate(this, elapsed)
		this.updateTimer = (this.updateTimer or 0) + elapsed
		if (this.updateTimer > 2) then
			this.updateTimer = 0
			this.text:SetText("Memory Usage: " .. Healcsi:GetMemoryUsageString())
		end
	end

	local function Hide(self)
		self.frame:Hide()
	end
	
	local function Show(self)
		self.frame:Show()
	end

	local function SetWindowShadowColor(self, r, g, b, a)
		local shadow = self.frame.shadow

		shadow.left:SetVertexColor(r, g, b, a or 1)
		shadow.right:SetVertexColor(r, g, b, a or 1)
		shadow.top:SetVertexColor(r, g, b, a or 1)
		shadow.bottom:SetVertexColor(r, g, b, a or 1)
		shadow.topLeft:SetVertexColor(r, g, b, a or 1)
		shadow.topRight:SetVertexColor(r, g, b, a or 1)
		shadow.bottomLeft:SetVertexColor(r, g, b, a or 1)
		shadow.bottomRight:SetVertexColor(r, g, b, a or 1)
		
	end

	local function SetWindowShadow(self, texture)
		local shadow = self.frame.shadow

		local shadowTexture = Healcsi:GetMedia("shadow", texture)

		shadow.left:SetTexture(shadowTexture)
		shadow.left:SetTexCoord(0, 0.25, 0.25, 0.75)
		shadow.left:SetWidth(128)

		shadow.right:SetTexture(shadowTexture)
		shadow.right:SetTexCoord(0.75, 1, 0.25, 0.75)
		shadow.right:SetWidth(128)

		shadow.top:SetTexture(shadowTexture)
		shadow.top:SetTexCoord(0.25, 0.75, 0, 0.25)
		shadow.top:SetHeight(128)

		shadow.bottom:SetTexture(shadowTexture)
		shadow.bottom:SetTexCoord(0.25, 0.75, 0.75, 1)
		shadow.bottom:SetHeight(128)

		shadow.topLeft:SetTexture(shadowTexture)
		shadow.topLeft:SetTexCoord(0, 0.25, 0, 0.25)
		shadow.topLeft:SetSize(128, 128)

		shadow.topRight:SetTexture(shadowTexture)
		shadow.topRight:SetTexCoord(0.75, 1, 0, 0.25)
		shadow.topRight:SetSize(128, 128)

		shadow.bottomLeft:SetTexture(shadowTexture)
		shadow.bottomLeft:SetTexCoord(0, 0.24, 0.75, 1)
		shadow.bottomLeft:SetSize(128, 128)

		shadow.bottomRight:SetTexture(shadowTexture)
		shadow.bottomRight:SetTexCoord(0.75, 1, 0.75, 1)
		shadow.bottomRight:SetSize(128, 128)

	end

	local function SetBgFile(self, file)
		self.content.background.mediaWidth  = Healcsi.Media.bgWidth[file]
		self.content.background.mediaHeight = Healcsi.Media.bgHeight[file]
		if (self.content.background.texture) then
			self.content.background.texture:Hide()
		end
		self.content.background.texture = Healcsi:SetMiddleBackground(self.content, "ARTWORK", Healcsi.Media.bgTextures[file], Healcsi.Media.bgWidth[file], Healcsi.Media.bgHeight[file], 0, 1024)
	end

	local function OnAcquire(self)

		self:SetWindowShadow("Shadow-3")
		self:SetWindowShadowColor(Healcsi:GetColor("DarkBlue", 1))

		self:UpdateMenu()

		TitleOnMouseUp(self.title)
	end

	local function Constructor()
		local frame = CreateFrame("Frame", nil, TopParent)
		local self = {}
		self.type = "Window"

		self.Hide = Hide
		self.Show = Show
		self.OnAcquire  = OnAcquire
		self.UpdateMenu = UpdateMenu
		self.SetBgFile  = SetBgFile

		self.SetWindowShadow = SetWindowShadow
		self.SetWindowShadowColor = SetWindowShadowColor

		self.frame = frame
		frame.obj = self

		self.Database = Healcsi.db.global.configWindow

		frame:SetWidth(self.Database.width)
		frame:SetHeight(self.Database.height)
		if (self.Database.offset.top ~= 0) or (self.Database.offset.left ~= 0) then
			frame:SetPoint("TOPLEFT", TopParent, "BOTTOMLEFT", self.Database.offset.left, self.Database.offset.top)
		else
			frame:SetPoint("CENTER", TopParent, "CENTER", 0, 0)
		end
		frame:EnableMouse()
		frame:SetMovable(true)
		frame:SetResizable(true)
		frame:SetMinResize(520,240)
		frame:SetToplevel(true)
		frame:SetFrameStrata("FULLSCREEN_DIALOG")
		frame:SetScript("OnMouseDown", FrameOnMouseDown)

		frame:SetScript("OnShow", FrameOnShow)
		frame:SetScript("OnHide", FrameOnClose)

		frame:SetScript("OnUpdate", OnWindowUpdate)

		local frameTexture = Healcsi:CreateTextureFull(frame, "BACKGROUND")
		Healcsi:DecorateTexture(frameTexture, "background", "Solid", "BlueBackground", 0.9)

		-- Shadow: ============================================
		--

		local leftShadow = frame:CreateTexture(nil, "BORDER")
		leftShadow:SetPoint("TOPRIGHT", frame, "TOPLEFT")
		leftShadow:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT")

		local rightShadow = frame:CreateTexture(nil, "BORDER")
		rightShadow:SetPoint("TOPLEFT", frame, "TOPRIGHT")
		rightShadow:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT")

		local topShadow = frame:CreateTexture(nil, "BORDER")
		topShadow:SetPoint("BOTTOMLEFT", frame, "TOPLEFT")
		topShadow:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT")

		local bottomShadow = frame:CreateTexture(nil, "BORDER")
		bottomShadow:SetPoint("TOPLEFT", frame, "BOTTOMLEFT")
		bottomShadow:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT")

		frame.shadow = frame.shadow or {}

		frame.shadow.left = leftShadow
		frame.shadow.right = rightShadow
		frame.shadow.top = topShadow
		frame.shadow.bottom = bottomShadow

		local topLeftShadow = frame:CreateTexture(nil, "BORDER")
		topLeftShadow:SetPoint("BOTTOMRIGHT", frame, "TOPLEFT")

		local topRightShadow = frame:CreateTexture(nil, "BORDER")
		topRightShadow:SetPoint("BOTTOMLEFT", frame, "TOPRIGHT")

		local bottomLeftShadow = frame:CreateTexture(nil, "BORDER")
		bottomLeftShadow:SetPoint("TOPRIGHT", frame, "BOTTOMLEFT")

		local bottomRightShadow = frame:CreateTexture(nil, "BORDER")
		bottomRightShadow:SetPoint("TOPLEFT", frame, "BOTTOMRIGHT")

		frame.shadow.topLeft = topLeftShadow
		frame.shadow.topRight = topRightShadow
		frame.shadow.bottomLeft = bottomLeftShadow
		frame.shadow.bottomRight = bottomRightShadow
		
		-- Close: ============================================
		--
		local close = CreateFrame("Frame", nil, frame)
		close:SetSize(24, 24)
		close:SetPoint("TOPRIGHT", -4, -4)
		close:SetScript("OnEnter", CloseOnEnter)
		close:SetScript("OnLeave", CloseOnLeave)
		close:SetScript("OnMouseDown", CloseWindow)
		close:SetFrameLevel(50)
		self.closeButton = close
		close.obj = self

		local closeTexture = Healcsi:SetMiddleBackground(close, "ARTWORK", [[Interface\Addons\Healcsi\Media\UI\close]], 512, 512, 3, 512)
		Healcsi:SetVertexColor(closeTexture, "DarkBlue", 1)
		self.closeButton.texture = closeTexture

		-- Sizer: ============================================
		--
		local sizer = CreateFrame("Frame", nil, frame)
		sizer:SetPoint("BOTTOMRIGHT",frame,"BOTTOMRIGHT", -2, 2)
		sizer:SetSize(15, 15)
		sizer:EnableMouse()
		sizer:SetScript("OnMouseDown", SizerOnMouseDown)
		sizer:SetScript("OnMouseUp", SizerOnMouseUp)
		self.sizer = sizer

		self.sizer.texture = Healcsi:SetMiddleBackground(sizer, "ARTWORK", [[Interface\Addons\Healcsi\Media\UI\resize]], 256, 256, 0, 256)

		-- Head Line: ============================================
		--
		local headLineTexture = Healcsi:CreateTextureHoriz(frame, "OVERLAY", {2, 2}, 32, 4)
		Healcsi:SetTexture(headLineTexture, "background", [[Interface\Addons\Healcsi\Media\UI\headLine]], "White", 0.6)

		local title = CreateFrame("Button", nil, frame)

		-- Title: ============================================
		--

		local title = CreateFrame("Frame", nil, frame)
		title:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0)
		title:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 0, -35)
		title:EnableMouse()
		title:SetFrameLevel(10)
		title:SetScript("OnMouseDown", TitleOnMouseDown)
		title:SetScript("OnMouseUp", TitleOnMouseUp)
		self.title = title

		local menu = {}
		self.menu = menu

		local modules = CreateFrame("Frame", nil, frame)
		modules:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -160, -6)
		modules:SetSize(90, 22)
		modules:SetFrameLevel(50)
		modules:SetScript("OnMouseDown", ModulesOnMouseDown)
		modules:SetScript("OnMouseUp", ModulesOnMouseUp)
		modules:SetScript("OnEnter", ModulesOnEnter)
		modules:SetScript("OnLeave", ModulesOnLeave)
		menu.modules = modules

		local modulesTexture = Healcsi:CreateTextureFull(modules, "BACKGROUND")
		--Healcsi:DecorateTexture(modulesTexture, "background", "Solid", "Red", 0.6)
		modulesTexture:Hide()
		menu.modules.texture = modulesTexture


		local modulesText = modules:CreateFontString("modules", "ARTWORK")
		modulesText:SetAllPoints(modules)
		modulesText:SetFont(Healcsi:GetFont("MoonBold"), 16)
		modulesText:SetJustifyH("CENTER")
		modulesText:SetJustifyV("MIDDLE")
		modulesText:SetText("Modules")
		modulesText:SetTextColor(0, 0, 0, 1)
		menu.modules.text = modulesText

		local menuSparatorTexture = frame:CreateTexture("MenuSeparator", "ARTWORK")
		Healcsi:SetTexture(menuSparatorTexture, "background", [[Interface\Addons\Healcsi\Media\UI\menu]])
		menuSparatorTexture:SetSize(4, 28)
		menuSparatorTexture:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -150, -2)

		local profiles = CreateFrame("Frame", nil, frame)
		profiles:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -50, -6)
		profiles:SetSize(90, 22)
		profiles:SetFrameLevel(50)
		profiles:SetScript("OnMouseDown", ProfilesOnMouseDown)
		profiles:SetScript("OnMouseUp", ProfilesOnMouseUp)
		profiles:SetScript("OnEnter", ProfilesOnEnter)
		profiles:SetScript("OnLeave", ProfilesOnLeave)
		menu.profiles = profiles

		local profilesTexture = Healcsi:CreateTextureFull(profiles, "BACKGROUND")
		profilesTexture:Hide()
		--Healcsi:DecorateTexture(profilesTexture, "background", "Solid", "Red", 0.6)
		menu.profiles.texture = profilesTexture

		local profilesText = profiles:CreateFontString("modules", "ARTWORK")
		profilesText:SetAllPoints(profiles)
		profilesText:SetFont(Healcsi:GetFont("MoonBold"), 16)
		profilesText:SetJustifyH("CENTER")
		profilesText:SetJustifyV("MIDDLE")
		profilesText:SetText("Profiles")
		profilesText:SetTextColor(0, 0, 0, 1)
		menu.profiles.text = profilesText

		--
		local dropDownHolder = CreateFrame("Frame", nil, frame)
		dropDownHolder:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, -6)
		dropDownHolder:SetPoint("BOTTOMRIGHT", frame, "TOPLEFT", 250, -35)
		dropDownHolder:SetFrameLevel(50)

		local moduleSelector = Gui:Create("Healcsi-ModuleSelector") 		
		moduleSelector:SetPoint("TOPLEFT", dropDownHolder, "TOPLEFT", 0, 0)
		
		moduleSelector.frame:SetParent(dropDownHolder)
		moduleSelector.frame:SetFrameLevel(80)
		
		moduleSelector.dropdown:SetParent(dropDownHolder)
		moduleSelector.dropdown:SetFrameLevel(80)

		moduleSelector.frame:Show()

		menu.modules.selector = moduleSelector
		menu.modules.selector.holder = dropDownHolder

		
		
		if (self.Database.selectedMenu == "profiles") then
			self.menu.profiles.selected = true
			self.menu.modules.selected  = nil
		end

		if (self.Database.selectedMenu == "modules") then
			self.menu.modules.selected  = true
			self.menu.profiles.selected = nil
		end

		-- Container Support: ====================================
		--
		local content = CreateFrame("Frame", nil, frame)
		content:SetFrameLevel(80)
		self.content = content
		content.obj = self
		Healcsi:SetPadding(content, frame, {40, 5, 18, 5})

		content.background = {}
		--content.background.texture = Healcsi:SetMiddleBackground(content, "ARTWORK", "ShamanBackground", 1024, 793, 0, 1024)

		-- Memory: ====================================
		--


		local memory = CreateFrame("Frame", nil, frame)
		self.memory = memory
		memory:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 5, 0)
		memory:SetSize(400, 18)
		memory:SetFrameLevel(50)
		memory:SetScript("OnUpdate", MemoryUpdate)

		local memoryText = memory:CreateFontString("memory", "ARTWORK")
		memoryText:SetAllPoints(memory)
		memoryText:SetFont(Healcsi:GetFont("Friz Quadrata TT"), 14) -- MoonBold
		memoryText:SetJustifyH("LEFT")
		memoryText:SetJustifyV("MIDDLE")
		memoryText:SetText("Memory Usage: " .. Healcsi:GetMemoryUsageString())
		memoryText:SetTextColor(0, 0, 0, 1)
		memory.text = memoryText

		--local contentTexture = Healcsi:CreateTextureFull(content, "BACKGROUND", {0, 0, 0, 0})
		--Healcsi:DecorateTexture(contentTexture, "background", "Solid", "Red", 1)

		Gui:RegisterAsContainer(self)
		return self
	end

	Gui:RegisterWidgetType(Type, Constructor, Version)
end