
local Healcsi       = Healcsi
local Console       = Healcsi.Console
local Media         = Healcsi.Media
local Gui           = Healcsi.Gui
local L             = Healcsi.Locale
local Modules       = Healcsi.Modules
local Editor        = Healcsi.Editor
local ClickCast     = Healcsi.Modules.ClickCast
local UnitButtons   = Healcsi.Modules.UnitButtons


function ClickCast:Initialize()

	self.BINDING_MODIFIERS = { "", "shift-", "ctrl-", "alt-", "ctrl-shift-", "alt-shift-", "alt-ctrl-", "alt-ctrl-shift-" }
	self.BINDING_MODIFIERS_DISPLAY = { L["direct"], "Shift-", "Ctrl-", "Alt-", "Ctrl-Shift-", "Alt-Shift-", "Alt-Ctrl-", "Alt-Ctrl-Shift-" }
	-- self.ACTION_TYPES = { togglemenu = 1, target = 1, assist = 1, focus = 1, spell = 2, macro = 2, buildin = 2 }

	self.playerClass = select(2, UnitClass("player"))

	self:RegisterForClicks()
	self:AssignAll()

	self.initDone = true
end

function ClickCast:IsInitDone()
	return self.initDone
end


function ClickCast:GetSpecialMacro()
	return "/stopcasting\n/target mouseover\n/click ExtraActionButton1\n/targetlasttarget"
end

-- local MACRO_FORMAT = "/stopcasting\n/cast %s\n/cast [@mouseover] %s"

local function AssignBindToUnitFrame(frame, modifier, id, action, extra)
	local bindType, bindKey, bindValue

	if action == "action" then

		bindType = extra

	elseif action == "spell" then

		bindType = "spell"
		bindKey = "spell"
		bindValue = extra

	elseif action == "transfer" then

		if (extra == "extraBtn") then
			bindType = "macro"
			bindKey = "macrotext"
			bindValue = ClickCast:GetSpecialMacro()
		end

	elseif action == "buildin" then

		bindType = "spell"
		bindKey = "spell"
		bindValue = extra

	elseif action == "macroText" then

		bindType = "macro"
		bindKey = "macrotext"
		bindValue = gsub(extra or "", "##", "@mouseover")

	elseif action == "macroName" then

		bindType = "macro"
		bindKey = "macro"
		bindValue = gsub(extra or "", "##", "@mouseover")

	end


	if (bindType) then
		frame:SetAttribute(modifier.."type"..id, bindType)
		if (bindKey) then
			frame:SetAttribute(modifier..bindKey..id, bindValue)
		end
	else
		bindType = ATTRIBUTE_NOOP
		if modifier == "" then
			id = tonumber(id)
			if id == 1 then
				bindType = "target"
			elseif id == 2 then
				bindType = "togglemenu"
			end
		end
		frame:SetAttribute(modifier.."type"..id, bindType)
	end

end

function ClickCast:AssignBind(modifier, id, action, extra)
	modifier, id = self:DecodeWheelButton(modifier, id)
	for _, frame in ipairs(UnitButtons.unitFrames) do
		AssignBindToUnitFrame(frame, modifier, id, action, extra)
	end
end

function ClickCast:RegisterForClicks()
	for _, frame in ipairs(UnitButtons.unitFrames) do
		if (not(Editor.InCombat)) then
			frame:RegisterForClicks(Healcsi.db.profile.ClickCast.clickDownMode)
		else
			ClickCast.needRegister = 1
		end
	end
end

Healcsi:RegisterEventCallback("OutOfCombat", function()
	if (ClickCast.needRegister) then
		for _, frame in ipairs(UnitButtons.unitFrames) do
			ClickCast:RegisterForClicks()
		end
		ClickCast.needRegister = nil
	end
	if (ClickCast.needClear) then
		ClickCast:ClearAll()
		ClickCast.needClear = nil
	end
	if (ClickCast.needAssign) then
		ClickCast:AssignAll()
		ClickCast.needAssign = nil
	end
end)

Healcsi:RegisterEventCallback("ClickCastConfigChanged", function()
	ClickCast:RegisterForClicks()
	ClickCast:AssignAll()
	ClickCast:UpdateWheelCast()
end)

Healcsi:RegisterEventCallback("GlobalConfigRefresh", function()
	ClickCast:RegisterForClicks()
	ClickCast:AssignAll()
	ClickCast:UpdateWheelCast()
end)

Healcsi:RegisterEventCallback("OnUnitNotify", function(frame, event)
	if (event == "UnitButtonCreated") then
		local db = Healcsi.db.profile.ClickCast.clickSet[self.playerClass]

		local key, value
		for key, value in pairs(db) do
			if type(key) == "string" and (type(value) == "string" or type(value) == "number") then

				local modifier, id, spec = strmatch(key, "(.-)(%d+)_-(.-)$")
				modifier, id = self:DecodeWheelButton(modifier, id)
				local bindStr = self:ToBindString(value, spec)

				if ((id) and (type(bindStr) == "string")) then
					local action, extra = strmatch(bindStr, "(.-):(.+)")
					self:AssignBind(modifier, tonumber(id), action, extra)
				end
			end
		end

		if ClickCast:HasWheelBinds() then
			ClickCast:EnableWheelCastOnFrame(frame)
		end
	end
end)

function ClickCast:ClearAll()
	if (not(Editor.InCombat)) then
		local _, modifier, id
		for _, modifier in ipairs(self.BINDING_MODIFIERS) do
			for id = 1, 7 do
				self:AssignBind(modifier, id)
			end
		end
	else
		self.needClear = 1
	end
end

function ClickCast:AssignAll()
	if (not(Editor.InCombat)) then
		self:ClearAll()
		local db = Healcsi.db.profile.ClickCast.clickSet[self.playerClass]
		local key, value
		for key, value in pairs(db) do
			if type(key) == "string" and (type(value) == "string" or type(value) == "number") then

				local modifier, id, spec = strmatch(key, "(.-)(%d+)_-(.-)$")
				local bindStr = self:ToBindString(value, spec)

				-- Healcsi:Print("bindStr: " .. (bindStr or ""))

				if ((id) and (type(bindStr) == "string")) then
					local action, extra = strmatch(bindStr, "(.-):(.+)")
					self:AssignBind(modifier, tonumber(id), action, extra)
				end
			end
		end
	else
		self.needAssign = 1
	end
end

-- all this action, extra, order, section, name..... VERY Confusing!!!
-- Let's talk about it here and now so ponies can live happily ever after:

-- In the editor.lua file in function called: CreateClickCastDropDown() I came up with my own tags
-- with those tags I could easily create the dropdowns, and the database uses those values
-- However the Bind section above in this file expects the format to be different
-- Since I don't want to spend hour and hours of time to figure out how to rewrite the whole binding (just some part of it...)
-- I created this conversion function

function ClickCast:ToBindString(dbValue, spec)
	local value

	if (spec == "_spellText") then
		value = "spell:" .. dbValue
		return value
	elseif (spec == "_macroText") then
		value = "macroText:".. dbValue
		return value
	elseif (spec == "_macroName") then
		value = "macroName:".. dbValue
		return value
	end

	local order, section, name = string.match(dbValue, "(%d+)-(%d+)-(.+)")

	if (section == "1") then

		if (name == "none") then
			value = ""
		elseif (name == "togglemenu" or name == "target" or name == "assist" or name == "focus" or name == "follow") then
			value = "action:" .. name
		else
			value = "transfer:extraBtn"
		end

	elseif (section == "2") then
		
		local spells = Healcsi.db.profile.ClickCast.spells[self.playerClass]
		value = "buildin:" .. spells[tonumber(order)] -- Buildin spell

	end

	return value
end

