
local Healcsi      = Healcsi
local Console      = Healcsi.Console
local Media        = Healcsi.Media
local Gui          = Healcsi.Gui
local L            = Healcsi.Locale
local Modules      = Healcsi.Modules
local ClickCast  = Healcsi.Modules.ClickCast
local Editor       = Healcsi.Editor

Healcsi:RegisterEventCallback("UpdateEditorLayout", function()
	ClickCast:RefreshScrollBar()
end)

Healcsi:RegisterEventCallback("GlobalConfigRefresh", function()
	ClickCast.parentFrame:ReleaseChildren()
	ClickCast:CreateClickCastEditorView(ClickCast.parentFrame)
end)

function ClickCast:RefreshScrollBar()
	if (not(self.scrollFrame)) then return end

	local pos = self.scrollFrame.scrollbar:GetValue()

	self.scrollFrame:DoLayout()
	self.scrollFrame:FixScroll() -- avoid jumpy scrollbar

	self.scrollFrame.scrollbar:SetValue(pos) -- set position
	self.scrollFrame:SetScroll(pos) -- just to be sure. don't think this is needed btw...
end


function ClickCast:DrawClickModeGroup(container)
	local holder = Editor:CreateGroup(container, true, nil, {10, 0, 10, 10}, {0, 5, 0, 5}, "BlueSelected", 0.3)
	local db = Healcsi.db.profile.ClickCast

	local modes = {
		["AnyDown"] = "Cast on mouse press",
		["AnyUp"]   = "Cast on mouse release"
	}
	local line1 = Editor:CreateGroup(holder, true, nil, {5, 0, 5, 10})
	Editor:CreateDropDown(line1, false, 300, L["ClickMode"], modes, db, "clickDownMode", "UnitButtonClickModeMode", "ClickCastConfigChanged", true)

end

function ClickCast:DrawLeftGroup(container)
	local holder = Editor:CreateGroup(container, true, nil, {10, 0, 10, 10}, {0, 5, 0, 5}, "BlueSelected", 0.3)
	local db = Healcsi.db.profile.ClickCast
	Editor:CreateClickCastEditor(holder, db, "1", "ClickCastConfigChanged", self.BINDING_MODIFIERS, self.BINDING_MODIFIERS_DISPLAY)
end

function ClickCast:DrawRightGroup(container)
	local holder = Editor:CreateGroup(container, true, nil, {10, 0, 10, 10}, {0, 5, 0, 5}, "BlueSelected", 0.3)
	local db = Healcsi.db.profile.ClickCast
	Editor:CreateClickCastEditor(holder, db, "2", "ClickCastConfigChanged", self.BINDING_MODIFIERS, self.BINDING_MODIFIERS_DISPLAY)
end

function ClickCast:DrawMiddleGroup(container)
	local holder = Editor:CreateGroup(container, true, nil, {10, 0, 10, 10}, {0, 5, 0, 5}, "BlueSelected", 0.3)
	local db = Healcsi.db.profile.ClickCast
	Editor:CreateClickCastEditor(holder, db, "3", "ClickCastConfigChanged", self.BINDING_MODIFIERS, self.BINDING_MODIFIERS_DISPLAY)
end

function ClickCast:DrawFourGroup(container)
	local holder = Editor:CreateGroup(container, true, nil, {10, 0, 10, 10}, {0, 5, 0, 5}, "BlueSelected", 0.3)
	local db = Healcsi.db.profile.ClickCast
	Editor:CreateClickCastEditor(holder, db, "4", "ClickCastConfigChanged", self.BINDING_MODIFIERS, self.BINDING_MODIFIERS_DISPLAY)
end

function ClickCast:DrawFiveGroup(container)
	local holder = Editor:CreateGroup(container, true, nil, {10, 0, 10, 10}, {0, 5, 0, 5}, "BlueSelected", 0.3)
	local db = Healcsi.db.profile.ClickCast
	Editor:CreateClickCastEditor(holder, db, "5", "ClickCastConfigChanged", self.BINDING_MODIFIERS, self.BINDING_MODIFIERS_DISPLAY)
end

function ClickCast:DrawScrollUpGroup(container)
	local holder = Editor:CreateGroup(container, true, nil, {10, 0, 10, 10}, {0, 5, 0, 5}, "BlueSelected", 0.3)
	local db = Healcsi.db.profile.ClickCast
	Editor:CreateClickCastEditor(holder, db, "6", "ClickCastConfigChanged", self.BINDING_MODIFIERS, self.BINDING_MODIFIERS_DISPLAY)
end

function ClickCast:DrawScrollDownGroup(container)
	local holder = Editor:CreateGroup(container, true, nil, {10, 0, 10, 10}, {0, 5, 0, 5}, "BlueSelected", 0.3)
	local db = Healcsi.db.profile.ClickCast
	Editor:CreateClickCastEditor(holder, db, "7", "ClickCastConfigChanged", self.BINDING_MODIFIERS, self.BINDING_MODIFIERS_DISPLAY)
end


local function SelectGroup(container, event, group)
	container:ReleaseChildren()

	local scrollFrame = Gui:Create("Healcsi-ScrollFrame")
	scrollFrame:SetLayout("Flow")
	scrollFrame:SetFullWidth(true)
	scrollFrame:SetFullHeight(true)
	ClickCast.scrollFrame = scrollFrame
	container:AddChild(scrollFrame)

	if (group == "clickmode") then
		ClickCast:DrawClickModeGroup(scrollFrame)
	elseif (group == "left") then
		ClickCast:DrawLeftGroup(scrollFrame)
	elseif (group == "right") then
		ClickCast:DrawRightGroup(scrollFrame)
	elseif (group == "mid") then
		ClickCast:DrawMiddleGroup(scrollFrame)
	elseif (group == "four") then
		ClickCast:DrawFourGroup(scrollFrame)
	elseif (group == "five") then
		ClickCast:DrawFiveGroup(scrollFrame)
	elseif (group == "scrup") then
		ClickCast:DrawScrollUpGroup(scrollFrame)
	elseif (group == "scrdown") then
		ClickCast:DrawScrollDownGroup(scrollFrame)
	end

	Healcsi.db.profile.ClickCast.selectedTab = group

	ClickCast:RefreshScrollBar()
end

function ClickCast:CreateClickCastEditorView(frame)

	local tab =  Gui:Create("Healcsi-TabGroup")
	tab:SetFullWidth(true)
	tab:SetFullHeight(true)
	tab:SetLayout("Flow") -- Do not set Fill or else it will freeze... brrrr...

	local tabs = {
		{text="Click Mode",  value="clickmode"},
		{text="Left",        value="left"},
		{text="Right",       value="right"},
		{text="Middle",      value="mid"},
		{text="4",           value="four"},
		{text="5",           value="five"},
		{text="Scroll Up",   value="scrup"},
		{text="Scroll Down", value="scrdown"},
	}

	tab:SetTabs(tabs)
	tab:SetCallback("OnGroupSelected", SelectGroup)

	local selectedTab = Healcsi.db.profile.ClickCast.selectedTab or "left"

	local found = false
	for i, currTab in ipairs(tabs) do
		if(currTab["value"] == selectedTab) then found = true end
	end

	if (found) then
		tab:SelectTab(selectedTab)
	end
	
	frame:AddChild(tab)

end

function ClickCast:CreateEditor()

	--[[
	local scrollFrame = Gui:Create("Healcsi-ScrollFrame")
	scrollFrame:SetLayout("Flow")
	scrollFrame:SetFullWidth(true)
	scrollFrame:SetFullHeight(true)
	self.scrollFrame = scrollFrame
	]]

	local parentFrame = Gui:Create("SimpleGroup")
	parentFrame:SetLayout("Flow")
	parentFrame:SetFullWidth(true)
	parentFrame:SetFullHeight(true)
	self.parentFrame = parentFrame

	-- Put page content here

	self:CreateClickCastEditorView(parentFrame)

	-- End of page content

	--scrollFrame:AddChild(parentFrame)
	--return scrollFrame

	return parentFrame
end