
local Healcsi       = Healcsi
local Console       = Healcsi.Console
local Media         = Healcsi.Media
local Gui           = Healcsi.Gui
local L             = Healcsi.Locale
local Modules       = Healcsi.Modules
local Editor        = Healcsi.Editor
local ToolBox       = Healcsi.Modules.ToolBox
local UnitButtons   = Healcsi.Modules.UnitButtons
local UnitFrames    = Healcsi.Modules.UnitFrames

local ConvertToParty = ConvertToParty
local ConvertToRaid = ConvertToRaid
local SetEveryoneIsAssistant = SetEveryoneIsAssistant
local IsEveryoneAssistant = IsEveryoneAssistant
local RegisterStateDriver = RegisterStateDriver


local function IsAllowed(officer)
	local group, leadship, num = UnitFrames:GetGroupStats() 
	if group == "raid" then
		return leadship == "leader" or (officer and leadship == "officer")
	elseif group == "party" then
		return leadship == "leader"
	end
end

function ToolBox:CreateGeneral()
	local frame = self:CreateToolbox("HealcsiToolboxGeneral", 1, 1, 0, GENERAL, L["GeneralToolBoxTooltip"])
	local menu = frame:CreateMenu(GENERAL)

	-- Refresh
	local button = menu:AddClickButton(L["RefreshFrames"])
	button:SetFrameRef("container", UnitFrames:GetContainer())
	button:SetAttribute("_onclick", [[
		self:GetParent():Hide()
		local container = self:GetFrameRef("container")
		container:Hide()
		container:Show()
	]])

	button = menu:AddClickButton(CONVERT_TO_RAID, "SecureHandlerStateTemplate")

	button:SetScript("OnUpdate", function(self)
		self:Grayout(not IsAllowed())
	end)

	function button:UpdateGroup(value)
		if value == 2 then
			self.text:SetText(CONVERT_TO_PARTY)
			self.func = ConvertToParty
		else
			self.text:SetText(CONVERT_TO_RAID)
			self.func = ConvertToRaid
		end
	end

	function button:OnClick()
		local group, leadship, num = UnitFrames:GetGroupStats() 
		if leadship == "leader" then
			self.func()
		else
			ToolBox:PrintPermissionError()
		end
	end

	button:SetAttribute("_onstate-groupstate", [[
		if newstate == 0 then
			self:Disable()
		else
			self:Enable()
		end
		self:CallMethod("UpdateGroup", newstate)
	]])

	RegisterStateDriver(button, "groupstate", "[group: raid] 2; [group: party] 1; 0")

	button = menu:AddClickButton(ALL_ASSIST_LABEL_LONG, "SecureHandlerStateTemplate")

	button:SetScript("OnUpdate", function(self)
		self:Grayout(not UnitIsGroupLeader("player"))
	end)

	function button:OnClick()
		local group, leadship, num = UnitFrames:GetGroupStats() 
		if leadship == "leader" then
			SetEveryoneIsAssistant(not IsEveryoneAssistant())
		else
			ToolBox:PrintPermissionError(ALL_ASSIST_NOT_LEADER_ERROR)
		end
	end

	button:SetScript("OnShow", function(self)
		if IsEveryoneAssistant() then
			self.check:Show()
		else
			self.check:Hide()
		end
	end)

	button:SetAttribute("_onstate-inraid", [[
		if newstate == 1 then
			self:Enable()
		else
			self:Disable()
		end
	]])

	RegisterStateDriver(button, "inraid", "[group: raid] 1; 0")

	local function GroupToolButton_OnClick(self)
		if self.grayed then
			ToolBox:PrintPermissionError()
		else
			self.func()
		end
	end

	local function CreateGroupToolButton(text, func)
		local button = menu:AddClickButton(text, "SecureHandlerStateTemplate")
		button.func = func
		button.OnClick = GroupToolButton_OnClick

		button:SetAttribute("_onstate-ingroup", [[
			if newstate == 1 then
				self:Enable()
			else
				self:Disable()
			end
		]])

		RegisterStateDriver(button, "ingroup", "[group] 1; 0")

		button:SetScript("OnUpdate", function(self)
			self:Grayout(not IsAllowed(1))
		end)

		return button
	end

	CreateGroupToolButton(ROLE_POLL, InitiateRolePoll)
	CreateGroupToolButton(READY_CHECK, DoReadyCheck)

	local lockButton = menu:AddClickButton(L["lockPositionCB"])


	function lockButton:OnClick()
		local db = Healcsi.db.profile.UnitFrames.general
		if (db.lockPosition) then
			db.lockPosition = false
		else
			db.lockPosition = true
		end
		Healcsi:BroadcastEvent("UnitFrameConfigChanged")
	end

	if (Healcsi.db.profile.UnitFrames.general.lockPosition) then
		lockButton.check:Show()
	else
		lockButton.check:Hide()
	end

	Healcsi:RegisterEventCallback("UnitFrameConfigChanged", function()
		if (Healcsi.db.profile.UnitFrames.general.lockPosition) then
			lockButton.check:Show()
		else
			lockButton.check:Hide()
		end
	end)

	button = menu:AddClickButton(SETTINGS.."...")
	function button:OnClick()
		if (Healcsi.configWindow:IsShown()) then
			Healcsi.configWindow:Hide()
		else
			Healcsi.configWindow:Show()
		end
	end

	menu:Finish()

end
