
local Healcsi       = Healcsi
local Console       = Healcsi.Console
local Media         = Healcsi.Media
local Gui           = Healcsi.Gui
local L             = Healcsi.Locale
local Modules       = Healcsi.Modules
local Editor        = Healcsi.Editor
local ToolBox       = Healcsi.Modules.ToolBox
local UnitButtons   = Healcsi.Modules.UnitButtons
local UnitFrames    = Healcsi.Modules.UnitFrames

local RegisterStateDriver = RegisterStateDriver
local UnregisterStateDriver = UnregisterStateDriver
local format = format

if not UnitFrames.Raid_GetRaidGroup then return end

function ToolBox:CreateGroupFilters()
	local frame = self:CreateToolbox("HealcsiToolboxGroupFilter", 1, 1, 1, RAID, L["SubGroupsToolBoxTooltip"])
	frame:Hide()

	if (Healcsi.db.profile.UnitFrames.raid.keepGroupsTogether) then
		RegisterStateDriver(frame, "visibility", "[group:raid] show; hide")
	else
		UnregisterStateDriver(frame, "visibility")
		frame:Hide()
	end

	Healcsi:RegisterEventCallback("UnitFrameConfigChanged", function()
		if (Healcsi.db.profile.UnitFrames.raid.keepGroupsTogether) then
			RegisterStateDriver(frame, "visibility", "[group:raid] show; hide")
		else
			UnregisterStateDriver(frame, "visibility")
			frame:Hide()
		end
	end)

	local menu = frame:CreateMenu(RAID)

	local button = menu:AddClickButton(L["GroupSwap"], nil, nil, 1)

	function button:OnClick()
		if ToolBox:GroupSwapEnabled() then
			ToolBox:DisableGroupSwap()
		else
			ToolBox:EnableGroupSwap()
		end
	end

	button:SetScript("OnShow", function(self)
		if ToolBox:CanGroupSwap() then
			self.text:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
			if ToolBox:GroupSwapEnabled() then
				self.check:Show()
			else
				self.check:Hide()
			end
		else
			self.text:SetTextColor(GRAY_FONT_COLOR.r, GRAY_FONT_COLOR.g, GRAY_FONT_COLOR.b)
		end
	end)

	local function Button_OnUpdate(self)
		local filterGroup = self.filterGroup
		if filterGroup:IsShown() then
			self.check:Show()
		else
			self.check:Hide()
		end

		local count = 0
		local i
		for i = 1, #filterGroup do
			if filterGroup[i]:IsVisible() then
				count = i
			else
				break
			end
		end
		self.count:SetText(count > 0 and count or nil)
	end

	local i
	for i = 1, 8 do
		local button = menu:AddButton(format(GROUP_NUMBER, i), "SecureHandlerClickTemplate")
		local count = button:CreateFontString(nil, "ARTWORK", "GameFontGreenSmall")
		button.count = count
		count:SetPoint("RIGHT", -4, 0)

		local filterGroup = UnitFrames:Raid_GetRaidGroup(i)
		button.filterGroup = filterGroup
		button:SetFrameRef("filtergroup", filterGroup)

		button:SetScript("OnUpdate", Button_OnUpdate)

		button:SetAttribute("_onclick", [[
			local group = self:GetFrameRef("filtergroup")
			if group:IsShown() then
				group:Hide()
			else
				group:Show()
			end
		]])
	end

	menu:Finish()

end


