
local Healcsi       = Healcsi
local Console       = Healcsi.Console
local Media         = Healcsi.Media
local Gui           = Healcsi.Gui
local L             = Healcsi.Locale
local Modules       = Healcsi.Modules
local Editor        = Healcsi.Editor
local ToolBox       = Healcsi.Modules.ToolBox
local UnitButtons   = Healcsi.Modules.UnitButtons
local UnitFrames    = Healcsi.Modules.UnitFrames

local InCombatLockdown = InCombatLockdown
local strfind = strfind
local tonumber = tonumber
local select = select
local GetRaidRosterInfo = GetRaidRosterInfo
local SwapRaidSubgroup = SwapRaidSubgroup
local SetRaidSubgroup = SetRaidSubgroup
local GetMouseFocus = GetMouseFocus
local GetCursorPosition = GetCursorPosition


local function NotifyFrame_SetText(self, text)
	self.text:SetText(text)
	self:SetHeight(self.text:GetHeight() + 30)
end

local function NotifyFrame_IsClosed(self)
	return self.manualClosed
end

local function NotifyFrameCloseButton_OnClick(self)
	local parent = self:GetParent()
	parent:Hide()
	parent.manualClosed = 1
	if type(parent.OnClose) == "function" then
		parent:OnClose()
	end
end

local function CreateNotifyFrame(name, parent, width, closable, arrowFlag)
	local frame = CreateFrame("Frame", name, parent, "GlowBoxTemplate")
	frame:SetSize(width, 56)
	frame:Hide()
	frame:SetFrameStrata("FULLSCREEN_DIALOG")

	local arrow = CreateFrame("Frame", frame:GetName().."Arrow", frame, "GlowBoxArrowTemplate")
	frame.arrow = arrow

	if arrowFlag == "LEFT" then
		arrow:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", 4, 4)
	elseif arrowFlag == "RIGHT" then
		arrow:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT", -4, 4)
	else
		arrow:SetPoint("TOP", frame, "BOTTOM", 0, 4)
	end

	local text = frame:CreateFontString(frame:GetName().."Text", "OVERLAY", "GameFontHighlightLeft")
	frame.text = text
	text:SetNonSpaceWrap(true)
	text:SetPoint("TOPLEFT", 12, -16)
	text:SetWidth(width - 24)

	if closable then
		local topClose = CreateFrame("Button", frame:GetName().."TopClose", frame, "UIPanelCloseButton")
		frame.topClose = topClose
		topClose:SetSize(32, 32)
		topClose:SetPoint("TOPRIGHT", 6, 6)
		topClose:SetScript("OnClick", NotifyFrameCloseButton_OnClick)
	end

	frame.SetText = NotifyFrame_SetText
	frame.IsClosed = NotifyFrame_IsClosed

	return frame
end


function ToolBox:CanGroupSwap()
	if not Healcsi.db.profile.UnitFrames.raid.keepGroupsTogether then
		return nil, "CONFIG"
	end

	if InCombatLockdown() then
		return nil, "COMBAT"
	end

	local group, leadship, num = UnitFrames:GetGroupStats() 

	if not leadship then
		return nil, "PRIV"
	end

	return 1
end

function ToolBox:EnableGroupSwap()
	local enable, reason = self:CanGroupSwap()
	if enable then
		self.GroupSwap.frame:Show()
	elseif reason == "COMBAT" then
		self:PrintPermissionError(L["CannotChangeGroupsInCombat"])
	elseif reason == "PRIV" then
		self:PrintPermissionError()
	end
end

function ToolBox:DisableGroupSwap()
	self.GroupSwap.frame:Hide()
end

function ToolBox:GroupSwapEnabled()
	return self.GroupSwap.frame:IsShown()
end


Healcsi:RegisterEventCallback("UnitButtonConfigChanged", function()
	ToolBox:DisableGroupSwap()
end)

Healcsi:RegisterEventCallback("UnitFrameConfigChanged", function()
	ToolBox:DisableGroupSwap()
end)


local Orig_Raid_GetFramesMatrix = UnitFrames.Raid_GetFramesMatrix

local function Hooked_Raid_GetFramesMatrix(self)
	if UnitFrames:GetLayoutData() then
		return 5, 8
	else
		return 8, 5
	end
end

local function IsRaidUnit(unitId)
	local _, _, id = strfind(unitId or "", "^raid(%d+)$")
	return id
end

local function UnitToIndex(unitId)
	local id = IsRaidUnit(unitId)
	if id then
		return tonumber(id)
	end
end

local function Button_OnDragStart(self)
	if not ToolBox:GroupSwapEnabled() then
		return
	end

	local id = UnitToIndex(self.unit)
	if not id then
		return
	end

	local group = select(3, GetRaidRosterInfo(id))
	local i
	for i = 1, 8 do
		ToolBox.GroupSwap.groupPanels[i]:EnableMouse(i ~= group)
	end

	ToolBox.GroupSwap.cursor.button = self
	ToolBox.GroupSwap.cursor:Show()
end

local function Button_OnDragStop(self)
	ToolBox.GroupSwap.cursor:Hide()

	local targetFrame = GetMouseFocus()
	local i
	for i = 1, 8 do
		ToolBox.GroupSwap.groupPanels[i]:EnableMouse(false)
	end

	if not targetFrame or not targetFrame._HealcsiSubGroupSwap or InCombatLockdown() or not ToolBox:GroupSwapEnabled() or not IsRaidUnit(self.unit) then
		return
	end

	local srcId = UnitToIndex(self.unit)
	if not srcId then
		return
	end

	local srcSubGroup = select(3, GetRaidRosterInfo(srcId))

	local targetSubGroup, targetId = targetFrame._HealcsiSubGroupSwap
	if targetSubGroup == 0 then
		targetId = UnitToIndex(targetFrame.unit)
		if not targetId then
			return
		end

		targetSubGroup = select(3, GetRaidRosterInfo(targetId))
	end

	if srcSubGroup == targetSubGroup then
		return
	end

	if targetId then
		SwapRaidSubgroup(srcId, targetId)
	else
		SetRaidSubgroup(srcId, targetSubGroup)
	end
end

function ToolBox:CreateGroupSwap()

	local frame = CreateFrame("Frame", "HealcsiGroupSwapFrame", HealcsiRaidGroupsFrame)
	self.GroupSwap = {}
	self.GroupSwap.frame = frame
	frame:Hide()

	local db = Healcsi.db.profile.ToolBox

	local notifyText = frame:CreateFontString(frame:GetName().."Text", "ARTWORK", "GameFontHighlightLeft")
	notifyText:SetPoint("TOPLEFT", HealcsiUnitFrameContainerHolderDecoration, "BOTTOMLEFT", 8, -2)
	notifyText:SetPoint("TOPRIGHT", HealcsiUnitFrameContainerHolderDecoration, "BOTTOMRIGHT", -8, -2)
	notifyText:SetHeight(100)
	notifyText:SetJustifyV("TOP")
	notifyText:SetFont(STANDARD_TEXT_FONT, 12)
	notifyText:SetText(L["GroupSwapPrompt"])

	local arrowFrame = CreateNotifyFrame(frame:GetName().."ArrowFrame", HealcsiToolboxGroupFilter, 280, 1, "LEFT")
	arrowFrame:SetPoint("BOTTOMLEFT", HealcsiToolboxGroupFilter, "TOPRIGHT", -40, 20)
	arrowFrame:SetText(L["GroupSwapNotify"])

	function arrowFrame:OnClose()
		db.groupSwapNoNotify = 1
	end

	Healcsi:RegisterEventCallback("OnInitialize", function(db)
		if not db.groupSwapNoNotify then
			arrowFrame:Show()
		end
	end)


	local groupPanels = {}
	self.GroupSwap.groupPanels = groupPanels

	local i
	for i = 1, 8 do
		local panel = CreateFrame("Button", frame:GetName().."Group"..i, frame)
		groupPanels[i] = panel
		panel._HealcsiSubGroupSwap = i
		panel:SetFrameLevel(1)
		panel:SetAlpha(0.75)
		panel:EnableMouse(false)

		panel:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight", "ADD")
		panel:SetPoint("TOPLEFT", UnitFrames:Raid_GetRaidGroup(i), "TOPLEFT")

		local border = CreateFrame("Frame", nil, panel)
		border:SetAllPoints(panel)
		border:SetScale(0.5)
		border:SetBackdrop({ edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5 } })

		local numText = panel:CreateFontString(nil, "ARTWORK", "GameFontDisable")
		numText:SetPoint("BOTTOMRIGHT", -8, 8)
		numText:SetFormattedText("%d", i)
	end

	local cursor = CreateFrame("Frame", frame:GetName().."Cursor", frame)
	self.GroupSwap.cursor = cursor
	cursor:Hide()
	cursor:SetMovable(true)
	cursor:SetFrameStrata("HIGH")

	local border = CreateFrame("Frame", nil, cursor)
	border:SetAllPoints(cursor)
	border:SetBackdrop({ bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", tile = true, tileSize = 16, edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5 } })
	border:SetScale(0.5)
	border:SetAlpha(0.75)

	cursor.nameText = cursor:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	cursor.nameText:SetPoint("CENTER")

	cursor:SetScript("OnShow", function(self)
		local button = self.button
		if not button then
			self:Hide()
			return
		end

		self:SetSize(Healcsi.db.profile.UnitButtons.size.unitWidth, Healcsi.db.profile.UnitButtons.size.unitHeight)
		self:ClearAllPoints()
		self:SetPoint("CENTER", button, "CENTER")
		self.nameText:SetText(button.nameText:GetText())
		self.nameText:SetTextColor(button.nameText:GetTextColor())
		self.nameText:SetFont(button.nameText:GetFont())
	end)

	cursor:SetScript("OnHide", function(self)
		self:Hide()
		self.button = nil
	end)

	cursor:SetScript("OnUpdate", function(self)
		local scale = self:GetEffectiveScale()
		local x, y = GetCursorPosition()
		self:SetPoint("CENTER", UIParent, "BOTTOMLEFT", x / scale, y / scale)
	end)

	frame:SetScript("OnShow", function(self)
		arrowFrame:Hide()

		local i
		if not InCombatLockdown() then
			if Healcsi.db.profile.UnitFrames.raid.showRaidPets then
				HealcsiRaidGroupHeaderPet:Hide()
			end

			for i = 1, 8 do
				UnitFrames:Raid_GetRaidGroup(i):Show()
			end
		end

		local horiz, spacing = UnitFrames:GetLayoutData()
		local width, height

		if horiz then
			width = Healcsi.db.profile.UnitButtons.size.unitWidth * 5 + spacing * 4
			height = Healcsi.db.profile.UnitButtons.size.unitHeight
		else
			width = Healcsi.db.profile.UnitButtons.size.unitWidth
			height = Healcsi.db.profile.UnitButtons.size.unitHeight * 5 + spacing * 4
		end

		for i = 1, 8 do
			groupPanels[i]:SetSize(width, height)
		end

		UnitFrames.Raid_GetFramesMatrix = Hooked_Raid_GetFramesMatrix
		Healcsi:BroadcastEvent("UpdateContainerSize")
	end)

	frame:SetScript("OnHide", function(self)
		if not InCombatLockdown() and Healcsi.db.profile.UnitFrames.raid.showRaidPets then
			HealcsiRaidGroupHeaderPet:Show()
		end

		UnitFrames.Raid_GetFramesMatrix = Orig_Raid_GetFramesMatrix
		Healcsi:BroadcastEvent("UpdateContainerSize")
	end)


end

Healcsi:RegisterEventCallback("OnUnitNotify", function(frame, event, ...)

	for _, frame in ipairs(UnitButtons.unitFrames) do

		local button = frame
		
		if strfind(button:GetName(), "^HealcsiRaidGroupHeaderSubGroup(%d+)UnitButton(%d+)$") then -- HealcsiRaidGroupHeaderSubGroup1UnitButton1
			button._HealcsiSubGroupSwap = 0 -- 
			button:RegisterForDrag("LeftButton")
			button:SetScript("OnDragStart", Button_OnDragStart)
			button:SetScript("OnDragStop", Button_OnDragStop)
		end

	end


end)

--[[

Healcsi:RegisterEventCallback("OnUnitNotify", function(frame, event, ...)
	if (event == "UnitButtonCreated") then
		
		if strfind(button:GetName(), "^HealcsiRaidGroupHeaderSubGroup(%d+)UnitButton(%d+)$") then -- HealcsiRaidGroupHeaderSubGroup1UnitButton1
			button._HealcsiSubGroupSwap = 0 -- 
			button:RegisterForDrag("LeftButton")
			button:SetScript("OnDragStart", Button_OnDragStart)
			button:SetScript("OnDragStop", Button_OnDragStop)
		end

	end
end)

]]