
local Healcsi       = Healcsi
local Console       = Healcsi.Console
local Media         = Healcsi.Media
local Gui           = Healcsi.Gui
local L             = Healcsi.Locale
local Modules       = Healcsi.Modules
local Editor        = Healcsi.Editor
local UnitAuras     = Healcsi.Modules.UnitAuras
local UnitButtons   = Healcsi.Modules.UnitButtons


local type = type
local tinsert = tinsert
local pairs = pairs
local ipairs = ipairs
local wipe = wipe
local strlower = strlower
local GetSpellInfo = GetSpellInfo
local floor = floor
local format = format
local strmatch = strmatch
local tonumber = tonumber
local GetMapContinents = GetMapContinents
local GetCurrentMapContinent = GetCurrentMapContinent
local tinsert = tinsert
local GetInstanceInfo = GetInstanceInfo
local GetRealZoneText = GetRealZoneText
local EJ_SelectInstance = EJ_SelectInstance
local EJ_GetEncounterInfoByIndex = EJ_GetEncounterInfoByIndex
local EJ_GetInstanceByIndex = EJ_GetInstanceByIndex
local EJ_GetNumTiers = EJ_GetNumTiers
local EJ_GetTierInfo = EJ_GetTierInfo
local EJ_GetCurrentTier = EJ_GetCurrentTier

local GetTime = GetTime
local CreateFrame = CreateFrame
local DebuffTypeColor = DebuffTypeColor


tierList = {}
continentList = {}
pendingList = {}

local activeDebuffs = nil
local debuffHighlightInitDone = nil

local function GetInstance(tierId, instanceId)
	local tier = tierList[tierId]
	if tier then
		return tier.instances[instanceId]
	end
end

function UnitAuras:GetNumTiers()
	return  #tierList
end

function UnitAuras:GetTierName(tierId)
	local tier = tierList[tierId]
	if tier then
		return tier.name
	end
end

function UnitAuras:GetCurrentTier()
	return EJ_GetCurrentTier()
end

function UnitAuras:RegisterDebuff(tierId, instanceId, bossId, spellId, level, custom)
	if debuffHighlightInitDone then
		local instance = GetInstance(tierId, instanceId)
		if not instance then
			return
		end

		local debuff = instance:RegisterDebuff(bossId, spellId, level, custom)
		if not debuff then return end

		local db = Healcsi.db.profile.UnitAuras.debuffHighlight
		if custom and db then
			db.customDebuffs[format("%d,%d,%d", tierId, instanceId, spellId)] = format("%d,%d", debuff.bossId, debuff.level)
		end

		return debuff
	else
		tinsert(pendingList, { tierId = tierId, instanceId = instanceId, bossId = bossId, spellId = spellId, level = level, custom = custom })
	end
end

function UnitAuras:EnumInstances(tierId, instanceType)
	local tier = tierList[tierId]
	if not tier then return end

	if type(instanceType) == "string" then
		instanceType = strlower(instanceType)
	end

	if instanceType ~= "raid" and instanceType ~= "party" then
		instanceType = nil
	end

	local list = {}

	local id, data
	for id, data in pairs(tier.instances) do
		if not instanceType or instanceType == data.type then
			tinsert(list, {id = id, name = data.name, debuffCount = data.debuffCount})
		end
	end

	return list
end

function UnitAuras:EnumBosses(tierId, instanceId)
	local instance = GetInstance(tierId, instanceId)
	if not instance then return end

	local list = {}

	local i
	for i = 1, #instance.bosses do
		local data = instance.bosses[i]
		tinsert(list, {id = data.id, name = data.name, debuffCount = data.debuffCount})
	end
	return list
end


function UnitAuras:EnumDebuffs(tierId, instanceId, bossId)
	local instance = GetInstance(tierId, instanceId)
	if not instance then return end

	if not bossId then
		bossId = 0
	end

	local list = {}

	local name, data
	for name, data in pairs(instance.debuffs) do
		if data.bossId == bossId then
			tinsert(list, {name = name, data = data})
		end
	end
	return list
end

function UnitAuras:IsDebuffRegistered(tierId, instanceId, debuffName)
	local instance = GetInstance(tierId, instanceId)
	if instance then
		return instance.debuffs[debuffName]
	end
end

function UnitAuras:GetInstanceName(tierId, instanceId)
	local instance = GetInstance(tierId, instanceId)
	if instance then
		return instance.name
	end
end

local function NormalizeLevel(level, default)
	if type(level) == "number" then
		level = floor(level)
		if level < 0 then
			level = 0
		end

		if level > 5 then
			level = 5
		end
	else
		level = default
	end

	return level
end

function UnitAuras:SetDebuffLevel(tierId, instanceId, spellId, level)
	if type(spellId) ~= "number" then return end

	local instance = GetInstance(tierId, instanceId)
	if not instance then return end

	local name = GetSpellInfo(spellId)
	if not name then return end

	local debuff = instance.debuffs[name]
	if not debuff then return end

	level = NormalizeLevel(level, debuff.defLevel)
	debuff.level = level

	local key = format("%d,%d,%d", tierId, instanceId, spellId)
	local db = Healcsi.db.profile.UnitAuras.debuffHighlight

	if debuff.custom then
		db.customDebuffs[key] = format("%d,%d", debuff.bossId, level)
	else
		if debuff.level == debuff.defLevel then
			db.userLevels[key] = nil
		else
			db.userLevels[key] = level
		end
	end

	return level
end

local function InitDebuffLevel(tierId, instanceId, spellId, level)
	local instance = GetInstance(tierId, instanceId)
	if not instance then return end

	level = NormalizeLevel(level)
	if not level then return end

	local debuff = instance.debuffs[GetSpellInfo(spellId)]
	if debuff then
		debuff.level = level
		return level
	end
end

local function ParseValues(value)
	if type(value) == "string" then
		local id1, id2, id3 = strmatch(value, "(%d+),(%d+),(%d+)")
		if id1 and id2 and id3 then
			return tonumber(id1), tonumber(id2), tonumber(id3)
		end
	end
end

local function ParseValues2(value)
	if type(value) == "string" then
		local id1, id2 = strmatch(value, "(%d+),(%d+)")
		if id1 and id2 then
			return tonumber(id1), tonumber(id2)
		end
	end
end

function UnitAuras:ApplyUserLevels()
	-- Apply user-modified priorities
	local db = Healcsi.db.profile.UnitAuras.debuffHighlight
	local key, level
	for key, level in pairs(db.userLevels) do
		local valid = type(key) == "string" and type(level) == "number"
		if valid then
			local tierId, instanceId, spellId = ParseValues(key)
			valid = InitDebuffLevel(tierId, instanceId, spellId, level)
		end

		if not valid then
			db.userLevels[key] = nil
		end
	end
end

function UnitAuras:ClearUserLevels()
	local _, tier, instance
	for _, tier in pairs(tierList) do
		for _, instance in pairs(tier.instances) do
			local debuff
			for _, debuff in pairs(instance.debuffs) do
				debuff.level = debuff.defLevel
			end
		end
	end
end

function UnitAuras:GetZoneDebuffs()

	if (not(Healcsi.db.profile.UnitAuras.debuffHighlight.show)) then return end

	local zone = GetInstanceInfo() or GetRealZoneText()
	if not zone then return end

	local continent = continentList[GetCurrentMapContinent()]

	local _, tier
	for _, tier in pairs(tierList) do
		local id, data
		for id, data in pairs(tier.instances) do
			if data.name == zone or data.name == continent then
				--print("zone debuff: ", data.name)
				return data.debuffs
			end
		end
	end
end

function UnitAuras:DeleteCustomDebuff(tierId, instanceId, spellId)
	local instance = GetInstance(tierId, instanceId)
	if not instance then return end

	local name = GetSpellInfo(spellId)
	if not name then return end

	local debuff = instance.debuffs[name]
	if not debuff or not debuff.custom then return end

	local db = Healcsi.db.profile.UnitAuras.debuffHighlight

	instance.debuffs[name] = nil
	instance.debuffCount = instance.debuffCount - 1
	db.customDebuffs[format("%d,%d,%d", tierId, instanceId, spellId)] = nil
	return debuff
end

function UnitAuras:ApplyCustomDebuffs()
	-- Apply custom debuffs
	local db = Healcsi.db.profile.UnitAuras.debuffHighlight
	local key, value
	for key, value in pairs(db.customDebuffs) do
		local valid = type(key) == "string" and type(value) == "string"
		if valid then
			local tierId, instanceId, spellId = ParseValues(key)
			local bossId, level = ParseValues2(value)
			if tierId and instanceId and spellId and bossId and level then
				valid = self:RegisterDebuff(tierId, instanceId, bossId, spellId, level, 1)
			end
		end

		if not valid then
			db.customDebuffs[key] = nil
		end
	end
end

function UnitAuras:ClearCustomDebuffs()
	local _, tier, instance
	for _, tier in pairs(tierList) do
		for _, instance in pairs(tier.instances) do
			local name, debuff
			for name, debuff in pairs(instance.debuffs) do
				if debuff.custom then
					instance.debuffs[name] = nil
				end
			end
		end
	end
end

------------------------------------------------
-- Internal Utilities
------------------------------------------------

local function FindBoss(bosses, id)
	local general
	local _, data
	for _, data in ipairs(bosses) do
		if data.id == id then
			return data
		end

		if data.id == 0 then
			general = data
		end
	end

	return general, 1
end

local function BuildBossList(instanceId)
	EJ_SelectInstance(instanceId)
	local list = {}
	local i
	for i = 1, 255 do
		local name, _, id = EJ_GetEncounterInfoByIndex(i)
		if not name or not id then break end

		tinsert(list, { id = id, name = name, debuffCount = 0 })
	end

	tinsert(list, { id = 0, name = GENERAL, debuffCount = 0 })
	return list
end

local function Instance_RegisterDebuff(self, bossId, spellId, level, custom)
	local name, _, icon = GetSpellInfo(spellId)
	if not name then return end

	level = NormalizeLevel(level, 2)

	local boss, general = FindBoss(self.bosses, bossId)
	if general then
		bossId = 0
	end

	local debuff = self.debuffs[name]
	if debuff then return end

	debuff = {}
	self.debuffs[name] = debuff
	self.debuffCount = self.debuffCount + 1
	boss.debuffCount = boss.debuffCount + 1

	debuff.id = spellId
	debuff.name = name
	debuff.icon = icon
	debuff.link = "|cff71d5ff|Hspell:"..spellId.."|h["..name.."]|h|r"
	debuff.level = level
	debuff.defLevel = level
	debuff.bossId = bossId
	debuff.custom = custom

	return debuff
end

local function BuildInstanceList(tier, instanceType, list)
	local arg = instanceType == "raid"
	EJ_SelectTier(tier)
	local i
	for i = 1, 255 do
		local id, name = EJ_GetInstanceByIndex(i, arg)
		if not id or not name then break end

		local data = { id = id, name = name, type = instanceType, debuffCount = 0, debuffs = {} }
		data.RegisterDebuff = Instance_RegisterDebuff
		list[id] = data
		data.bosses = BuildBossList(id)
	end
end

function UnitAuras:FindTopDebuff(unit)
	if not unit or not activeDebuffs then return end

	local i, data, maxLevel, maxName, maxIcon, maxCount, maxDisType, maxExpires
	for i = 1, 20 do
		local name, _, icon, count, dispelType, _, expires = UnitDebuff(unit, i)
		if not name then break end

		data = activeDebuffs[name]
		if data then
			if name == maxName and count > maxCount then
				maxCount = count
			end

			local level = data.level
			if level > (maxLevel or 0) then
				maxLevel, maxName, maxIcon, maxCount, maxDisType, maxExpires = level, name, icon, count, dispelType, expires
			end
		end
	end

	return maxLevel, maxIcon, maxCount, maxDisType, maxExpires
end


local function RaidDebuffFrame_OnUpdate(self, elapsed)
	self.elapsed = (self.elapsed or 0) + elapsed
	if self.elapsed > 0.2 then

		--Healcsi:Print("RaidDebuffFrame_OnUpdate")

		self.elapsed = 0
		local timeLeft = (self.expires or 0) - GetTime()
		if timeLeft > 0 and timeLeft < 15 then
			self.timeText:SetFormattedText("%d", timeLeft)
		else
			self.timeText:SetText()
		end
	end
end

local function RaidDebuffFrame_UpdateDebuff(self)
	local level, icon, count, dispelType, expires = UnitAuras:FindTopDebuff(self.unit)
	if not level then

		self:SetAlpha(0)
		self:SetScript("OnUpdate", nil)
		return
	end

	local color
	if dispelType and dispelType ~= "none" then
		color = DebuffTypeColor[dispelType]
	end

	if color then
		self.bkgnd:SetVertexColor(color.r, color.g, color.b, 1)
	else
		self.bkgnd:SetVertexColor(0, 0, 0, 0.75)
	end

	self.icon:SetTexture(icon)
	if count and count > 1 then
		self.countText:SetFormattedText("%d", count)
	else
		self.countText:SetText()
	end

	self.expires = expires
	if expires and expires > 0 then
		self:SetScript("OnUpdate", RaidDebuffFrame_OnUpdate)
		RaidDebuffFrame_OnUpdate(self, 1)
	else
		self.timeText:SetText()
		self:SetScript("OnUpdate", nil)
	end

	self:SetAlpha(Healcsi.db.profile.UnitAuras.debuffHighlight.alpha)

end

function UnitAuras:InitDebuffHighLight()
	local i
	local numTiers = EJ_GetNumTiers()
	
	for i = 1, numTiers do
		local tier = {}
		tier.name = EJ_GetTierInfo(i)
		tier.id = i
		tier.instances = {}
		BuildInstanceList(i, "raid", tier.instances)
		BuildInstanceList(i, "party", tier.instances)
		tinsert(tierList, tier)
	end

	-- World bosses do not stay in "instances" since 6.0
	local list = { GetMapContinents() }
	local i
	for i = 1, #list do
		local name = list[i]
		if type(name) == "string" then
			tinsert(continentList, name)
		end
	end

	debuffHighlightInitDone = 1
	local _, data
	for _, data in pairs(pendingList) do
		self:RegisterDebuff(data.tierId, data.instanceId, data.bossId, data.spellId, data.level, data.custom)
	end

	wipe(pendingList)

	self:ApplyCustomDebuffs()
	self:ApplyUserLevels()

	self.eventFrame = CreateFrame("Frame", "DebuffHighLightEventFrame")

	self.eventFrame:RegisterEvent("ZONE_CHANGED")
	self.eventFrame:RegisterEvent("ZONE_CHANGED_INDOORS")
	self.eventFrame:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	self.eventFrame:RegisterEvent("PLAYER_ENTERING_WORLD")

	self.eventFrame:SetScript("OnEvent", function(self, event, ...)
		UnitAuras:UpdateActiveDebuffs()
	end)

end

function UnitAuras:CreateDebuffHighlightVisualFrame(frame)
	local visual = CreateFrame("Frame", frame.visualParent:GetName().."DebuffHighlight", frame.visualParent)
	frame.debuffHighlightVisual = visual
	visual:Hide() -- important! so the OnShow can get called...

	local db = Healcsi.db.profile.UnitAuras.debuffHighlight

	visual:SetSize(20, 20)
	visual:SetScale(db.scale)
	visual:SetPoint(db.anchor, frame.visualParent ,db.anchor, db.offset.x, db.offset.y)

	visual.bkgnd = visual:CreateTexture(nil, "BORDER")
	visual.bkgnd:SetAllPoints(visual)
	visual.bkgnd:SetTexture(Healcsi:GetMedia("background", "Solid"))

	visual.icon = visual:CreateTexture(nil, "ARTWORK")
	visual.icon:SetTexCoord(db.zoom, 1-db.zoom, db.zoom, 1-db.zoom)
	visual.icon:SetPoint("TOPLEFT", 1, -1)
	visual.icon:SetPoint("BOTTOMRIGHT", -1, 1)

	visual.countText = visual:CreateFontString(nil, "OVERLAY", "TextStatusBarText")
	visual.countText:SetFont(STANDARD_TEXT_FONT, 12, "OUTLINE")
	visual.countText:SetPoint("BOTTOMRIGHT", 0, 1)

	visual.timeText = visual:CreateFontString(nil, "OVERLAY", "TextStatusBarText")
	visual.timeText:SetFont(STANDARD_TEXT_FONT, 9, "OUTLINE")
	visual.timeText:SetPoint("TOP", visual, "BOTTOM")

	visual.UpdateDebuff = RaidDebuffFrame_UpdateDebuff

	visual:SetScript("OnEvent", function(self, event, unit)
		if unit and unit == self.unit then
			RaidDebuffFrame_UpdateDebuff(self)
		end
	end)

	visual:SetScript("OnShow", function(self)
		self:RegisterEvent("UNIT_AURA")
		RaidDebuffFrame_UpdateDebuff(self)

	end)

	visual:SetScript("OnHide", function(self)
		self:UnregisterAllEvents()
	end)

	-- activeDebuffs = self:GetZoneDebuffs()

	if activeDebuffs then
		visual:Show()
	end
end

function UnitAuras:DebuffHighLightReset()
	self:ClearUserLevels()
	self:ClearCustomDebuffs()
end

function UnitAuras:UpdateDebuffHighlightVisualFrames()
	local db = Healcsi.db.profile.UnitAuras.debuffHighlight
	for _, frame in ipairs(UnitButtons.unitFrames) do

		if (not(frame.debuffHighlightVisual)) then
			UnitAuras:CreateDebuffHighlightVisualFrame(frame)
		end

		local visual = frame.debuffHighlightVisual

		visual:SetScale(db.scale)
		visual:ClearAllPoints()
		visual:SetPoint(db.anchor, frame.visualParent ,db.anchor, db.offset.x, db.offset.y)

		visual.icon:SetTexCoord(db.zoom, 1-db.zoom, db.zoom, 1-db.zoom)

	end
end

function UnitAuras:UpdateActiveDebuffs()

	activeDebuffs = self:GetZoneDebuffs()

	for _, frame in ipairs(UnitButtons.unitFrames) do
		if (not(frame.debuffHighlightVisual)) then
			UnitAuras:CreateDebuffHighlightVisualFrame(frame)
		end
		if (activeDebuffs) then
			frame.debuffHighlightVisual:Show()
		else
			frame.debuffHighlightVisual:Hide()
		end
	end
end

-- ==============================================================
-- Notify


Healcsi:RegisterEventCallback("OnUnitNotify", function(frame, event)
	if (event == "OnUnitChange" ) then -- or event == "OnUnitFlagChange"
		if (not(frame.debuffHighlightVisual)) then
			UnitAuras:CreateDebuffHighlightVisualFrame(frame)
		end
		local visual = frame.debuffHighlightVisual
		local displayedUnit, unit, inVehicle, unitClass = frame:GetUnitInfo()
		visual.unit = displayedUnit
		visual:UpdateDebuff()
	end
end)

Healcsi:RegisterEventCallback("DebuffHighlightConfigChanged", function(frame, event)
	UnitAuras:UpdateActiveDebuffs()
	UnitAuras:UpdateDebuffHighlightVisualFrames()
end)
