
local Healcsi       = Healcsi
local Console       = Healcsi.Console
local Media         = Healcsi.Media
local Gui           = Healcsi.Gui
local L             = Healcsi.Locale
local Modules       = Healcsi.Modules
local Editor        = Healcsi.Editor
local UnitButtons   = Healcsi.Modules.UnitButtons

local strlen = string.utf8len or strlen
local strsub = string.utf8sub or strsub
local mfloor = math.floor

function UnitButtons:Initialize()

	self.initDone = true
end

function UnitButtons:IsInitDone()
	return self.initDone
end

function UnitButtons:GetUnitSpacing()
	return Healcsi.db.profile.UnitButtons.size.spacing
end

function UnitButtons:GetUnitWidth()
	return Healcsi.db.profile.UnitButtons.size.unitWidth
end

function UnitButtons:GetUnitHeight()
	return Healcsi.db.profile.UnitButtons.size.unitHeight
end








local _G = _G
local type = type
local ipairs = ipairs
local UnitIsConnected = UnitIsConnected
local UnitIsDead = UnitIsDead
local UnitIsPlayer = UnitIsPlayer
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local select = select
local UnitClass = UnitClass
local UnitPlayerControlled = UnitPlayerControlled
local UnitCanAttack = UnitCanAttack
local UnitReaction = UnitReaction
local DebuffTypeColor = DebuffTypeColor
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local wipe = wipe
local UnitGroupRolesAssigned = UnitGroupRolesAssigned
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitGetIncomingHeals = UnitGetIncomingHeals
local UnitGetTotalAbsorbs = UnitGetTotalAbsorbs
local UnitGetTotalHealAbsorbs = UnitGetTotalHealAbsorbs
local UnitHasVehicleUI = UnitHasVehicleUI
local strmatch = strmatch
local UnitPowerType = UnitPowerType
local PowerBarColor = PowerBarColor
local UnitPowerMax = UnitPowerMax
local UnitName = UnitName
local UnitIsUnit = UnitIsUnit
local UnitThreatSituation = UnitThreatSituation
local GetThreatStatusColor = GetThreatStatusColor
local GetReadyCheckStatus = GetReadyCheckStatus
local GetTime = GetTime
local GetRaidTargetIndex = GetRaidTargetIndex
local SetRaidTargetIconTexture = SetRaidTargetIconTexture
local UnitHasIncomingResurrection = UnitHasIncomingResurrection
local UnitInRange = UnitInRange
local UnitIsVisible = UnitIsVisible
local InCombatLockdown = InCombatLockdown
local GameTooltip = GameTooltip
local CreateFrame = CreateFrame
local tinsert = tinsert
local RegisterStateDriver = RegisterStateDriver
local RegisterUnitWatch = RegisterUnitWatch
local strlen = string.utf8len or strlen
local strsub = string.utf8sub or strsub
local GetTexCoordsForRoleSmallCircle = GetTexCoordsForRoleSmallCircle
local GetNumTalents = GetNumTalents
local GetTalentInfo = GetTalentInfo
local UnitGUID = UnitGUID
local tonumber = tonumber
local GetRaidRosterInfo = GetRaidRosterInfo
local UnitIsGroupLeader = UnitIsGroupLeader
local GetLootMethod = GetLootMethod
local UnitIsTapDenied = UnitIsTapDenied
local max = max
local min = min












-- ###

UnitButtons.unitFrames = {} -- Stores all unit frames that were created by this addon

function UnitButtons:NumButtons()
	return #(UnitButtons.unitFrames)
end

function UnitButtons:GetButton(index)
	return UnitButtons.unitFrames[index]
end

local function BroadcastUnitNotify(frame, event, ...)
	if (UnitButtons:IsInitDone()) then
		Healcsi:BroadcastEvent("OnUnitNotify", frame, event, ...)
	end
end

-- ###

-- ==============================================================
-- Pet Owner finder

local petOwnerScanTool = CreateFrame("GameTooltip", "HealcsiPetOwnerScanTooltip", nil, "GameTooltipTemplate")
petOwnerScanTool:SetOwner(WorldFrame, "ANCHOR_NONE")
local petOwnerScanText = _G["HealcsiPetOwnerScanTooltipTextLeft2"] -- This is the line with <[Player]'s Pet>

local function GetPetOwner(petUnit)
	petOwnerScanTool:ClearLines()
	petOwnerScanTool:SetUnit(petUnit)
	local ownerText = petOwnerScanText:GetText()

	if (not(ownerText)) then return nil end

	local owner, _ = string.split("'",ownerText)
	return owner
end

-- ==============================================================
-- Utils

local function GetUnitColor(unit, isName, ignoreOffline)
	-- is unit a nil value?
	if (not(unit)) then return 1, 1, 1 end

	-- is unit offline?
	if (not(isName) and not(UnitIsConnected(unit)) and not(ignoreOffline)) then
		return "offline" --Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix) -- offline
	end

	if (UnitIsDead(unit)) then
		--return Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.dead.fix) -- dead
	end

	if UnitIsPlayer(unit) then
		local color = RAID_CLASS_COLORS[select(2, UnitClass(unit))]
		return color.r or 0, color.g or 0, color.b or 0
	elseif (not(isName) and (dead or UnitIsTapDenied(unit))) then
		return 0.5, 0.5, 0.5 -- WTF is this?
	elseif UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then
			return Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.hostile.fix) -- hostile
		elseif UnitCanAttack("player", unit) then
			return Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.neutral.fix) -- neutral
		else
			return "pet" -- pet is handled differently!
		end
	else
		local reaction = UnitReaction(unit, "player") -- Returns the reaction of one unit with regards to another as a number
		--[[
			1 - Hated
			2 - Hostile
			3 - Unfriendly
			4 - Neutral
			5 - Friendly
			6 - Honored
			7 - Revered
			8 - Exalted
		]]
		if (not(reaction)) then
			return 0.5, 0.5, 1 -- should never get here? i don't see a nil return value defined for function UnitReaction :/
		elseif (reaction < 4) then
			return Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.hostile.fix) -- below neutral
		elseif (reaction > 4) then
			return Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.friendly.fix) -- above neutral
		else
			return Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.neutral.fix) -- neutral
		end
	end
end

local function GetCalculatedColor(self, unitDb, petDb, vehicleDb, ignoreOffline, ignoreVehicle)
	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local r, g, b, a

	if (self.inVehicle and not(ignoreVehicle)) then
		if (vehicleDb.color.mode == "fix") then
			r, g, b, a = vehicleDb.color.fix.r, vehicleDb.color.fix.g, vehicleDb.color.fix.b, vehicleDb.color.fix.a
		elseif (vehicleDb.color.mode == "health") then
			local healthMax = UnitHealthMax(self.displayedUnit)
			local health = UnitHealth(self.displayedUnit)
			local healthPercent = health / healthMax

			r = math.min(1, (1-healthPercent) * 2.5) -- the trick is use here is that i multiple by 2
			g = math.min(1, healthPercent * 2)     -- and then call math.min that forces the max value to 1
			b = 0
			a = alpha

			if (unitDb.color.shadeMode == "hsvShade") then
				r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
			elseif (unitDb.color.shadeMode == "rgbShade") then
				r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.rgb)
			end
		elseif (vehicleDb.color.mode == "class") then
			if (GetUnitColor(unit, nil, ignoreOffline) == "pet") then
				-- pet section begins
				local petOwner = GetPetOwner(unit)
				r, g, b, a = GetUnitColor(petOwner, nil, ignoreOffline)
				a = petDb.color.alpha

				if (petDb.color.shadeMode == "hsvShade") then
					r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, petDb.color.hsv)
				elseif (petDb.color.shadeMode == "rgbShade") then
					r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, petDb.color.rgb)
				end
				-- pet section ends
			elseif (GetUnitColor(unit, nil, ignoreOffline) == "offline") then
				r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix)
			else
				r, g, b, a = GetUnitColor(unit, nil, ignoreOffline)
				a = alpha

				if (unitDb.color.shadeMode == "hsvShade") then
					r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
				elseif (unitDb.color.shadeMode == "rgbShade") then
					r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.rgb)
				end
			end
		end
	elseif (GetUnitColor(unit, nil, ignoreOffline) == "pet") then
		-- pet section begins
		if (petDb.color.mode == "fix") then
			r, g, b, a = Healcsi:ConvColorObj(petDb.color.fix)
		else
			if (petDb.color.mode == "class") then
				local petOwner = GetPetOwner(unit)
				if (GetUnitColor(petOwner, nil, ignoreOffline) == "offline") then
					r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix)
				else
					r, g, b, a = GetUnitColor(petOwner, nil, ignoreOffline)
				end
			elseif (petDb.color.mode == "health") then
				local healthMax = UnitHealthMax(self.displayedUnit)
				local health = UnitHealth(self.displayedUnit)
				local healthPercent = health / healthMax

				r = math.min(1, (1-healthPercent) * 2.5) -- the trick is use here is that i multiple by 2
				g = math.min(1, healthPercent * 2)     -- and then call math.min that forces the max value to 1
				b = 0
			end

			a = petDb.color.alpha

			if (petDb.color.shadeMode == "hsvShade") then
				r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, petDb.color.hsv)
			elseif (petDb.color.shadeMode == "rgbShade") then
				r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, petDb.color.rgb)
			end
		end
		-- pet section ends
	elseif (GetUnitColor(unit, nil, ignoreOffline) == "offline") then
		r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix)
	elseif (unitDb.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(unitDb.color.fix)
	else
		if (unitDb.color.mode == "class") then
			r, g, b, a = GetUnitColor(unit, nil, ignoreOffline)
		elseif (unitDb.color.mode == "health") then
			local healthMax = UnitHealthMax(self.displayedUnit)
			local health = UnitHealth(self.displayedUnit)
			local healthPercent = health / healthMax

			r = math.min(1, (1-healthPercent) * 2.5) -- the trick is use here is that i multiple by 2
			g = math.min(1, healthPercent * 2)     -- and then call math.min that forces the max value to 1
			b = 0
		end

		a = unitDb.color.alpha

		if (unitDb.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
		elseif (unitDb.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.gb)
		end
	end

	return r, g, b, a
end

local function GetCalculatedClassColor(self, unitDb, petDb, vehicleDb, ignoreOffline, ignoreVehicle)
	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local r, g, b, a

	if (self.inVehicle and not(ignoreVehicle)) then
		if (vehicleDb.color.mode == "fix") then
			r, g, b, a = vehicleDb.color.fix.r, vehicleDb.color.fix.g, vehicleDb.color.fix.b, vehicleDb.color.fix.a
		elseif (vehicleDb.color.mode == "health") then
			local healthMax = UnitHealthMax(self.displayedUnit)
			local health = UnitHealth(self.displayedUnit)
			local healthPercent = health / healthMax

			r = math.min(1, (1-healthPercent) * 2.5) -- the trick is use here is that i multiple by 2
			g = math.min(1, healthPercent * 2)     -- and then call math.min that forces the max value to 1
			b = 0
			a = alpha

			if (unitDb.color.shadeMode == "hsvShade") then
				r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
			elseif (unitDb.color.shadeMode == "rgbShade") then
				r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.rgb)
			end
		elseif (vehicleDb.color.mode == "class") then
			if (GetUnitColor(unit, nil, ignoreOffline) == "pet") then
				-- pet section begins
				local petOwner = GetPetOwner(unit)
				r, g, b, a = GetUnitColor(petOwner, nil, ignoreOffline)
				a = petDb.color.alpha

				if (petDb.color.shadeMode == "hsvShade") then
					r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, petDb.color.hsv)
				elseif (petDb.color.shadeMode == "rgbShade") then
					r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, petDb.color.rgb)
				end
				-- pet section ends
			elseif (GetUnitColor(unit, nil, ignoreOffline) == "offline") then
				r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix)
			else
				r, g, b, a = GetUnitColor(unit, nil, ignoreOffline)
				a = alpha

				if (unitDb.color.shadeMode == "hsvShade") then
					r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
				elseif (unitDb.color.shadeMode == "rgbShade") then
					r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.rgb)
				end
			end
		end
	elseif (GetUnitColor(unit, nil, ignoreOffline) == "pet") then
		-- pet section begins
		local petOwner = GetPetOwner(unit)
		if (GetUnitColor(petOwner, nil, ignoreOffline) == "offline")then
			r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix)
		else
			r, g, b, a = GetUnitColor(petOwner, nil, ignoreOffline)
		end
		a = petDb.color.alpha

		if (petDb.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, petDb.color.hsv)
		elseif (petDb.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, petDb.color.rgb)
		end
		-- pet section ends
	elseif (GetUnitColor(unit, nil, ignoreOffline) == "offline") then
		r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix)
	else
		r, g, b, a = GetUnitColor(unit, nil, ignoreOffline)
		a = alpha

		if (unitDb.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
		elseif (unitDb.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.rgb)
		end
	end

	return r, g, b, a
end

local function GetCalculatedHealthColor(self, unitDb, petDb, vehicleDb)
	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local r, g, b, a

	if (self.inVehicle) then
		if (vehicleDb.mode == "fix") then
			r, g, b, a = vehicleDb.fix.r, vehicleDb.fix.g, vehicleDb.fix.b, vehicleDb.fix.a
		elseif (vehicleDb.mode == "health") then
			local healthMax = UnitHealthMax(self.displayedUnit)
			local health = UnitHealth(self.displayedUnit)
			local healthPercent = health / healthMax

			r = math.min(1, (1-healthPercent) * 2.5) -- the trick is use here is that i multiple by 2
			g = math.min(1, healthPercent * 2)     -- and then call math.min that forces the max value to 1
			b = 0
			a = alpha

			if (unitDb.color.shadeMode == "hsvShade") then
				r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
			elseif (unitDb.color.shadeMode == "rgbShade") then
				r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.rgb)
			end
		elseif (vehicleDb.mode == "class") then
			if (GetUnitColor(unit) == "pet") then
				-- pet section begins
				local petOwner = GetPetOwner(unit)
				r, g, b, a = GetUnitColor(petOwner)
				a = petDb.color.alpha

				if (petDb.color.shadeMode == "hsvShade") then
					r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, petDb.color.hsv)
				elseif (petDb.color.shadeMode == "rgbShade") then
					r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, petDb.color.rgb)
				end
				-- pet section ends
			else
				r, g, b, a = GetUnitColor(unit)
				a = alpha

				if (unitDb.color.shadeMode == "hsvShade") then
					r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
				elseif (unitDb.color.shadeMode == "rgbShade") then
					r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.rgb)
				end
			end
		end
	--elseif (GetUnitColor(unit) == "pet") then pet has health, so no need to go extra rounds...
	else
		
		local healthMax = UnitHealthMax(self.displayedUnit)
		local health = UnitHealth(self.displayedUnit)
		local healthPercent = health / healthMax

		r = math.min(1, (1-healthPercent) * 2.5) -- the trick is use here is that i multiple by 2
		g = math.min(1, healthPercent * 2)     -- and then call math.min that forces the max value to 1
		b = 0
		a = alpha

		if (unitDb.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, unitDb.color.hsv)
		elseif (unitDb.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, unitDb.color.rgb)
		end
	end

	return r, g, b, a
end

-- ==============================================================
-- Updates

local function UnitButton_UpdateSize(self)
	if (Editor.InCombat) then return end

	local db = Healcsi.db.profile.UnitButtons.size

	self:SetWidth(db.unitWidth)
	self:SetHeight(db.unitHeight)

	BroadcastUnitNotify(self, "OnUnitSizeChange", db.unitWidth, db.unitHeight)
end

local function UnitButton_UpdateMedia(self)
	-- Health
	self.healthBar:SetStatusBarTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.health.file))
	self.healthBarUnderlay:SetTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.health.underlay.file))

	-- Power
	self.powerBar:SetStatusBarTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.power.file))
	self.powerBarUnderlay:SetTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.power.underlay.file))

	-- shieldBar
	self.shieldBar:SetStatusBarTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.shield.file))

	-- shieldBar border
	self.shieldBorderLeft:SetTexture(Media.borderTextures[Healcsi.db.profile.UnitButtons.shield.border.left.file])
	self.shieldBorderRight:SetTexture(Media.borderTextures[Healcsi.db.profile.UnitButtons.shield.border.right.file])
	self.shieldBorderTop:SetTexture(Media.borderTextures[Healcsi.db.profile.UnitButtons.shield.border.top.file])
	self.shieldBorderBottom:SetTexture(Media.borderTextures[Healcsi.db.profile.UnitButtons.shield.border.bottom.file])

	-- absorbBar
	self.absorbsBar:SetStatusBarTexture(Media.absorbTextures[Healcsi.db.profile.UnitButtons.absorb.file])

	-- incomingHealBar
	self.incomingHealBar:SetStatusBarTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.incHeal.file))
	self.incomingHeal2AbsorbBar:SetStatusBarTexture(Media.absorbTextures[Healcsi.db.profile.UnitButtons.incHeal.absorb.file])
end

local function UnitButton_UpdateHealthColor(self)
	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local r, g, b, a = GetCalculatedColor(
		self,
		Healcsi.db.profile.UnitButtons.health,
		Healcsi.db.profile.UnitButtons.health.color.pet,
		Healcsi.db.profile.UnitButtons.health.color.inVehicle
	)
	self.healthBar:SetStatusBarColor(r, g, b, a or 1)
end

local function UnitButton_UpdateHealthUnderlayColor(self)
	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local r, g, b, a =  GetCalculatedColor(
		self,
		Healcsi.db.profile.UnitButtons.health.underlay,
		Healcsi.db.profile.UnitButtons.health.underlay.color.pet,
		Healcsi.db.profile.UnitButtons.health.underlay.color.inVehicle
	)

	self.healthBarUnderlay:SetVertexColor(r, g, b, a or 1)
end

local function UnitButton_UpdateShieldColor(self)
	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local r, g, b, a

	local db = Healcsi.db.profile.UnitButtons.shield.color
	if (db.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.fix)
	else
		if (db.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		elseif (db.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		end

		a = db.alpha

		if (db.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.hsv)
		elseif (db.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.rgb)
		end

	end

	self.shieldBar:SetStatusBarColor(r, g, b, a)
end

local function UnitButton_UpdateAbsorbColor(self)
	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local r, g, b, a

	local db = Healcsi.db.profile.UnitButtons.absorb.color
	if (db.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.fix)
	else
		if (db.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		elseif (db.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		end

		a = db.alpha

		if (db.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.hsv)
		elseif (db.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.rgb)
		end
	end

	self.absorbsBar:SetStatusBarColor(r, g, b, a)
end

local function UnitButton_UpdateIncHealColor(self)

	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local r, g, b, a

	local db = Healcsi.db.profile.UnitButtons.incHeal.color
	if (db.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.fix)
	else
		if (db.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		elseif (db.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		end

		a = db.alpha

		if (db.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.hsv)
		elseif (db.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.rgb)
		end
	end

	self.incomingHealBar:SetStatusBarColor(r, g, b, a)

	local db = Healcsi.db.profile.UnitButtons.incHeal.absorb.color
	if (db.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.fix)
	else
		if (db.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		elseif (db.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		end

		a = db.alpha

		if (db.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.hsv)
		elseif (db.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.rgb)
		end
	end

	self.incomingHeal2AbsorbBar:SetStatusBarColor(r, g, b, a)
end

local function UnitButton_UpdateShieldBorderColor(self)
	local unit = self.unit
	local displayedUnit = self.displayedUnit
	if (not(unit) or not(displayedUnit)) then return end

	local shield = 0

	if (displayedUnit) then
		shield = UnitGetTotalAbsorbs(displayedUnit) or 0
	end

	local db = Healcsi.db.profile.UnitButtons.shield.border

	local minWidth = 0
	if(db.left.show) then minWidth = minWidth + db.left.size end
	if(db.right.show) then minWidth = minWidth + db.right.size end

	if (db.left.show and db.left.file ~= "None" and shield > 5) then
		if (Healcsi.db.profile.UnitButtons.shield.showAbsorbBar) then
			self.shieldBorderLeft:SetPoint("TOPLEFT", self.shieldBar, "TOPLEFT", 0, 0)
			self.shieldBorderLeft:SetPoint("BOTTOMRIGHT", self.shieldBar, "BOTTOMLEFT", db.left.size, 0)
		else
			self.shieldBorderLeft:SetPoint("TOPLEFT", self.healthBar, "TOPLEFT", 0, 0)
			self.shieldBorderLeft:SetPoint("BOTTOMRIGHT", self.healthBar, "BOTTOMLEFT", db.left.size, 0)
		end

--[[
		if (false and (minWidth > self.shieldBar:GetStatusBarTexture():GetWidth())e and Healcsi.db.profile.UnitButtons.shield.showAbsorbBar) then
			self.shieldBorderLeft:SetVertexColor(0, 0, 0, 0)
			--
			Healcsi:Print("Unable to show border: " .. minWidth .. " ? > ? " .. self.shieldBar:GetStatusBarTexture():GetWidth() )
			--
		else]]
			local r, g, b, a
			if (db.left.color.mode == "fix") then
				r, g, b, a = Healcsi:ConvColorObj(db.left.color.fix)
			else
				if (db.left.color.mode == "class") then
					r, g, b, a = GetCalculatedClassColor(
						self,
						Healcsi.db.profile.UnitButtons.health,
						Healcsi.db.profile.UnitButtons.health.color.pet,
						Healcsi.db.profile.UnitButtons.health.color.inVehicle
					)
				elseif (db.left.color.mode == "health") then
					r, g, b, a = GetCalculatedHealthColor(
						self,
						Healcsi.db.profile.UnitButtons.health,
						Healcsi.db.profile.UnitButtons.health.color.pet,
						Healcsi.db.profile.UnitButtons.health.color.inVehicle
					)
				end
				a = db.left.color.alpha
				if (db.left.color.shadeMode == "hsvShade") then
					r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.left.color.hsv)
				elseif (db.left.color.shadeMode == "rgbShade") then
					r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.left.color.rgb)
				end
			end
			self.shieldBorderLeft:SetVertexColor(r, g, b, a)
		--end
	else
		self.shieldBorderLeft:SetVertexColor(0, 0, 0, 0)
	end


	if (db.right.show and db.right.file ~= "None" and shield > 5) then
		if (Healcsi.db.profile.UnitButtons.shield.showAbsorbBar) then
			self.shieldBorderRight:SetPoint("TOPRIGHT", self.shieldBar:GetStatusBarTexture(), "TOPRIGHT", 0, 0)
			self.shieldBorderRight:SetPoint("BOTTOMLEFT", self.shieldBar:GetStatusBarTexture(), "BOTTOMRIGHT", -db.right.size, 0)
		else
			self.shieldBorderRight:SetPoint("TOPRIGHT", self.healthBar:GetStatusBarTexture(), "TOPRIGHT", 0, 0)
			self.shieldBorderRight:SetPoint("BOTTOMLEFT", self.healthBar:GetStatusBarTexture(), "BOTTOMRIGHT", -db.right.size, 0)
		end
		
		--[[if (minWidth > self.shieldBar:GetStatusBarTexture():GetWidth() and Healcsi.db.profile.UnitButtons.shield.showAbsorbBar) then
			self.shieldBorderRight:SetVertexColor(0, 0, 0, 0)
			Healcsi:Print("Right does not fit")
		else]]
			local r, g, b, a
			if (db.right.color.mode == "fix") then
				r, g, b, a = Healcsi:ConvColorObj(db.right.color.fix)
			else
				if (db.right.color.mode == "class") then
					r, g, b, a = GetCalculatedClassColor(
						self,
						Healcsi.db.profile.UnitButtons.health,
						Healcsi.db.profile.UnitButtons.health.color.pet,
						Healcsi.db.profile.UnitButtons.health.color.inVehicle
					)
				elseif (db.right.color.mode == "health") then
					r, g, b, a = GetCalculatedHealthColor(
						self,
						Healcsi.db.profile.UnitButtons.health,
						Healcsi.db.profile.UnitButtons.health.color.pet,
						Healcsi.db.profile.UnitButtons.health.color.inVehicle
					)
				end
				a = db.right.color.alpha
				if (db.right.color.shadeMode == "hsvShade") then
					r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.right.color.hsv)
				elseif (db.right.color.shadeMode == "rgbShade") then
					r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.right.color.rgb)
				end
			end
			self.shieldBorderRight:SetVertexColor(r, g, b, a)
		--end
	else
		self.shieldBorderRight:SetVertexColor(0, 0, 0, 0)
	end

	if (db.top.show and db.top.file ~= "None" and shield > 5) then
		if (Healcsi.db.profile.UnitButtons.shield.showAbsorbBar) then
			self.shieldBorderTop:SetPoint("TOPLEFT", self.shieldBar, "TOPLEFT", 0, 0)
			self.shieldBorderTop:SetPoint("BOTTOMRIGHT", self.shieldBar:GetStatusBarTexture(), "TOPRIGHT", 0, -db.top.size)
		else
			self.shieldBorderTop:SetPoint("TOPLEFT", self.healthBar, "TOPLEFT", 0, 0)
			self.shieldBorderTop:SetPoint("BOTTOMRIGHT", self.healthBar:GetStatusBarTexture(), "TOPRIGHT", 0, -db.top.size)
		end

		local r, g, b, a
		if (db.top.color.mode == "fix") then
			r, g, b, a = Healcsi:ConvColorObj(db.top.color.fix)
		else
			if (db.top.color.mode == "class") then
				r, g, b, a = GetCalculatedClassColor(
					self,
					Healcsi.db.profile.UnitButtons.health,
					Healcsi.db.profile.UnitButtons.health.color.pet,
					Healcsi.db.profile.UnitButtons.health.color.inVehicle
				)
			elseif (db.top.color.mode == "health") then
				r, g, b, a = GetCalculatedHealthColor(
					self,
					Healcsi.db.profile.UnitButtons.health,
					Healcsi.db.profile.UnitButtons.health.color.pet,
					Healcsi.db.profile.UnitButtons.health.color.inVehicle
				)
			end
			a = db.top.color.alpha
			if (db.top.color.shadeMode == "hsvShade") then
				r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.top.color.hsv)
			elseif (db.top.color.shadeMode == "rgbShade") then
				r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.top.color.rgb)
			end
		end

		self.shieldBorderTop:SetVertexColor(r, g, b, a)
	else
		self.shieldBorderTop:SetVertexColor(0, 0, 0, 0)
	end

	if (db.bottom.show and db.bottom.file ~= "None" and shield > 5) then
		if (Healcsi.db.profile.UnitButtons.shield.showAbsorbBar) then
			self.shieldBorderBottom:SetPoint("BOTTOMLEFT", self.shieldBar, "BOTTOMLEFT", 0, 0)
			self.shieldBorderBottom:SetPoint("TOPRIGHT", self.shieldBar:GetStatusBarTexture(), "BOTTOMRIGHT", 0, db.bottom.size)
		else
			self.shieldBorderBottom:SetPoint("BOTTOMLEFT", self.healthBar, "BOTTOMLEFT", 0, 0)
			self.shieldBorderBottom:SetPoint("TOPRIGHT", self.healthBar:GetStatusBarTexture(), "BOTTOMRIGHT", 0, db.bottom.size)
		end

		local r, g, b, a
		if (db.bottom.color.mode == "fix") then
			r, g, b, a = Healcsi:ConvColorObj(db.bottom.color.fix)
		else
			if (db.bottom.color.mode == "class") then
				r, g, b, a = GetCalculatedClassColor(
					self,
					Healcsi.db.profile.UnitButtons.health,
					Healcsi.db.profile.UnitButtons.health.color.pet,
					Healcsi.db.profile.UnitButtons.health.color.inVehicle
				)
			elseif (db.bottom.color.mode == "health") then
				r, g, b, a = GetCalculatedHealthColor(
					self,
					Healcsi.db.profile.UnitButtons.health,
					Healcsi.db.profile.UnitButtons.health.color.pet,
					Healcsi.db.profile.UnitButtons.health.color.inVehicle
				)
			end
			a = db.bottom.color.alpha
			if (db.bottom.color.shadeMode == "hsvShade") then
				r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.bottom.color.hsv)
			elseif (db.bottom.color.shadeMode == "rgbShade") then
				r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.bottom.color.rgb)
			end
		end

		self.shieldBorderBottom:SetVertexColor(r, g, b, a)
	else
		self.shieldBorderBottom:SetVertexColor(0, 0, 0, 0)
	end

end

--[[
local shieldBorderUpdater = CreateFrame("Frame", "ShieldBorderUpdaterFrame")
shieldBorderUpdater:SetScript("OnUpdate", function(self, elapsed)
	self.updateElapsed = (self.updateElapsed or 0) + elapsed
	if (self.updateElapsed > 0.6 and self.updateNeeded) then
		
		-- we do backward iteratation so we can remove elements without conflict! and ponies won't die...
		for i=#(shieldBorderUpdater.updateThis),1,-1 do
		    local frame = shieldBorderUpdater.updateThis[i]
		    UnitButton_UpdateShieldBorderColor(frame)
		    table.remove(shieldBorderUpdater.updateThis, i)
		end

		self.updateNeeded = nil
		self.updateElapsed = 0
	end
end)
]]

local function UnitButton_UpdateHealthShieldAbsorbIncomingHeals(self)
	local displayedUnit = self.displayedUnit
	if (not(displayedUnit) or not(UnitExists(displayedUnit))) then return end

	local healthBar  = self.healthBar
	local shieldBar  = self.shieldBar
	local absorbsBar = self.absorbsBar
	local incomingHealBar = self.incomingHealBar
	local incomingHeal2AbsorbBar = self.incomingHeal2AbsorbBar

	local health = UnitHealth(displayedUnit)
	local healthMax = UnitHealthMax(displayedUnit)
	local shield = UnitGetTotalAbsorbs(displayedUnit) or 0
	local absorbs = UnitGetTotalHealAbsorbs(displayedUnit) or 0
	local heal = UnitGetIncomingHeals(displayedUnit) or 0

	--health    = 4000000 --4000000
	--healthMax = 10000000
	--shield    = 5000000

	if (Healcsi.db.profile.UnitButtons.shield.showAbsorbBar) then
		healthBar:SetMinMaxValues(0, healthMax + shield)
	else
		healthBar:SetMinMaxValues(0, healthMax)
	end

	healthBar:SetValue(health)

	if (Healcsi.db.profile.UnitButtons.shield.showAbsorbBar) then
		if (health < healthMax) then
			shieldBar:SetMinMaxValues(0, (healthMax + shield) - health)
			shieldBar:SetValue(shield)
		else
			shieldBar:SetMinMaxValues(0,100)
			shieldBar:SetValue(100)
		end

		-- We need to delay the border updates, cause for some reason the shieldbar doesn't get it's width update immediately
		--shieldBorderUpdater.updateElapsed = 0
		--shieldBorderUpdater.updateNeeded = 1
		--shieldBorderUpdater.updateThis = shieldBorderUpdater.updateThis or {}
		--tinsert(shieldBorderUpdater.updateThis, self)
	else
		shieldBar:SetMinMaxValues(0,100)
		shieldBar:SetValue(0)
	end

	if absorbs > 0 then
		absorbsBar:SetMinMaxValues(0, healthMax)
		absorbsBar:SetValue(math.min(absorbs, healthMax))
		absorbsBar:Show()
	else
		absorbsBar:Hide()
	end

	if (absorbs > 0) then
		incomingHeal2AbsorbBar:Show()
		if (absorbs > heal) then
			incomingHeal2AbsorbBar:SetMinMaxValues(0, absorbs)
			incomingHeal2AbsorbBar:SetValue(math.min(heal, absorbs))
			heal = 0
		else
			incomingHeal2AbsorbBar:SetMinMaxValues(0, 100)
			incomingHeal2AbsorbBar:SetValue(100)
			heal = heal - absorbs
		end
	else
		incomingHeal2AbsorbBar:Hide()
	end

	if ((health < healthMax) and (heal > 0)) then
		incomingHealBar:SetMinMaxValues(0, healthMax - health)
		incomingHealBar:SetValue(heal)
		incomingHealBar:Show()
	else
		incomingHealBar:Hide()
	end

end

local function UnitButton_UpdateVehicleStatus(self)
	local unit = self.unit
	if (not(unit)) then return end

	if UnitHasVehicleUI(unit) then
		self.inVehicle = 1
		if unit == "player" then
			self.displayedUnit = "vehicle"
		else
			local prefix, id, suffix = strmatch(unit, "([^%d]+)([%d]*)(.*)")
			self.displayedUnit = prefix.."pet"..id..suffix
		end
	else
		self.inVehicle = nil
		self.displayedUnit = unit
	end
end

local function UnitButton_UpdateInRange(self)
	local unit = self.displayedUnit
	if (not(unit)) then return end

	local inRange, checked = UnitInRange(unit)
	if (not(checked)) then
		inRange = UnitIsVisible(unit)
	end

	inRange = inRange and 1 or nil
	if (not(UnitIsConnected(unit))) then
		inRange = 1 -- we don't want to apply outrange for offline players, they are simply... offline xD
	end

	self.artFrame:SetAlpha(inRange and 1 or Healcsi.db.profile.UnitButtons.size.outrangeOpacity)

	if inRange ~= self.inRange then
		self.inRange = inRange
		BroadcastUnitNotify(self, "OnRangeChange", inRange)
	end

end

local function UnitButton_UpdateResurrect(self)
	local unit = self.unit
	if (not(unit)) then return end

	local icon = self.resIcon
	local db = Healcsi.db.profile.UnitButtons.icons.resurrect

	if (UnitIsDeadOrGhost(unit) and UnitHasIncomingResurrection(unit)) then

		icon:ClearAllPoints()
		icon:SetPoint(db.anchor, self.layerFrame, db.anchor, db.offset.x, db.offset.y)
		icon:SetSize(db.size, db.size)
		icon:SetVertexColor(1, 1, 1, db.alpha)
		icon:SetTexCoord(db.zoom, 1-db.zoom, db.zoom, 1-db.zoom) 

		icon:Show()
		self.flagIcon:SetAlpha(0)
	else
		icon:Hide()
		self.flagIcon:SetAlpha(1)
	end
end

local function UnitButton_UpdatePowerColor(self)
	local unit = self.displayedUnit
	if (not(unit)) then return end

	local r, g, b, a

	if (not(UnitIsConnected(unit))) then
		r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix)
	elseif (Healcsi.db.profile.UnitButtons.power.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.power.color.fix)
	else
		a = Healcsi.db.profile.UnitButtons.power.color.alpha
		local powerType, powerToken, altR, altG, altB = UnitPowerType(unit)
		local prefix = _G[powerToken]
		local info = PowerBarColor[powerToken]
		if info then
			r, g, b = info.r, info.g, info.b
		else
			if not altR then
				info = PowerBarColor[powerType] or PowerBarColor["MANA"]
				r, g, b = info.r, info.g, info.b
			else
				r, g, b = altR, altG, altB
			end
		end

		if (Healcsi.db.profile.UnitButtons.power.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, Healcsi.db.profile.UnitButtons.power.color.hsv)
		elseif (Healcsi.db.profile.UnitButtons.power.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, Healcsi.db.profile.UnitButtons.power.color.rgb)
		end
	end

	self.powerBar:SetStatusBarColor(r, g, b, a)
end

local function UnitButton_UpdatePowerUnderlayColor(self)
	local unit = self.displayedUnit
	if (not(unit)) then return end

	local r, g, b, a

	if (not(UnitIsConnected(unit))) then
		r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.health.color.offline.fix)
	elseif (Healcsi.db.profile.UnitButtons.power.underlay.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.power.underlay.color.fix)
	else
		a = Healcsi.db.profile.UnitButtons.power.underlay.color.alpha
		local powerType, powerToken, altR, altG, altB = UnitPowerType(unit)
		local prefix = _G[powerToken]
		local info = PowerBarColor[powerToken]
		if info then
			r, g, b = info.r, info.g, info.b
		else
			if not altR then
				info = PowerBarColor[powerType] or PowerBarColor["MANA"]
				r, g, b = info.r, info.g, info.b
			else
				r, g, b = altR, altG, altB
			end
		end

		if (Healcsi.db.profile.UnitButtons.power.underlay.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, Healcsi.db.profile.UnitButtons.power.underlay.color.hsv)
		elseif (Healcsi.db.profile.UnitButtons.power.underlay.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, Healcsi.db.profile.UnitButtons.power.underlay.color.rgb)
		end
	end

	self.powerBarUnderlay:SetVertexColor(r, g, b, a)
end

local function UnitButton_UpdatePowerMax(self)
	local unit = self.displayedUnit
	if (not(unit)) then return end

	local value = UnitPowerMax(unit)

	if value > 0 then
		self.powerBar:SetMinMaxValues(0, value)
		self.powerBar:Show()
		self.powerBarUnderlay:Show()

		if (GetUnitColor(unit) == "pet") then
			self.healthBar:ClearAllPoints()
			self.healthBar:SetPoint("TOPLEFT", 1, -1)
			self.healthBar:SetPoint("BOTTOMRIGHT", -1, Healcsi.db.profile.UnitButtons.size.petPowerHeight + 2)
		else
			self.healthBar:ClearAllPoints()
			self.healthBar:SetPoint("TOPLEFT", 1, -1)
			self.healthBar:SetPoint("BOTTOMRIGHT", -1, Healcsi.db.profile.UnitButtons.size.powerHeight + 2)
		end
	else
		self.powerBar:Hide()
		self.powerBarUnderlay:Hide()

		self.healthBar:ClearAllPoints()
		self.healthBar:SetPoint("TOPLEFT", 1, -1)
		self.healthBar:SetPoint("BOTTOMRIGHT", -1, 2)
	end
end

local function UnitButton_UpdatePower(self)
	local unit = self.displayedUnit
	if (not(unit)) then return end

	self.powerBar:SetValue(UnitPower(unit))
end

local function UnitButton_UpdateNameText(self)
	local unit = self.unit
	if (not(unit)) then return end
	
	local db = Healcsi.db.profile.UnitButtons.nameText
	local nameText = self.nameText

	self.unitClass = self.unit and select(2, UnitClass(self.unit))
	BroadcastUnitNotify(self, "OnUnitChange", self.displayedUnit, self.unit, self.inVehicle, self.unitClass)

	nameText:ClearAllPoints()
	nameText:SetPoint(db.anchor, self.healthBar, db.anchor, db.offset.x, db.offset.y)

	if (db.font.outline) then
		nameText:SetFont(Healcsi:GetMedia("font", db.font.file), db.font.size, "OUTLINE")
	else
		nameText:SetFont(Healcsi:GetMedia("font", db.font.file), db.font.size, nil)
	end

	local r, g, b, a
	if (db.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.color.fix)
	else
		if (db.color.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle,
				true, true
			)
		elseif (db.color.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle,
				true, true
			)
		end
		a = db.color.alpha
		if (db.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.color.hsv)
		elseif (db.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.color.rgb)
		end
	end
	nameText:SetTextColor(r, g, b, a)

	db = Healcsi.db.profile.UnitButtons.nameText.font.shadow

	if (db.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.color.fix)
	else
		if (db.color.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle,
				true
			)
		elseif (db.color.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle,
				true
			)
		end
		a = db.color.alpha
		if (db.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.color.hsv)
		elseif (db.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.color.rgb)
		end
	end
	nameText:SetShadowColor(r, g, b, a) 
	nameText:SetShadowOffset(db.offset.x, db.offset.y)

	local text = UnitName(unit)

	if (not(text)) then
		nameText:SetText()
		return
	end

	local buttonWidth = self:GetWidth()
	if buttonWidth < 1 then return end

	local lengthLimit = db.lengthLimit
	if type(widthLimit) ~= "number" then lengthLimit = 0.75 end

	local i
	for i = strlen(text), 0, -1 do
		nameText:SetText(strsub(text, 1, i))
		if nameText:GetWidth() / buttonWidth <= lengthLimit then
			break
		end
	end

end

local function UnitButton_UpdateHealthText(self)
	local unit = self.unit
	local displayedUnit = self.displayedUnit

	if (not(unit) or not(displayedUnit) or not(UnitExists(displayedUnit))) then return end

	local db = Healcsi.db.profile.UnitButtons.healthText
	local healthText = self.healthText

	healthText:ClearAllPoints()
	healthText:SetPoint(db.anchor, self.healthBar, db.anchor, db.offset.x, db.offset.y)

	if (db.font.outline) then
		healthText:SetFont(Healcsi:GetMedia("font", db.font.file), db.font.size, "OUTLINE")
	else
		healthText:SetFont(Healcsi:GetMedia("font", db.font.file), db.font.size, nil)
	end

--[[
	if (not(db.show)) then
		healthText:Hide()
		return
	else
		healthText:Show()
	end
]]

	local r, g, b, a
	if (db.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.color.fix)
	else
		if (db.color.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		elseif (db.color.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		end
		a = db.color.alpha
		if (db.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.color.hsv)
		elseif (db.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.color.rgb)
		end
	end
	healthText:SetTextColor(r, g, b, a)

	db = Healcsi.db.profile.UnitButtons.healthText.font.shadow

	if (db.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.color.fix)
	else
		if (db.color.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		elseif (db.color.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		end
		a = db.color.alpha
		if (db.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.color.hsv)
		elseif (db.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.color.rgb)
		end
	end
	healthText:SetShadowColor(r, g, b, a) 
	healthText:SetShadowOffset(db.offset.x, db.offset.y)

	db = Healcsi.db.profile.UnitButtons.healthText

	local health = UnitHealth(displayedUnit)
	local healthMax = UnitHealthMax(displayedUnit)
	local shield = UnitGetTotalAbsorbs(displayedUnit) or 0

	local mode    = db.mode
	local display = db.display
	local value   = db.value

	local text = ""
	if (mode == "percent") then
		if (value == "healthOnly") then
			text = mfloor(health/healthMax) * 100
		else
			text = mfloor((health+shield)/healthMax) * 100
		end
		text = text .. "%"
	elseif (mode == "missing") then
		text = mfloor(healthMax - health)
		if (display == "sH") then text = Healcsi:NumToShort(text) end
	else -- remaining
		if (value == "healthOnly") then
			text = mfloor(health)
			if (display == "sH") then text = Healcsi:NumToShort(text) end
		elseif (value == "healthPlusShield") then
			local textH = ""
			local textS = ""
			textH = mfloor(health)
			textS = mfloor(shield)
			if (display == "sHsS") then
				textH = Healcsi:NumToShort(textH)
				textS = Healcsi:NumToShort(textS)
			elseif (display == "sHfS") then
				textH = Healcsi:NumToShort(textH)
			elseif (display == "fHsS") then
				textS = Healcsi:NumToShort(textS)
			end
			if (textS ~= "0") then
				text = textH .. " + " .. textS
			else
				text = textH
			end
		else
			text = health + shield
			if (display == "sH") then text = Healcsi:NumToShort(text) end
		end
	end
	healthText:SetText(text)
end

local function UnitButton_UpdateThreat(self)
	local unit = self.displayedUnit
	if (not(unit) or not(UnitExists(unit))) then return end

	local status = UnitThreatSituation(unit)
	--[[
		nil = unit is not on any other unit's threat table.
		0 = not tanking anything.
		1 = not tanking anything, but have higher threat than tank on at least one unit.
		2 = insecurely tanking at least one unit, but not securely tanking anything.
		3 = securely tanking at least one unit.

		0 = "low on threat"
		1 = "overnuking"
		2 = "losing threat"
		3 = "tanking securely"
	]]

	if status and status >= 1 then
		local db = Healcsi.db.profile.UnitButtons.threat

		self.aggroHighlight:SetTexture(Healcsi.Media.aggroTextures[db.file])

		if (db.color.default.use) then
			local alpha = 1
			if (status == 1) then
				alpha = db.color.default.alpha.overnuking
			elseif (status == 2) then
				alpha = db.color.default.alpha.losing
			elseif (status == 3) then
				alpha = db.color.default.alpha.tanking
			end
			self.aggroHighlight:SetVertexColor(GetThreatStatusColor(status), alpha)
		else
			if (status == 1) then
				self.aggroHighlight:SetVertexColor(Healcsi:ConvColorObj(db.color.overnuking))
			elseif (status == 2) then
				self.aggroHighlight:SetVertexColor(Healcsi:ConvColorObj(db.color.losing))
			elseif (status == 3) then
				self.aggroHighlight:SetVertexColor(Healcsi:ConvColorObj(db.color.tanking))
			end
		end
		
		self.aggroHighlight:Show()
	else
		self.aggroHighlight:Hide()
	end

end

local function UnitButton_UpdateTarget(self)
	local unit = self.displayedUnit
	if (not(unit)) then return end

	local db = Healcsi.db.profile.UnitButtons.selection

	self.selectionHighlight:SetTexture(Healcsi.Media.selectionTextures[db.file])
	self.selectionHighlight:SetVertexColor(Healcsi:ConvColorObj(db.color))

	if UnitIsUnit(unit, "target") then
		self.selectionHighlight:Show()
	else
		self.selectionHighlight:Hide()
	end
end

local GHOST_AURA, _, GHOST_TEXTURE = GetSpellInfo(8326)
local DEATH_TEXTURE = "Interface\\TargetingFrame\\UI-TargetingFrame-Skull"
local SPIRIT_OF_REDEMPTION = GetSpellInfo(27827)
local SPIRIT_TEXTURE = "Interface\\Icons\\Spell_Holy_GuardianSpirit"
local CAUTERIZE_AURA, _, CAUTERIZE_TEXTURE = GetSpellInfo(87023) -- mage
local PURGATORY_AURA, _, PURGATORY_TEXTURE = GetSpellInfo(116888) -- death knight
local CHEATING_DEATH_AURA, _, CHEATING_DEATH_TEXTURE = GetSpellInfo(45182) -- rogue

local function UnitButton_UpdateFlags(self)
	local unit = self.unit
	if (not(unit)) then return end

	local texture, flag, text

	if (not(UnitIsConnected(unit))) then
		flag = "offline"
		text = PLAYER_OFFLINE
	elseif UnitIsDead(unit) then
		flag = "dead"
		texture = DEATH_TEXTURE
		text = DEAD
	elseif UnitDebuff(unit, GHOST_AURA) then
		flag = "ghost"
		texture = GHOST_TEXTURE
		text = DEAD
	elseif self.unitClass == "PRIEST" and UnitBuff(unit, SPIRIT_OF_REDEMPTION) then
		flag = "spirit"
		texture = SPIRIT_TEXTURE
		text = DEAD
	elseif self.unitClass == "MAGE" and select(11, UnitDebuff(unit, CAUTERIZE_AURA)) == 87023 then
		flag = "dying"
		texture = CAUTERIZE_TEXTURE
		text = CAUTERIZE_AURA
	elseif self.unitClass == "DEATHKNIGHT" and UnitDebuff(unit, PURGATORY_AURA) then
		flag = "dying"
		texture = PURGATORY_TEXTURE
		text = PURGATORY_AURA
	elseif self.unitClass == "ROGUE" and UnitBuff(unit, CHEATING_DEATH_AURA) then
		flag = "dying"
		texture = CHEATING_DEATH_TEXTURE
		text = CHEATING_DEATH_AURA
	end

	local db         = Healcsi.db.profile.UnitButtons.statusText
	local flagDb     = Healcsi.db.profile.UnitButtons.icons.flag
	local statusText = self.statusText
	local flagIcon   = self.flagIcon

	flagIcon:ClearAllPoints()
	flagIcon:SetPoint(flagDb.anchor, self.layerFrame, flagDb.anchor, flagDb.offset.x, flagDb.offset.y)
	flagIcon:SetSize(flagDb.size, flagDb.size)
	flagIcon:SetVertexColor(1, 1, 1, flagDb.alpha)
	flagIcon:SetTexCoord(flagDb.zoom, 1-flagDb.zoom, flagDb.zoom, 1-flagDb.zoom) 

	statusText:ClearAllPoints()
	statusText:SetPoint(db.anchor, self.healthBar, db.anchor, db.offset.x, db.offset.y)

	if (db.font.outline) then
		statusText:SetFont(Healcsi:GetMedia("font", db.font.file), db.font.size, "OUTLINE")
	else
		statusText:SetFont(Healcsi:GetMedia("font", db.font.file), db.font.size, nil)
	end

	local r, g, b, a
	if (db.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.color.fix)
	else
		if (db.color.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		elseif (db.color.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		end
		a = db.color.alpha
		if (db.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.color.hsv)
		elseif (db.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.color.rgb)
		end
	end
	statusText:SetTextColor(r, g, b, a)

	db = Healcsi.db.profile.UnitButtons.healthText.font.shadow

	if (db.color.mode == "fix") then
		r, g, b, a = Healcsi:ConvColorObj(db.color.fix)
	else
		if (db.color.mode == "class") then
			r, g, b, a = GetCalculatedClassColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		elseif (db.color.mode == "health") then
			r, g, b, a = GetCalculatedHealthColor(
				self,
				Healcsi.db.profile.UnitButtons.health,
				Healcsi.db.profile.UnitButtons.health.color.pet,
				Healcsi.db.profile.UnitButtons.health.color.inVehicle
			)
		end
		a = db.color.alpha
		if (db.color.shadeMode == "hsvShade") then
			r, g, b, a = Healcsi:ApplyHsvShade(r, g, b, a, db.color.hsv)
		elseif (db.color.shadeMode == "rgbShade") then
			r, g, b, a = Healcsi:ApplyRgbShade(r, g, b, a, db.color.rgb)
		end
	end
	statusText:SetShadowColor(r, g, b, a) 
	statusText:SetShadowOffset(db.offset.x, db.offset.y)

	flagIcon:SetTexture(texture)
	statusText:SetText(text)

	if (text) then
		self.healthText:Hide()
	else
		self.healthText:Show()
	end

	UnitButton_UpdateResurrect(self)

	if flag ~= self.unitFlag then
		self.unitFlag = flag
		BroadcastUnitNotify(self, "OnUnitFlagChange", flag)
	end

end

local READYCHECK_STATUS = { ready = READY_CHECK_READY_TEXTURE, notready = READY_CHECK_NOT_READY_TEXTURE, waiting = READY_CHECK_WAITING_TEXTURE }

local function UnitButton_UpdateReadyCheck(self)
	local unit = self.unit
	if not unit then return end
	local status = GetReadyCheckStatus(unit)
	self.readyCheckStatus = status
	local texture = READYCHECK_STATUS[status]
	if texture then
		self.readyCheckIcon:SetTexture(texture)

		local db = Healcsi.db.profile.UnitButtons.icons.readyCheck

		self.readyCheckIcon:ClearAllPoints()
		self.readyCheckIcon:SetPoint(db.anchor, self.healthBar, db.anchor, db.offset.x, db.offset.y)
		self.readyCheckIcon:SetSize(db.size, db.size)
		self.readyCheckIcon:SetVertexColor(1, 1, 1, db.alpha)

		self.readyCheckIcon:Show()
	else
		self.readyCheckIcon:Hide()
	end
end

local function UnitButton_FinishReadyCheck(self)
	if self.readyCheckStatus == "waiting" then
		self.readyCheckIcon:SetTexture(READYCHECK_STATUS.notready)
	end
	self.readyCheckHideTime = GetTime() + 6
end

local function UnitButton_DelayHideReadyCheck(self)
	local hideTime = self.readyCheckHideTime
	if hideTime and GetTime() > hideTime then
		self.readyCheckHideTime = nil
		self.readyCheckIcon:Hide()
	end
end

local function UnitButton_UpdateRole(self)
	local unit = self.unit
	if (not(unit)) then return end

	local icon = self.roleIcon
	local status
	local db = Healcsi.db.profile.UnitButtons.icons.role

	icon:ClearAllPoints()
	icon:SetPoint(db.anchor, self.layerFrame, db.anchor, db.offset.x, db.offset.y)
	icon:SetSize(db.size, db.size)
	icon:SetVertexColor(1, 1, 1, db.alpha)
	icon:SetTexCoord(db.zoom, 1-db.zoom, db.zoom, 1-db.zoom) 

	if (db.show) then
		if self.inVehicle then
			icon:SetTexture("Interface\\Vehicles\\UI-Vehicles-Raid-Icon")
			icon:SetTexCoord(0, 1, 0, 1)
			status = "VEHICLE"
		else
			status = UnitGroupRolesAssigned(unit)
			if status == "TANK" or status == "HEALER" or status == "DAMAGER" then
				icon:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-PORTRAITROLES")
				icon:SetTexCoord(GetTexCoordsForRoleSmallCircle(status))
			else
				status = nil
			end
		end
	end

	if (status) then icon:Show() else icon:Hide() end

	if (self.roleIconStatus ~= status) then
		self.roleIconStatus = status
		BroadcastUnitNotify(self, "OnRoleIconChange", status)
	end
end

local function UnitButton_UpdateRaidIcon(self)
	local unit = self.displayedUnit
	if (not(unit)) then return end

	local icon = self.raidIcon
	local db = Healcsi.db.profile.UnitButtons.icons.raidTarget

	icon:ClearAllPoints()
	icon:SetPoint(db.anchor, self.layerFrame, db.anchor, db.offset.x, db.offset.y)
	icon:SetSize(db.size, db.size)
	icon:SetVertexColor(1, 1, 1, db.alpha)
	--icon:SetTexCoord(db.zoom, 1-db.zoom, db.zoom, 1-db.zoom) 

	local index
	if (db.show) then index = GetRaidTargetIndex(unit) end

	if (index) then
		SetRaidTargetIconTexture(icon, index)
		icon:Show()
	else
		icon:Hide()
	end

	if (self.raidIconStatus ~= index) then
		self.raidIconStatus = index
		BroadcastUnitNotify(self, "OnRaidIconChange", index)
	end
end


local function UnitButton_UpdatePrivilege(self)
	local privFrame = self.privFrame

	local db = Healcsi.db.profile.UnitButtons.icons.privilege

	--if (not(db.show)) then
		--privFrame:Hide()
		--return
	--end

	--privFrame:SetPoint(db.anchor, self.healthBar, db.anchor, db.offset.x, db.offset.y)
	--privFrame:SetSize(db.size, db.size)

	local raidIndex, partyIndex = self.raidIndex, self.partyIndex
	local rank, loot, role , _

	if raidIndex then
		_, rank, _, _, _, _, _, _, _, role, loot = GetRaidRosterInfo(raidIndex)
	elseif partyIndex then
		if UnitIsGroupLeader(self.unit or "") then
			rank = 2
		end

		local lootMethod, lootMaster = GetLootMethod()
		if lootMethod == "master" and partyIndex == lootMaster then
			loot = 1
		end
	end

	local leaderIcon, lootIcon, mainTankIcon = self.leaderIcon, self.lootIcon, self.mainTankIcon
	local iconCount = 0

	if ((rank == 2) and db.leader.show) then -- Leader
		iconCount = iconCount + 1

		leaderIcon:SetPoint(db.leader.anchor, privFrame, db.leader.anchor, db.leader.offset.x, db.leader.offset.y)
		leaderIcon:SetSize(db.leader.size, db.leader.size)
		leaderIcon:SetSize(db.leader.size, db.leader.size)
		leaderIcon:SetVertexColor(1, 1, 1, db.leader.alpha)

		leaderIcon:SetTexture("Interface\\GroupFrame\\UI-Group-LeaderIcon")
		leaderIcon:Show()
	elseif (role == "MAINASSIST" and db.mainAssist.show) then -- I think main assist is higher than simple assist
		iconCount = iconCount + 1

		leaderIcon:SetPoint(db.mainAssist.anchor, privFrame, db.mainAssist.anchor, db.mainAssist.offset.x, db.mainAssist.offset.y)
		leaderIcon:SetSize(db.mainAssist.size, db.mainAssist.size)
		leaderIcon:SetSize(db.mainAssist.size, db.mainAssist.size)
		leaderIcon:SetVertexColor(1, 1, 1, db.mainAssist.alpha)

		rank = 3

		leaderIcon:SetTexture("Interface\\GroupFrame\\UI-Group-MainAssistIcon")
		leaderIcon:Show()
	elseif ((rank == 1) and (db.assist.show)) then -- Assist
		iconCount = iconCount + 1

		leaderIcon:SetPoint(db.assist.anchor, privFrame, db.assist.anchor, db.assist.offset.x, db.assist.offset.y)
		leaderIcon:SetSize(db.assist.size, db.assist.size)
		leaderIcon:SetSize(db.assist.size, db.assist.size)
		leaderIcon:SetVertexColor(1, 1, 1, db.assist.alpha)

		leaderIcon:SetTexture("Interface\\GroupFrame\\UI-Group-AssistantIcon")
		leaderIcon:Show()
	else
		rank = nil
		leaderIcon:Hide()
	end

	--[[
		rank nil = none
		rank 1   = assist
		rank 2   = leader
		rank 3   = main assist
	]]

	local mainTank = nil
	if (role == "MAINTANK") then
		iconCount = iconCount + 1
		mainTank = 1

		mainTankIcon:SetPoint(db.mainTank.anchor, privFrame, db.mainTank.anchor, db.mainTank.offset.x, db.mainTank.offset.y)
		mainTankIcon:SetSize(db.mainTank.size, db.mainTank.size)
		mainTankIcon:SetSize(db.mainTank.size, db.mainTank.size)
		mainTankIcon:SetVertexColor(1, 1, 1, db.mainTank.alpha)

		mainTankIcon:Show()
	else
		mainTank = nil
		mainTankIcon:Hide()
	end

	if (loot and (db.loot.show)) then
		iconCount = iconCount + 1
		loot = 1

		lootIcon:SetPoint(db.loot.anchor, privFrame, db.loot.anchor, db.loot.offset.x, db.loot.offset.y)
		lootIcon:SetSize(db.loot.size, db.loot.size)
		lootIcon:SetSize(db.loot.size, db.loot.size)
		lootIcon:SetVertexColor(1, 1, 1, db.loot.alpha)

		lootIcon:Show()
	else
		loot = nil
		lootIcon:Hide()
	end

	-- privFrame:Show()

	if ((self.privilegeLeader ~= rank) or (self.privilegeLoot ~= loot) or (self.mainTank ~= mainTank)) then
		self.privilegeLeader, self.privilegeLoot, self.mainTank = rank, loot, mainTank
		BroadcastUnitNotify(self, "OnPrivilegeChange", rank, loot, mainTank)
	end
end


local function UnitButton_UpdateAuras(self)
	local unit = self.displayedUnit
	if(not(unit)) then return end

	BroadcastUnitNotify(self, "OnAurasChange", unit, self.unit, self.inVehicle, self.unitClass)
end

-- ==============================================================
-- Update All

local function UnitButton_UpdateAll(self)
	if not self:IsVisible() then
		return
	end

	self.inRange = 0
	UnitButton_UpdateVehicleStatus(self)
	-- UnitButton_UpdateSize(self) don't allow here!
	UnitButton_UpdateMedia(self)

	UnitButton_UpdateFlags(self)
	UnitButton_UpdateRole(self)
	UnitButton_UpdateResurrect(self)

	UnitButton_UpdateAuras(self)

	UnitButton_UpdateHealthShieldAbsorbIncomingHeals(self)
	UnitButton_UpdateHealthColor(self)
	UnitButton_UpdateHealthUnderlayColor(self)
	UnitButton_UpdateShieldColor(self)
	UnitButton_UpdateShieldBorderColor(self)
	UnitButton_UpdateAbsorbColor(self)
	UnitButton_UpdateIncHealColor(self)
	UnitButton_UpdateHealthText(self)

	UnitButton_UpdatePowerMax(self)
	UnitButton_UpdatePower(self)
	UnitButton_UpdatePowerColor(self)
	UnitButton_UpdatePowerUnderlayColor(self)

	UnitButton_UpdateNameText(self)
	UnitButton_UpdateTarget(self)
	UnitButton_UpdateThreat(self)
	UnitButton_UpdateReadyCheck(self)

	UnitButton_UpdateInRange(self)
	UnitButton_UpdateRaidIcon(self)
	UnitButton_UpdatePrivilege(self)

end

-- ==============================================================
-- Register/Unregister Events

local function UnitButton_RegisterEvents(self)
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("GROUP_ROSTER_UPDATE")
	self:RegisterEvent("UNIT_HEALTH")
	self:RegisterEvent("UNIT_MAXHEALTH")
	self:RegisterEvent("UNIT_POWER")
	self:RegisterEvent("UNIT_MAXPOWER")
	self:RegisterEvent("UNIT_DISPLAYPOWER")
	self:RegisterEvent("UNIT_POWER_BAR_SHOW")
	self:RegisterEvent("UNIT_POWER_BAR_HIDE")
	self:RegisterEvent("UNIT_NAME_UPDATE")
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE")
	self:RegisterEvent("UNIT_CONNECTION")
	self:RegisterEvent("UNIT_HEAL_PREDICTION")
	self:RegisterEvent("PLAYER_ROLES_ASSIGNED")
	self:RegisterEvent("UNIT_ENTERED_VEHICLE")
	self:RegisterEvent("UNIT_EXITED_VEHICLE")
	self:RegisterEvent("READY_CHECK")
	self:RegisterEvent("READY_CHECK_FINISHED")
	self:RegisterEvent("READY_CHECK_CONFIRM")
	self:RegisterEvent("PARTY_MEMBER_DISABLE")
	self:RegisterEvent("PARTY_MEMBER_ENABLE")
	self:RegisterEvent("UNIT_HEALTH_FREQUENT")
	self:RegisterEvent("INCOMING_RESURRECT_CHANGED")
	self:RegisterEvent("RAID_TARGET_UPDATE")
	self:RegisterEvent("UNIT_FLAGS")
	self:RegisterEvent("UNIT_OTHER_PARTY_CHANGED")
	self:RegisterEvent("UNIT_ABSORB_AMOUNT_CHANGED")
	self:RegisterEvent("UNIT_HEAL_ABSORB_AMOUNT_CHANGED")
	UnitButton_UpdateAll(self)
end

local function UnitButton_UnregisterEvents(self)
	self:UnregisterAllEvents()
end

-- ==============================================================
-- UnitButton Event Handlers

local function UnitButton_OnAttributeChanged(self, name, value)
	if name == "unit" then
		self.unit, self.displayedUnit, self.raidIndex, self.partyIndex = nil
		if type(value) == "string" then
			self.unit = value
			self.displayedUnit = value

			if value == "player" then
				self.partyIndex = 0
			else
				local _, _, index = strfind(value, "^raid(%d+)$")
				if index then
					self.raidIndex = tonumber(index)
				else
					_, _, index = strfind(value, "^party(%d+)$")
					if index then
						self.partyIndex = tonumber(index)
					end
				end
			end
		end
	end
end

local function UnitButton_OnShow(self)
	self.needUpdate = 1
	UnitButton_RegisterEvents(self)
	Healcsi:BroadcastEvent("UpdateContainerSize")
end

local function UnitButton_OnHide(self)
	UnitButton_UnregisterEvents(self)
	Healcsi:BroadcastEvent("UpdateContainerSize")
end

local function UnitButton_OnTick(self)
	local e = (self.__tickCount or 0) + 1
	if e >= 5 then
		e = 0
		local guid = UnitGUID(self.displayedUnit or "")
		if guid ~= self.__origGuid then
			self.__origGuid = guid
			self.needUpdate = 1
		end
	end
	self.__tickCount = e

	UnitButton_UpdateInRange(self)
	UnitButton_DelayHideReadyCheck(self)

	if self.needUpdate then
		self.needUpdate = nil
		UnitButton_UpdateAll(self)
	elseif self.flagChanged then
		self.flagChanged = nil
		UnitButton_UpdateFlags(self)
	end
end

local function UnitButton_OnUpdate(self, elapsed)
	local e = (self.__updateElapsed or 0) + elapsed
	if e > 0.2 then
		UnitButton_OnTick(self)
		e = 0
	end
	self.__updateElapsed = e
end

local function UnitButton_OnEvent(self, event, unit)
	if event == "PLAYER_ENTERING_WORLD" or event == "PARTY_MEMBERS_CHANGED" or event == "GROUP_ROSTER_UPDATE" then
		self.needUpdate = 1
	elseif event == "PLAYER_TARGET_CHANGED" then
		UnitButton_UpdateTarget(self)
	elseif event == "PLAYER_ROLES_ASSIGNED" then
		UnitButton_UpdateRole(self)
	elseif event == "RAID_TARGET_UPDATE" then
		UnitButton_UpdateRaidIcon(self)
	elseif event == "READY_CHECK" then
		UnitButton_UpdateReadyCheck(self)
	elseif event == "READY_CHECK_FINISHED" then
		UnitButton_FinishReadyCheck(self)
	elseif event == "PARTY_MEMBER_DISABLE" or event == "PARTY_MEMBER_ENABLE" then
			UnitButton_UpdatePowerMax(self)
			UnitButton_UpdatePower(self)
			UnitButton_UpdatePowerColor(self)
			UnitButton_UpdatePowerUnderlayColor(self)
	elseif unit and (self.displayedUnit == unit or self.unit == unit) then
		if event == "UNIT_NAME_UPDATE" or event == "UNIT_ENTERED_VEHICLE" or event == "UNIT_EXITED_VEHICLE" or event == "UNIT_CONNECTION" or event == "UNIT_OTHER_PARTY_CHANGED" then
			self.needUpdate = 1
		elseif (
			 event == "UNIT_MAXHEALTH" or
			 event == "UNIT_HEALTH" or
			 event == "UNIT_HEALTH_FREQUENT" or
			 event == "UNIT_HEAL_PREDICTION" or
			 event == "UNIT_ABSORB_AMOUNT_CHANGED" or
			 event == "UNIT_HEAL_ABSORB_AMOUNT_CHANGED"
		) then
			UnitButton_UpdateHealthShieldAbsorbIncomingHeals(self)
			UnitButton_UpdateHealthColor(self)
			UnitButton_UpdateHealthUnderlayColor(self)
			UnitButton_UpdateShieldColor(self)
			UnitButton_UpdateShieldBorderColor(self)
			UnitButton_UpdateAbsorbColor(self)
			UnitButton_UpdateIncHealColor(self)
			UnitButton_UpdateHealthText(self)
		elseif event == "UNIT_MAXPOWER" then
			UnitButton_UpdatePowerMax(self)
			UnitButton_UpdatePower(self)
		elseif event == "UNIT_POWER" then
			UnitButton_UpdatePower(self)
		elseif event == "UNIT_DISPLAYPOWER" or event == "UNIT_POWER_BAR_SHOW" or event == "UNIT_POWER_BAR_HIDE" then
			UnitButton_UpdatePowerMax(self)
			UnitButton_UpdatePower(self)
			UnitButton_UpdatePowerColor(self)
			UnitButton_UpdatePowerUnderlayColor(self)
		elseif event == "UNIT_AURA" or event == "UNIT_FLAGS" then
			UnitButton_UpdateAuras(self)
			self.flagChanged = 1
		elseif event == "UNIT_THREAT_SITUATION_UPDATE" then
			UnitButton_UpdateThreat(self)
		elseif event == "READY_CHECK_CONFIRM" then
			UnitButton_UpdateReadyCheck(self)
		elseif event == "INCOMING_RESURRECT_CHANGED" then
			UnitButton_UpdateResurrect(self)
		end
	end
end

local function UnitButton_OnEnter(self)
	local unit = self.displayedUnit
	if (not(unit)) then return end
	self.needUpdate = 1

	self.highlight:SetVertexColor(Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.selection.highlightColor))
	self.highlight:Show()

	local db = Healcsi.db.profile.UnitButtons.tooltip
	if (db.show == "always" or (db.show == "outOfCombat" and not(InCombatLockdown()))) then
		if (db.position == "default") then
			GameTooltip_SetDefaultAnchor(GameTooltip, self)
		else
			local right = self:GetLeft() < 300
			local bottom = GetScreenHeight() - self:GetTop() < 360

			local pointX = right and "LEFT" or "RIGHT"
			local pointY = bottom and "TOP" or "BOTTOM"
			local relPointX = right and "RIGHT" or "LEFT"
			local relPointY = bottom and "BOTTOM" or "TOP"

			GameTooltip:SetOwner(self, "ANCHOR_NONE")
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint(pointY..pointX, self, relPointY..relPointX, right and 4 or -4, bottom and -4 or 4)
		end
		GameTooltip:SetUnit(unit)
	end

end

local function UnitButton_OnLeave(self)
	GameTooltip:Hide()
	self.highlight:Hide()
end

-- ==============================================================
-- UnitButton Getters
local function UnitButton_GetUnitInfo(self)
	return self.displayedUnit, self.unit, self.inVehicle, self.unitClass
end

local function UnitButton_GetUnitFlag(self)
	return self.unitFlag
end

local function UnitButton_IsInRange(self)
	return self.inRange
end

local function UnitButton_GetRoleIconStatus(self)
	return self.roleIconStatus
end

local function UnitButton_GetRaidIconStatus(self)
	return self.raidIconStatus
end

local function UnitButton_GetPrivilege(self)
	return self.privilegeLeader, self.privilegeLoot, self.mainTank
end

-- ==============================================================
-- UnitButton Constructor (OnLoad)

function UnitButton_OnLoad(frame)

	local name = frame:GetName()

	-- Art frame
	local artFrame = CreateFrame("Frame", name.."ArtFrame", frame)
	frame.artFrame = artFrame
	artFrame:SetAllPoints(frame)

	-- Background
	local background = artFrame:CreateTexture(name.."Background", "BACKGROUND")
	frame.background = background
	background:SetAllPoints(artFrame)

	background:SetTexture(Healcsi:GetMedia("background", Healcsi.db.profile.UnitButtons.background.file))
	background:SetVertexColor(Healcsi:ConvColorObj(Healcsi.db.profile.UnitButtons.background.color))

	-- Health bar
	local healthBar = CreateFrame("StatusBar", name.."HealthBar", artFrame)
	frame.healthBar = healthBar
	healthBar:SetPoint("TOPLEFT", 1, -1)
	healthBar:SetPoint("BOTTOMRIGHT", -1, 2)

	healthBar:SetStatusBarTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.health.file))
	-- color will be updated somewhere else

	local underlay = artFrame:CreateTexture(name.."HealthBarUnderlay", "BORDER")
	frame.healthBarUnderlay = underlay
	underlay:SetAllPoints(healthBar)
	underlay:SetTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.health.underlay.file))
	-- color will be updated somewhere else

	-- Power bar
	local powerBar = CreateFrame("StatusBar", name.."PowerBar", artFrame)
	frame.powerBar = powerBar
	powerBar:SetPoint("TOPLEFT", healthBar, "BOTTOMLEFT", 0, -1)
	powerBar:SetPoint("BOTTOMRIGHT", -1, 1)

	powerBar:SetStatusBarTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.power.file))
	-- color will be updated somewhere else

	underlay = artFrame:CreateTexture(name.."PowerBarUnderlay", "BORDER")
	frame.powerBarUnderlay = underlay
	underlay:SetAllPoints(powerBar)
	underlay:SetTexture(Healcsi:GetMedia("background", Healcsi.db.profile.UnitButtons.power.underlay.file))
	-- color will be updated somewhere else

	-- Shield bar
	local shieldBar = CreateFrame("StatusBar", name.."ShieldBar", healthBar)
	frame.shieldBar = shieldBar
	shieldBar:SetPoint("TOPLEFT", healthBar:GetStatusBarTexture(), "TOPRIGHT")
	shieldBar:SetPoint("BOTTOMRIGHT", 0, 0)

	shieldBar:SetStatusBarTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.shield.file))
	-- shieldBar:Hide() -- never hide cause incomingHeal frame sits on it!

	local shieldBorderLeft   = shieldBar:CreateTexture(name.."ShieldBarBorderLeft",   "OVERLAY")
	shieldBorderLeft:SetPoint("TOPLEFT", shieldBar, "TOPLEFT", 0, 0)
	shieldBorderLeft:SetPoint("BOTTOMRIGHT", shieldBar, "BOTTOMLEFT", Healcsi.db.profile.UnitButtons.shield.border.left.size, 0)
	shieldBorderLeft:SetTexture(Media.borderTextures[Healcsi.db.profile.UnitButtons.shield.border.left.file])

	local shieldBorderRight  = shieldBar:CreateTexture(name.."ShieldBarBorderRight",  "OVERLAY")
	shieldBorderRight:SetPoint("TOPRIGHT", shieldBar:GetStatusBarTexture(), "TOPRIGHT", 0, 0)
	shieldBorderRight:SetPoint("BOTTOMLEFT", shieldBar:GetStatusBarTexture(), "BOTTOMRIGHT", -Healcsi.db.profile.UnitButtons.shield.border.right.size, 0)
	shieldBorderRight:SetTexture(Media.borderTextures[Healcsi.db.profile.UnitButtons.shield.border.right.file])

	local shieldBorderTop    = shieldBar:CreateTexture(name.."ShieldBarBorderTop",    "OVERLAY")
	shieldBorderTop:SetPoint("TOPLEFT", shieldBar, "TOPLEFT", 0, 0)
	shieldBorderTop:SetPoint("BOTTOMRIGHT", shieldBar:GetStatusBarTexture(), "TOPRIGHT", 0, -Healcsi.db.profile.UnitButtons.shield.border.top.size)
	shieldBorderTop:SetTexture(Media.borderTextures[Healcsi.db.profile.UnitButtons.shield.border.top.file])

	local shieldBorderBottom = shieldBar:CreateTexture(name.."ShieldBarBorderBottom", "OVERLAY")
	shieldBorderBottom:SetPoint("BOTTOMLEFT", shieldBar, "BOTTOMLEFT", 0, 0)
	shieldBorderBottom:SetPoint("TOPRIGHT", shieldBar:GetStatusBarTexture(), "BOTTOMRIGHT", 0, Healcsi.db.profile.UnitButtons.shield.border.bottom.size)
	shieldBorderBottom:SetTexture(Media.borderTextures[Healcsi.db.profile.UnitButtons.shield.border.bottom.file])

	frame.shieldBorderLeft   = shieldBorderLeft
	frame.shieldBorderRight  = shieldBorderRight
	frame.shieldBorderTop    = shieldBorderTop
	frame.shieldBorderBottom = shieldBorderBottom

	-- Absorbs bar
	local absorbsBar = CreateFrame("StatusBar", name.."AbsorbsBar", healthBar)
	absorbsBar:SetFrameLevel(100)
	frame.absorbsBar = absorbsBar
	absorbsBar:SetAllPoints(healthBar)
	absorbsBar:SetReverseFill(true)
	absorbsBar:SetStatusBarTexture(Media.absorbTextures[Healcsi.db.profile.UnitButtons.absorb.file])
	absorbsBar:Hide()

	-- Incoming healing bar
	local incomingHealBar = CreateFrame("StatusBar", name.."IncomingHealBar", healthBar)
	frame.incomingHealBar = incomingHealBar
	incomingHealBar:SetPoint("TOPLEFT", shieldBar:GetStatusBarTexture(), "TOPRIGHT")
	incomingHealBar:SetPoint("BOTTOMRIGHT", 0, 0)
	incomingHealBar:SetStatusBarTexture(Healcsi:GetMedia("statusbar", Healcsi.db.profile.UnitButtons.incHeal.file))
	incomingHealBar:Hide()

	local incomingHeal2AbsorbBar = CreateFrame("StatusBar", name.."IncomingHeal2AbsorbBar", absorbsBar)
	frame.incomingHeal2AbsorbBar = incomingHeal2AbsorbBar
	incomingHeal2AbsorbBar:SetPoint("TOPLEFT", absorbsBar:GetStatusBarTexture(), "TOPLEFT")
	incomingHeal2AbsorbBar:SetPoint("BOTTOMRIGHT", absorbsBar, "BOTTOMRIGHT")
	incomingHeal2AbsorbBar:SetStatusBarTexture(Media.absorbTextures[Healcsi.db.profile.UnitButtons.incHeal.absorb.file])
	incomingHeal2AbsorbBar:Hide()

	-- Layer frame
	local layerFrame = CreateFrame("Frame", name.."LayerFrame", artFrame)
	frame.layerFrame = layerFrame
	layerFrame:SetAllPoints(artFrame)
	layerFrame:SetFrameLevel(50)

	-- Highlight texture
	local highlight = layerFrame:CreateTexture(name.."Highlight", "BACKGROUND")
	frame.highlight = highlight
	highlight:SetAllPoints(frame)
	highlight:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
	highlight:SetBlendMode("ADD")
	highlight:Hide()

	-- Name text
	local nameText = layerFrame:CreateFontString(name.."NameText", "BORDER")
	frame.nameText = nameText

	-- Health text
	local healthText = layerFrame:CreateFontString(name.."StatusText", "BORDER") -- OVERLAY?
	frame.healthText = healthText

	-- Status text
	local statusText = layerFrame:CreateFontString(name.."StatusText", "BORDER") -- OVERLAY?
	frame.statusText = statusText

	-- Role icon
	local roleIcon = layerFrame:CreateTexture(name.."RoleIcon", "ARTWORK")
	frame.roleIcon = roleIcon

	-- Flag icon (dead/ghost/spirit)
	local flagIcon = layerFrame:CreateTexture(name.."FlagIcon", "ARTWORK")
	frame.flagIcon = flagIcon

	-- Resurrect icon
	local resIcon = layerFrame:CreateTexture(name.."ResIcon", "ARTWORK")
	frame.resIcon = resIcon
	resIcon:SetTexture("Interface\\RaidFrame\\Raid-Icon-Rez")
	resIcon:Hide()

	-- Raid icon
	local raidIcon = layerFrame:CreateTexture(name.."RaidIcon", "OVERLAY")
	frame.raidIcon = raidIcon
	--raidIcon:SetSize(14, 14)
	--raidIcon:SetPoint("TOP", 0, 3)
	raidIcon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
	raidIcon:Hide()

	-- Ready check icon
	local readyCheckIcon = layerFrame:CreateTexture(name.."ReadyCheckIcon", "OVERLAY")
	frame.readyCheckIcon = readyCheckIcon
	readyCheckIcon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
	readyCheckIcon:Hide()

	-- Container of privilege icons
	local privFrame = CreateFrame("Frame", name.."PrivilegeFrame", layerFrame)
	privFrame:SetAllPoints(healthBar)
	frame.privFrame = privFrame

	local leaderIcon = privFrame:CreateTexture(name.."LeaderIcon", "OVERLAY")
	frame.leaderIcon = leaderIcon
	leaderIcon:Hide()

	local lootIcon = privFrame:CreateTexture(name.."LootIcon", "OVERLAY")
	frame.lootIcon = lootIcon
	lootIcon:SetTexture("Interface\\GroupFrame\\UI-Group-MasterLooter")
	lootIcon:Hide()

	local mainTankIcon = privFrame:CreateTexture(name.."MainTankIcon", "OVERLAY")
	frame.mainTankIcon = mainTankIcon
	mainTankIcon:SetTexture("Interface\\GroupFrame\\UI-Group-MainTankIcon")
	mainTankIcon:Hide()

	-- Aggro highlight
	local aggroHighlight = layerFrame:CreateTexture(name.."AggroHighlight", "BACKGROUND")
	frame.aggroHighlight = aggroHighlight
	aggroHighlight:SetAllPoints(frame)
	aggroHighlight:Hide()

	-- Selection highlight
	local selectionHighlight = layerFrame:CreateTexture(name.."SelectionHighlight", "BORDER")
	frame.selectionHighlight = selectionHighlight
	selectionHighlight:SetAllPoints(frame)
	selectionHighlight:Hide()

	-- Aura frames
	local buffParent   = CreateFrame("Frame", name.."BuffParent", layerFrame)
	local debuffParent = CreateFrame("Frame", name.."DeBuffParent", layerFrame)
	local dispelParent = CreateFrame("Frame", name.."DispelParent", healthBar)

	frame.buffParent   = buffParent
	frame.debuffParent = debuffParent
	frame.dispelParent = dispelParent
	frame.dispelable   = {}

	-- Visual parent frame
	local visualParent = CreateFrame("Frame", name.."VisualParent", layerFrame)
	frame.visualParent = visualParent
	visualParent:SetAllPoints(layerFrame)
	visualParent:SetFrameLevel(150)


	-- Basic scripts
	frame:SetScript("OnAttributeChanged", UnitButton_OnAttributeChanged)
	frame:SetScript("OnShow", UnitButton_OnShow)
	frame:SetScript("OnHide", UnitButton_OnHide)
	frame:SetScript("OnUpdate", UnitButton_OnUpdate)
	frame:SetScript("OnEvent", UnitButton_OnEvent)

	-- OnEnter/OnLeave already used by SecureHandlerEnterLeaveTemplate so we take HookScript instead
	frame:HookScript("OnEnter", UnitButton_OnEnter)
	frame:HookScript("OnLeave", UnitButton_OnLeave)

	-- TODO: add all functions... but i am lazzyyyy

	frame.GetUnitInfo = UnitButton_GetUnitInfo -- this one i needed :)

	-- UnitButton_UpdateMedia(self)
	-- UnitButton_UpdateHealthColor
	-- UnitButton_UpdateHealthUnderlayColor
	-- UnitButton_UpdateHealthShieldAbsorbIncomingHeals
	-- UnitButton_UpdateVehicleStatus(self)
	-- UnitButton_UpdateInRange

	-- UnitButton_UpdatePowerMax(self)
	-- UnitButton_UpdatePower(self)
	-- UnitButton_UpdatePowerColor(self)
	-- UnitButton_UpdatePowerUnderlayColor(self)
	-- UnitButton_UpdateNameText(self)
	-- UnitButton_UpdateSize(self)

	tinsert(UnitButtons.unitFrames, frame)
	Healcsi:BroadcastEvent("UnitButtonCreated", frame)
end


Healcsi:RegisterEventCallback("UnitButtonConfigChanged", function()
	for _, frame in ipairs(UnitButtons.unitFrames) do
		UnitButton_UpdateAll(frame)
		if (not(Editor.InCombat)) then
			UnitButton_UpdateSize(frame)
		else
			frame.updateOutOfCombat = 1
		end
	end
end)

Healcsi:RegisterEventCallback("GlobalConfigRefresh", function()
	for _, frame in ipairs(UnitButtons.unitFrames) do
		UnitButton_UpdateAll(frame)
		if (not(Editor.InCombat)) then
			UnitButton_UpdateSize(frame)
		else
			frame.updateOutOfCombat = 1
		end
	end
end)

Healcsi:RegisterEventCallback("OutOfCombat", function()
	for _, frame in ipairs(UnitButtons.unitFrames) do
		if (frame.updateOutOfCombat) then
			UnitButton_UpdateSize(frame)
			frame.updateOutOfCombat = nil
		end
	end
end)


