local Healcsi     = Healcsi
local UnitFrames  = Healcsi.Modules.UnitFrames
local UnitButtons = Healcsi.Modules.UnitButtons
local Editor      = Healcsi.Editor

local max = max
local RegisterStateDriver = RegisterStateDriver
local UnregisterStateDriver = UnregisterStateDriver

function UnitFrames:FriendlyNpc_CreateFrame()
	
	local frame = self:Container_CreateGroupParent("HealcsiFriendlyNpcFrame", "SecureHandlerStateTemplate")
	self.Friendly = {}
	self.Friendly.frame = frame
	frame:Hide()

	function frame:UpdateContainerSize()
		Healcsi:BroadcastEvent("UpdateContainerSize")
	end

	HealcsiRaidGroupHeaderPet:SetSize(1, 36) -- Ensures NPC buttons always have a valid anchor even if the user did not enable the option "showRaidPets"

	local anchors = {
		solo = HealcsiSoloFramePlayer,
		party = HealcsiPartyFramePet,
		raid = HealcsiRaidGroupHeaderPet,
	}

	local k, v
	for k, v in pairs(anchors) do
		frame:SetFrameRef("anchor"..k, v)
	end

	self.Friendly.anchors = anchors
	frame.buttons = {}

	local i
	for i = 1, 5 do
		local button = CreateFrame("Button", frame:GetName().."Button"..i, frame, "HealcsiUnitButtonTemplate")
		button:Hide(i)
		button:SetAttribute("unit", "boss"..i)

		tinsert(frame.buttons, button)

		if i == 1 then
			button:SetPoint("TOPLEFT")
		end
	end

	frame:SetAttribute("_onstate-groupstate", [[
		local horiz = self:GetAttribute("horiz")
		local spacing = self:GetAttribute("spacing") or 0
		local showPartyPets = self:GetAttribute("showPartyPets")
		local showRaidPets = self:GetAttribute("showRaidPets")
		local anchor = self:GetFrameRef("anchor"..newstate)

		if anchor then
			self:ClearAllPoints()
			if newstate == "raid" and not showRaidPets then
				self:SetPoint("TOPLEFT", anchor, "TOPLEFT", 0, 0)
			elseif newstate == "party" and not showPartyPets then
				self:SetPoint("TOPLEFT", anchor, "TOPLEFT", 0, 0)
			elseif horiz then
				self:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, -spacing)
			else
				self:SetPoint("TOPLEFT", anchor, "TOPRIGHT", spacing, 0)
			end

			self:CallMethod("UpdateContainerSize")
		end
	]])

	RegisterStateDriver(frame, "groupstate", "[group:raid] raid; [group:party] party; solo")

	

end

-- Frame matrix functions need to be redesigned
local Orig_Solo_GetFramesMatrix  = UnitFrames.Solo_GetFramesMatrix
local Orig_Party_GetFramesMatrix = UnitFrames.Party_GetFramesMatrix
local Orig_Raid_GetFramesMatrix  = UnitFrames.Raid_GetFramesMatrix


function UnitFrames:FriendlyNpc_GetFramesMatrix()

	if (not(self.Friendly)) then self:Friendly_CreateFrame() end

	local frame = self.Friendly.frame

	if not frame:IsShown() then
		return
	end

	local count
	local i
	for i = 1, 5 do
		if frame.buttons[i]:IsShown() then
			count = i
		end
	end

	if not count then
		return
	end

	if (self:GetAlignHoriz()) then
		return count, 1
	else
		return 1, count
	end
end

function UnitFrames:FriendlyNpc_AppendMatrix(cols, rows, hasPet, group, show)
	local c1, r1 = self:FriendlyNpc_GetFramesMatrix()
	if not c1 then
		return cols, rows, hasPet
	end

	local reserve = 1

	local currGroup = self:GetGroupStats()
	if group == currGroup and not show then
		reserve = 0
	end

	if (self:GetAlignHoriz()) then
		if hasPet then
			return max(cols, c1), rows + 1, hasPet
		else
			return max(cols, c1), rows + 1 + reserve, hasPet
		end
	else
		if hasPet then
			return cols + 1, max(rows, r1), hasPet
		else
			return cols + 1 + reserve, max(rows, r1), hasPet
		end
	end
end

function UnitFrames:Solo_GetFramesMatrix()
	local cols, rows, hasPet = Orig_Solo_GetFramesMatrix(self)
	local c1, r1 = self:FriendlyNpc_GetFramesMatrix()
	if not c1 then
		return cols, rows, hasPet
	end

	if (self:GetAlignHoriz()) then
		return max(cols, c1), rows + 1, hasPet
	else
		return cols + 1, max(rows, r1), hasPet
	end
end

function UnitFrames:Party_GetFramesMatrix()
	local cols, rows, hasPet = Orig_Party_GetFramesMatrix(self)
	return self:FriendlyNpc_AppendMatrix(cols, rows, hasPet, "party", Healcsi.db.profile.UnitFrames.party.showPartyPets)
end

function UnitFrames:Raid_GetFramesMatrix()
	local cols, rows, hasPet = Orig_Raid_GetFramesMatrix(self)
	return self:FriendlyNpc_AppendMatrix(cols, rows, hasPet, "raid", Healcsi.db.profile.UnitFrames.raid.showRaidPets)
end

function UnitFrames:FriendlyNpc_UpdateLayout()

	if (not(self.Friendly)) then self:Friendly_CreateFrame() end

	local horiz, spacing = self:GetLayoutData()
	local group = self:GetGroupStats()
	local anchor = self.Friendly.anchors[group] or self.Friendly.anchors.solo

	local frame = self.Friendly.frame

	frame:ClearAllPoints()
	if group == "raid" and not Healcsi.db.profile.UnitFrames.raid.showRaidPets then
		frame:SetPoint("TOPLEFT", anchor, "TOPLEFT")
	elseif group == "party" and not Healcsi.db.profile.UnitFrames.party.showPartyPets then
		frame:SetPoint("TOPLEFT", anchor, "TOPLEFT")
	elseif horiz then
		frame:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, -spacing)
	else
		frame:SetPoint("TOPLEFT", anchor, "TOPRIGHT", spacing, 0)
	end

	frame:SetAttribute("horiz", horiz)
	frame:SetAttribute("spacing", spacing)

	local i
	for i = 2, 5 do
		local button = frame.buttons[i]
		button:ClearAllPoints()
		if horiz then
			button:SetPoint("LEFT", frame.buttons[i - 1], "RIGHT", spacing, 0)
		else
			button:SetPoint("TOP", frame.buttons[i - 1], "BOTTOM", 0, -spacing)
		end
	end

end

Healcsi:RegisterEventCallback("UnitButtonConfigChanged", function()
	if (not(Editor.InCombat)) then
		UnitFrames:FriendlyNpc_UpdateLayout()
	end
end)

Healcsi:RegisterEventCallback("UnitFrameConfigChanged", function()
	if (not(Editor.InCombat)) then
		UnitFrames:FriendlyNpc_UpdateLayout()
		local showFriendlyNPC = Healcsi.db.profile.UnitFrames.friendly.showFriendlyNPC
		local frame = UnitFrames.Friendly.frame
		local i
		for i = 1, 5 do
			local button = frame.buttons[i]
			if showFriendlyNPC then
				RegisterStateDriver(button, "visibility", "[@boss"..i..", help] show; hide")
			else
				UnregisterStateDriver(button, "visibility")
				button:Hide()
			end
		end

		if showFriendlyNPC then
			frame:Show()
		else
			frame:Hide()
		end

		Healcsi:BroadcastEvent("UpdateContainerSize")
	end
end)

Healcsi:RegisterEventCallback("GlobalConfigRefresh", function()
	if (not(Editor.InCombat)) then
		UnitFrames:FriendlyNpc_UpdateLayout()
		local showFriendlyNPC = Healcsi.db.profile.UnitFrames.friendly.showFriendlyNPC
		local frame = UnitFrames.Friendly.frame
		local i
		for i = 1, 5 do
			local button = frame.buttons[i]
			if showFriendlyNPC then
				RegisterStateDriver(button, "visibility", "[@boss"..i..", help] show; hide")
			else
				UnregisterStateDriver(button, "visibility")
				button:Hide()
			end
		end

		if showFriendlyNPC then
			frame:Show()
		else
			frame:Hide()
		end

		Healcsi:BroadcastEvent("UpdateContainerSize")
	end
end)