local Healcsi     = Healcsi
local UnitFrames  = Healcsi.Modules.UnitFrames
local UnitButtons = Healcsi.Modules.UnitButtons
local Editor      = Healcsi.Editor

function UnitFrames:Raid_GetRaidGroup(key)
	if (not(self.Raid)) then self:Raid_CreateFrame() end
	return self.Raid.groupHeaders[key]
end

local function ForceUpdateHeader(header)
	header:SetAttribute("unitsPerColumn", 5)
end

local function Header_GetButtonCount(header)
	local count = 0
	local i
	for i = 1, #header do
		if header[i]:IsVisible() then
			count = i
		else
			break
		end
	end
	return count
end

local function Header_GetMatrix(header)
	local units = Header_GetButtonCount(header)
	if (UnitFrames:GetAlignHoriz()) then
		return min(units, 5), ceil(units / 5)
	else
		return ceil(units / 5), min(units, 5)
	end
end

local function ForceCreateButtons(header, count)
	local Raid = UnitFrames.Raid
	if (Raid.initalizedHeaders[header] or InCombatLockdown() or (count <= #header)) then
		return
	end

	Raid.initalizedHeaders[header] = 1
	header:SetAttribute("startingIndex", 1 - count)
	header:Show()
	header:SetAttribute("startingIndex", 1)
	header:SetAttribute("showRaid", 1)

	local helper = header.monitorFrame
	header.monitorFrame = nil
	local button = header[1]
	if helper and button then
		helper:SetParent(button)
		helper:SetAllPoints(button)
		helper:Show()
	end
end

local function CreateGroupHeader(key, name, count, template, parent)
	local Raid = UnitFrames.Raid
	local header = CreateFrame("Frame", name, parent or UnitFrames:GetContainer(), template)
	header:Hide()

	Raid.groupHeaders[key] = header
	header:SetPoint("TOPLEFT", UnitFrames:GetContainer(), "TOPLEFT")

	header:SetAttribute("initialConfigFunction", [[
		local parent = self:GetParent()
		self:SetWidth(parent:GetAttribute("unitwidth") or 80)
		self:SetHeight(parent:GetAttribute("unitheight") or 50)
	]])

	header:SetAttribute("template", "HealcsiUnitButtonTemplate")
	header:SetAttribute("point", "TOP")
	header:SetAttribute("xOffset", 0)
	header:SetAttribute("yOffset", -1)
	header:SetAttribute("unitsPerColumn", 5)
	header:SetAttribute("columnAnchorPoint", "LEFT")
	header:SetAttribute("columnSpacing", 1)
	header:SetAttribute("maxColumns", count / 5)

	return header
end

function UnitFrames:Raid_CreateFrame()
	local frame = self:Container_CreateGroupParent("HealcsiRaidGroupsFrame", "SecureHandlerAttributeTemplate")
	frame:Hide()

	self.Raid = {}
	self.Raid.frame = frame
	self.Raid.groupHeaders = {}
	self.Raid.initalizedHeaders = {}

	-- This snippet allows repositioning the raid pet group in combat
	frame:SetAttribute("_onattributechanged", [[
		if name ~= "helpershow" or not self:IsShown() then
			return
		end

		local i, maxGroup
		for i = 1, 8 do
			if self:GetFrameRef("buttonhelper"..i):IsVisible() then
				maxGroup = i
			end
		end

		if not maxGroup then
			return
		end

		local header = self:GetFrameRef("subgroup"..maxGroup)
		local petGroup = self:GetFrameRef("petgroup")
		local spacing = self:GetAttribute("spacing") or 1
		local horiz = self:GetAttribute("horiz")

		petGroup:ClearAllPoints()
		if horiz then
			petGroup:SetPoint("TOPLEFT", header, "BOTTOMLEFT", 0, -spacing)
		else
			petGroup:SetPoint("TOPLEFT", header, "TOPRIGHT", spacing, 0)
		end
	]])

	-- Normal group
	local raidGroup = CreateGroupHeader("group", "HealcsiRaidGroupHeaderGroup", 40, "SecureGroupHeaderTemplate")
	self.Raid.frame.raidGroup = raidGroup

	-- Pets
	local petGroup = CreateGroupHeader("pet", "HealcsiRaidGroupHeaderPet", 10, "SecureGroupPetHeaderTemplate")
	petGroup:SetAttribute("filterOnPet", 1)
	self.Raid.frame:SetFrameRef("petgroup", petGroup)
	self.Raid.frame.petGroup = petGroup

	-- 8 subgroups
	local i
	for i = 1, 8 do
		local header = CreateGroupHeader(i, "HealcsiRaidGroupHeaderSubGroup"..i, 5, "SecureGroupHeaderTemplate", raidFrame)
		self.Raid.groupHeaders[i] = header
		header:SetAttribute("groupFilter", i)
		self.Raid.frame:SetFrameRef("subgroup"..i, header)

		local helper = CreateFrame("Frame", nil, header, "SecureHandlerShowHideTemplate")
		header.monitorFrame = helper
		helper:Hide()
		helper:SetFrameRef("targetframe", self.Raid.frame)
		self.Raid.frame:SetFrameRef("buttonhelper"..i, helper)
		helper:SetAttribute("_onshow", [[ self:GetFrameRef("targetframe"):SetAttribute("helpershow", 1) ]])
		helper:SetAttribute("_onhide", [[ self:GetFrameRef("targetframe"):SetAttribute("helpershow", 0) ]])
	end

	self:Raid_UpdateLayout()
	self:Raid_UpdateFrames()

end

local function GetSubGroupsMatrix()
	local maxGroup, maxUnits = 0, 0
	local i
	for i = 1, 8 do
		local units = Header_GetButtonCount(UnitFrames.Raid.groupHeaders[i])
		if (units > 0) then
			if (maxGroup < i) then maxGroup = i end
			if (maxUnits < units) then maxUnits = units end
		end
	end

	if (UnitFrames:GetLayoutData()) then
		return maxUnits, maxGroup
	else
		return maxGroup, maxUnits
	end
end

function UnitFrames:Raid_GetFramesMatrix()
	local cols, rows
	if (self:GetKeepGroupsTogether()) then
		cols, rows = GetSubGroupsMatrix()
	else
		cols, rows = Header_GetMatrix(self.Raid.frame.raidGroup)
	end

	local petCols, petRows = Header_GetMatrix(self.Raid.frame.petGroup)
	local hasPet = petCols > 0 or petRows > 0

	if (self:GetLayoutData()) then
		rows = rows + petRows
	else
		cols = cols + petCols
	end

	return cols, rows, hasPet
end

function UnitFrames:Raid_UpdateLayout()
	local horiz, spacing = self:GetLayoutData()
	local key, header
	for key, header in pairs(self.Raid.groupHeaders) do
		if (horiz) then
			header:SetAttribute("columnAnchorPoint", "TOP")
			header:SetAttribute("point", "LEFT")
			header:SetAttribute("xOffset", spacing)
			header:SetAttribute("yOffset", 0)
		else
			header:SetAttribute("columnAnchorPoint", "LEFT")
			header:SetAttribute("point", "TOP")
			header:SetAttribute("xOffset", 0)
			header:SetAttribute("yOffset", -spacing)
		end

		local i
		for i = 1, #header do header[i]:ClearAllPoints() end
		ForceUpdateHeader(header)
	end

	local i
	for i = 2, 8 do
		local header = self.Raid.groupHeaders[i]
		header:ClearAllPoints()
		if (horiz) then
			header:SetPoint("TOPLEFT", self.Raid.groupHeaders[i - 1], "BOTTOMLEFT", 0, -spacing)
		else
			header:SetPoint("TOPLEFT", self.Raid.groupHeaders[i - 1], "TOPRIGHT", spacing, 0)
		end
	end

	self.Raid.frame:SetAttribute("horiz", horiz)
	self.Raid.frame:SetAttribute("spacing", spacing)

	if (self:GetKeepGroupsTogether()) then
		self.Raid.frame:SetAttribute("helpershow", 1)
	else
		self.Raid.frame.petGroup:ClearAllPoints()
		if (horiz) then
			self.Raid.frame.petGroup:SetPoint("TOPLEFT", self.Raid.frame.raidGroup, "BOTTOMLEFT", 0, -spacing)
		else
			self.Raid.frame.petGroup:SetPoint("TOPLEFT", self.Raid.frame.raidGroup, "TOPRIGHT", spacing, 0)
		end
	end

end

local function SortHeader(header, groupBy, groupingOrder, sortMethod)
	local wasShow = header:IsShown()
	header:Hide()

	header:SetAttribute("sortMethod", sortMethod)
	header:SetAttribute("groupingOrder", groupingOrder)
	header:SetAttribute("groupBy", groupBy)
	header:SetAttribute("sortDir", "ASC")

	if wasShow then
		header:Show()
	end
end

function UnitFrames:Raid_UpdateFrames()
	-- showRaidPets:
	if (Healcsi.db.profile.UnitFrames.raid.showRaidPets) then
		ForceCreateButtons(UnitFrames.Raid.frame.petGroup, 20)
		self.Raid.frame.petGroup:Show()
	else
		UnitFrames.Raid.frame.petGroup:Hide()
	end

	-- keepgroupstogether
	if (self:GetKeepGroupsTogether()) then
		self.Raid.frame.raidGroup:Hide()
		self.Raid.frame:Show()
		local i
		for i = 1, 8 do
			local header = self.Raid.groupHeaders[i]
			ForceCreateButtons(header, 5)
			header:Show()
		end
	else
		self.Raid.frame:Hide()
		ForceCreateButtons(self.Raid.frame.raidGroup, 40)
		self.Raid.frame.raidGroup:Show()

		local horiz, spacing = self:GetLayoutData()
		self.Raid.frame.petGroup:ClearAllPoints()
		if (horiz) then
			self.Raid.frame.petGroup:SetPoint("TOPLEFT", self.Raid.frame.raidGroup, "BOTTOMLEFT", 0, -spacing)
		else
			self.Raid.frame.petGroup:SetPoint("TOPLEFT", self.Raid.frame.raidGroup, "TOPRIGHT", spacing, 0)
		end
	end

	-- grouphoriz, spacing
	self:Raid_UpdateLayout()

	-- raidfilter
	local sortValue = Healcsi.db.profile.UnitFrames.raid.sortMethod -- class, role, name, group
	local groupBy, groupingOrder, sortMethod
	
	if (sortValue == "class") then
		groupBy = "CLASS"
		groupingOrder = "WARRIOR,DEATHKNIGHT,PALADIN,MONK,PRIEST,SHAMAN,DRUID,ROGUE,MAGE,WARLOCK,HUNTER,DEMONHUNTER"

	elseif (sortValue == "role") then
		groupBy = "ROLE"
		groupingOrder = "MAINTANK,MAINASSIST,TANK,HEALER,DAMAGER,NONE"

	elseif (sortValue == "name") then
		sortMethod = "NAME"

	elseif (sortValue == "group") then
		groupBy = "GROUP"
		groupingOrder = "1,2,3,4,5,6,7,8"
	end

	SortHeader(self.Raid.frame.raidGroup, groupBy, groupingOrder, sortMethod)
	SortHeader(self.Raid.frame.petGroup, groupBy, groupingOrder, sortMethod)
end

local function UpdateUnitSize()
	local header
	for _, header in pairs(UnitFrames.Raid.groupHeaders) do
		ForceUpdateHeader(header)
	end
end

local function SetAllHeadersAttribute(name, value)
	local _, header
	for _, header in pairs(UnitFrames.Raid.groupHeaders) do
		header:SetAttribute(name, value)
	end
end


Healcsi:RegisterEventCallback("UnitButtonConfigChanged", function()
	if (not(Editor.InCombat)) then
		UpdateUnitSize()
		SetAllHeadersAttribute("unitwidth", UnitButtons:GetUnitWidth())
		SetAllHeadersAttribute("unitheight", UnitButtons:GetUnitHeight())
	end
end)

Healcsi:RegisterEventCallback("UnitFrameConfigChanged", function()
	if (not(Editor.InCombat)) then
		UnitFrames:Raid_UpdateLayout()
		UnitFrames:Raid_UpdateFrames()
	end
end)

Healcsi:RegisterEventCallback("GlobalConfigRefresh", function()
	if (not(Editor.InCombat)) then
		UnitFrames:Raid_UpdateLayout()
		UnitFrames:Raid_UpdateFrames()

		UpdateUnitSize()

		SetAllHeadersAttribute("unitwidth", UnitButtons:GetUnitWidth())
		SetAllHeadersAttribute("unitheight", UnitButtons:GetUnitHeight())
	end
end)
